(* Content-type: application/mathematica *)

(*** Wolfram Notebook File ***)
(* http://www.wolfram.com/nb *)

(* CreatedBy='Mathematica 7.0' *)

(*CacheID: 234*)
(* Internal cache information:
NotebookFileLineBreakTest
NotebookFileLineBreakTest
NotebookDataPosition[       145,          7]
NotebookDataLength[    286219,       7418]
NotebookOptionsPosition[    283566,       7338]
NotebookOutlinePosition[    283902,       7353]
CellTagsIndexPosition[    283859,       7350]
WindowFrame->Normal*)

(* Beginning of Notebook Content *)
Notebook[{
Cell[TextData[{
 StyleBox["This is a ",
  FontSize->18],
 StyleBox["Mathematica",
  FontSize->18,
  FontSlant->"Italic"],
 StyleBox[" notebook which can be used to help verify and debug potentials \
implemented in LAMMPS. It works by taking output from a LAMMPS simulation in \
the form of coordinates and forces (from forces.dat) at a particular \
timestep. The forces that are due to a particular potential term are obtained \
by subtracting the total forces before that term was calculated from the \
total forces after that term was calculated. The total potential energy due \
to that term is also available in forces.dat. To test a potential term, you \
must implement the form of the potential in the notebook below exactly as it \
is implemented in LAMMPS, including the same values of any parameters that \
were used in the LAMMPS simulation. The advantage of using ",
  FontSize->18],
 StyleBox["Mathematica",
  FontSize->18,
  FontSlant->"Italic"],
 StyleBox[" is that, unlike in LAMMPS where you have to manually calculate \
the derivative of the potential function and implement that separately to get \
the forces on each atom, ",
  FontSize->18],
 StyleBox["Mathematica",
  FontSize->18,
  FontSlant->"Italic"],
 StyleBox[" will calculate the derivative automatically and analytically and \
output what it thinks the forces should be. If there is disagreement between \
the energy of a particular potential term, then it is most likely an error in \
either the LAMMPS or ",
  FontSize->18],
 StyleBox["Mathematica",
  FontSize->18,
  FontSlant->"Italic"],
 StyleBox[" implementations and both must be checked. If there is agreement \
in the energies but disagreement in the forces then it is almost certainly a \
problem with the form of the forces being calculated in LAMMPS. If there is \
agreement to within machine precision in both the forces and the energies \
then, congratulations, you (most likely) have a working potential.\n\nCare \
must be taken that all the parameters used in the potentials are the same. \
Differences in the parameters can easily cause false negatives in the \
agreement between LAMMPS and ",
  FontSize->18],
 StyleBox["Mathematica",
  FontSize->18,
  FontSlant->"Italic"],
 StyleBox[". Because the derivatives are calculated analytically in ",
  FontSize->18],
 StyleBox["Mathematica",
  FontSize->18,
  FontSlant->"Italic"],
 StyleBox[", it can be quite slow, so a small protein should probably be \
used. In ",
  FontSize->18],
 StyleBox["Mathematica",
  FontSize->18,
  FontSlant->"Italic"],
 StyleBox[", each block that contains information that is going to be used \
later must be executed. Simply \[OpenCurlyDoubleQuote]being there\
\[CloseCurlyDoubleQuote] is not enough. For a typical tests, most of the \
blocks can be executed sequentially from top to bottom. Exceptions are noted \
in the comments at the top of the blocks.\n\nMost of the initial ",
  FontSize->18],
 StyleBox["Mathematica",
  FontSize->18,
  FontSlant->"Italic"],
 StyleBox[" code was written by Aram Davtyan (davtyan.aram@gmail.com) and \
most of the initial commenting was done by Nick Schafer (nschafer@ucsd.edu). \
All hail Aram, ",
  FontSize->18],
 StyleBox["Mathematica-",
  FontSize->18,
  FontSlant->"Italic"],
 StyleBox["man.",
  FontSize->18]
}], "Text",
 CellChangeTimes->{{3.512331963053382*^9, 3.512332007741375*^9}, {
  3.5123320617638483`*^9, 3.512332082951221*^9}, {3.512332422522584*^9, 
  3.512332771412112*^9}, {3.512332821642769*^9, 3.5123328996107903`*^9}, {
  3.5123329719186697`*^9, 3.512333128943853*^9}, {3.512333239722148*^9, 
  3.512333308884424*^9}, {3.512333367244809*^9, 3.512333367433116*^9}, {
  3.5123334203677073`*^9, 3.512333511994548*^9}, {3.5129328225443172`*^9, 
  3.5129328269707613`*^9}, {3.5137106624870453`*^9, 3.513710679429757*^9}}],

Cell[BoxData[
 RowBox[{
  StyleBox[
   RowBox[{"(*", " ", 
    RowBox[{"COPY", " ", "THE", " ", "COORDINATES", " ", "FROM", " ", 
     RowBox[{"forces", ".", "dat"}], " ", "GENERATED", " ", "BY", " ", 
     "LAMMPS", " ", "HERE"}], " ", "*)"}], "Text"], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{"rca", "=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"43.28651118", ",", "5.32396973", ",", "36.98381128"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"41.36252886", ",", "8.37662806", ",", "35.16013115"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"37.76125400", ",", "7.79291999", ",", "34.42132200"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"37.22702550", ",", "5.76432491", ",", "30.48368063"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"33.40931396", ",", "7.19758304", ",", "30.86203449"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"32.30311515", ",", "10.61458959", ",", "29.82197821"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"29.86849163", ",", "12.59603068", ",", "31.97637288"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"26.45280811", ",", "11.62502098", ",", "32.86066220"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"23.33750332", ",", "12.65340764", ",", "31.61476696"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"21.05075264", ",", "14.88353463", ",", "33.94356157"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"18.24038969", ",", "12.69246821", ",", "33.43207474"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"16.92110649", ",", "8.99746297", ",", "33.81164222"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"15.17778286", ",", "6.50379035", ",", "31.27437876"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"14.72079454", ",", "2.95977187", ",", "30.20481303"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"14.73732123", ",", "0.39686130", ",", "27.53315539"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"12.51128840", ",", 
         RowBox[{"-", "3.02118480"}], ",", "27.42159587"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"15.47909622", ",", 
         RowBox[{"-", "5.01383984"}], ",", "27.89254377"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"15.49261133", ",", 
         RowBox[{"-", "3.93366906"}], ",", "31.57715101"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"12.45508818", ",", 
         RowBox[{"-", "3.19986785"}], ",", "33.82319835"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"14.76632867", ",", 
         RowBox[{"-", "0.82039890"}], ",", "35.65448592"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"15.66535096", ",", "2.97744458", ",", "35.18372773"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"18.93623629", ",", "4.28237838", ",", "34.55685421"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"20.83757064", ",", "7.49396472", ",", "34.24821076"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"20.57174697", ",", "8.54231807", ",", "30.63176524"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"19.54288786", ",", "5.28516675", ",", "29.65805085"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"22.91649838", ",", "4.62117855", ",", "29.56045150"}], 
        "}"}]}], "}"}]}], ";"}], "\[IndentingNewLine]", "\n", 
   RowBox[{
    RowBox[{"rcb", "=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"44.73316612", ",", "5.27066515", ",", "35.94734574"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"41.50966812", ",", "9.16981544", ",", "34.02995004"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"36.84865085", ",", "9.02635192", ",", "34.56779446"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"36.89518809", ",", "4.45904705", ",", "29.51855735"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"32.37598557", ",", "6.33039556", ",", "31.30357011"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"32.62165901", ",", "11.07490484", ",", "28.14374651"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"31.07587090", ",", "12.93328146", ",", "30.87458611"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"26.34687547", ",", "11.01638747", ",", "34.34874728"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"24.06033816", ",", "12.41584339", ",", "29.61519632"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"20.83898482", ",", "16.35246980", ",", "33.18498164"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"19.07053894", ",", "12.09925703", ",", "32.36659758"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"16.76689003", ",", "8.55329607", ",", "35.31007160"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"15.79462414", ",", "6.14505467", ",", "29.30769565"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"14.78822471", ",", "1.47121212", ",", "29.36223124"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"15.77076668", ",", "0.49410647", ",", "26.54820219"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"11.92770278", ",", 
         RowBox[{"-", "3.52665638"}], ",", "26.06891142"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"16.87326644", ",", 
         RowBox[{"-", "4.79632577"}], ",", "27.10473207"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"16.75560278", ",", 
         RowBox[{"-", "4.19255840"}], ",", "32.01277523"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"11.69903788", ",", 
         RowBox[{"-", "4.20541301"}], ",", "34.67575860"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"15.82826715", ",", 
         RowBox[{"-", "1.61867904"}], ",", "36.55388496"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"14.44638634", ",", "3.85165703", ",", "35.52421383"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"19.65209292", ",", "3.22734762", ",", "33.58455691"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"22.10213235", ",", "7.81602130", ",", "34.91299655"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"21.63667435", ",", "9.29780429", ",", "29.80671172"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"18.61476517", ",", "5.04528797", ",", "28.62638798"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"24.13250339", ",", "5.30728741", ",", "29.81121407"}], 
        "}"}]}], "}"}]}], ";"}], "\[IndentingNewLine]", "\n", 
   RowBox[{
    RowBox[{"ro", "=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"42.56190235", ",", "6.29718177", ",", "35.07052793"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"39.21071237", ",", "7.26899612", ",", "36.61099404"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"38.09081205", ",", "7.91076728", ",", "31.83555425"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"34.95652219", ",", "5.78429670", ",", "29.90652815"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"33.99912683", ",", "9.26794994", ",", "32.04415773"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"30.64591706", ",", "10.12847956", ",", "31.95545838"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"27.73466177", ",", "12.05806492", ",", "30.75170811"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"25.53147099", ",", "13.83114454", ",", "33.10010966"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"22.76650225", ",", "13.23331285", ",", "34.64409060"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"18.92763232", ",", "14.93176194", ",", "35.20624446"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"18.88890367", ",", "10.69334042", ",", "33.61274588"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"15.04259857", ",", "8.95941118", ",", "32.90947429"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"16.19999567", ",", "4.47126182", ",", "31.83795472"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"14.17657348", ",", "2.84224017", ",", "28.23275370"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"14.19450738", ",", 
         RowBox[{"-", "1.21913183"}], ",", "28.62916488"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"13.52871787", ",", 
         RowBox[{"-", "4.57409117"}], ",", "28.58074396"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"16.20442352", ",", 
         RowBox[{"-", "5.96478555"}], ",", "29.99040321"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"13.94752323", ",", 
         RowBox[{"-", "2.76384302"}], ",", "32.26010749"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"12.34924176", ",", 
         RowBox[{"-", "1.56897779"}], ",", "35.42947921"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"15.51917254", ",", "0.42799516", ",", "34.03210301"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"17.59195506", ",", "3.12933335", ",", "36.60625345"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"19.12765889", ",", "6.16540592", ",", "33.36125077"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"22.15647139", ",", "7.24635849", ",", "32.10113744"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"20.89982189", ",", "7.14962281", ",", "28.82054548"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"20.99881578", ",", "3.98162373", ",", "28.52239169"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"23.54877418", ",", "2.37459527", ",", "30.12300897"}], 
        "}"}]}], "}"}]}], ";"}], "\[IndentingNewLine]", "\n", 
   RowBox[{
    RowBox[{"rn", "=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0.00000000", ",", "0.00000000", ",", "0.00000000"}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{"42.06872885", ",", "7.28897110", ",", "36.05763485"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"39.23107050", ",", "8.17230083", ",", "34.37088782"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"37.32439357", ",", "6.34489805", ",", "32.13411292"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"34.96548790", ",", "6.76819062", ",", "30.85701376"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"32.52195100", ",", "9.21465040", ",", "29.91100000"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"30.89981366", ",", "12.09795914", ",", "30.93969035"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"27.86415847", ",", "12.01347249", ",", "32.82581592"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"24.43411754", ",", "11.93763782", ",", "31.94026062"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"21.83612379", ",", "14.11332454", ",", "32.68845537"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"19.46996787", ",", "13.33424466", ",", "33.34912731"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"17.19231000", ",", "10.46671730", ",", "33.66530585"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"16.04495245", ",", "7.25148875", ",", "32.19574925"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"14.66638848", ",", "4.39039566", ",", "30.41764892"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"14.83363792", ",", "1.17990664", ",", "28.69344074"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"13.27338353", ",", 
         RowBox[{"-", "1.70545038"}], ",", "27.25087128"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"14.40841541", ",", 
         RowBox[{"-", "4.13318071"}], ",", "27.53707070"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"15.35368646", ",", 
         RowBox[{"-", "4.07760388"}], ",", "30.09262081"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"13.64461894", ",", 
         RowBox[{"-", "3.63539446"}], ",", "33.02909238"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"14.09957768", ",", 
         RowBox[{"-", "1.83044673"}], ",", "34.81182527"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"15.25831238", ",", "1.61728377", ",", "35.62530613"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"17.60643498", ",", "3.86656173", ",", "34.47844586"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"20.23726973", ",", "6.18988045", ",", "34.56226010"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"20.40538106", ",", "8.27701188", ",", "32.10520520"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"19.78743708", ",", "6.51158201", ",", "30.28413090"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"21.64373346", ",", "5.06084205", ",", "29.80067111"}], 
        "}"}]}], "}"}]}], ";"}], "\[IndentingNewLine]", "\n", 
   RowBox[{
    RowBox[{"rcp", "=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"42.60177590", ",", "6.35374550", ",", "35.94108250"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"39.82549617", ",", "7.88390681", ",", "35.45223799"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"37.74149843", ",", "7.35402913", ",", "32.66594483"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"35.60103173", ",", "6.10755151", ",", "30.38732784"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"33.33868717", ",", "8.66516735", ",", "30.99708283"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"31.19901804", ",", "10.92418693", ",", "31.01310809"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"28.38084662", ",", "12.19515639", ",", "31.79106348"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"25.42496227", ",", "12.57485751", ",", "32.64473945"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"22.61682557", ",", "13.36363705", ",", "33.13444658"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"19.70879572", ",", "14.38411499", ",", "34.22868337"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"18.13853113", ",", "11.18242196", ",", "33.57926854"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"15.90842439", ",", "8.39923528", ",", "32.92578940"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"15.39852092", ",", "5.01850435", ",", "31.20393870"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"14.54998335", ",", "2.31976021", ",", "28.94394244"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"14.03996027", ",", 
         RowBox[{"-", "0.92511332"}], ",", "27.85875795"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"13.53531810", ",", 
         RowBox[{"-", "3.98794168"}], ",", "27.90435516"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"15.71510233", ",", 
         RowBox[{"-", "5.06525328"}], ",", "29.43183935"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"14.28282011", ",", 
         RowBox[{"-", "3.38571162"}], ",", "32.32420117"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"12.96425300", ",", 
         RowBox[{"-", "2.11717694"}], ",", "34.76916708"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"15.21926179", ",", "0.47282881", ",", "35.02307283"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"17.05244893", ",", "3.33286032", ",", "35.49304321"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"19.44449652", ",", "5.64155301", ",", "34.10053430"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"21.19846922", ",", "7.66084615", ",", "32.70913553"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"20.43527713", ",", "7.32983233", ",", "29.82154081"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"20.80303920", ",", "4.71069221", ",", "29.27056841"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"0.00000000", ",", "0.00000000", ",", "0.00000000"}], "}"}]}],
       "}"}]}], ";"}], "\[IndentingNewLine]", "\n", 
   RowBox[{
    RowBox[{"rh", "=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0.00000000", ",", "0.00000000", ",", "0.00000000"}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{"42.10025785", ",", "7.33565598", ",", "36.75961491"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"39.72602371", ",", "8.66837860", ",", "33.43544122"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"37.04229234", ",", "5.89488467", ",", "32.80296179"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"35.48435853", ",", "7.02956512", ",", "31.24518055"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"31.98855578", ",", "8.72932059", ",", "29.06558577"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"31.34537974", ",", "12.74082125", ",", "30.17987820"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"28.38451736", ",", "12.12378589", ",", "33.66494718"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"24.34710238", ",", "10.92407557", ",", "31.57232332"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"21.71454974", ",", "14.21926062", ",", "31.47085699"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"20.09946884", ",", "12.89151348", ",", "32.55999459"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"16.58626053", ",", "10.86018833", ",", "33.63841588"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"16.74314330", ",", "6.79854333", ",", "32.20805721"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"14.01938808", ",", "4.83085778", ",", "29.90559181"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"15.13501113", ",", "0.75736101", ",", "29.26651430"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"13.14789928", ",", 
         RowBox[{"-", "1.46931491"}], ",", "26.62906177"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"14.41474755", ",", 
         RowBox[{"-", "3.66079248"}], ",", "26.99133764"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"14.95877497", ",", 
         RowBox[{"-", "3.35125319"}], ",", "29.64306921"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"13.91419782", ",", 
         RowBox[{"-", "4.13703690"}], ",", "33.08158123"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"14.59671721", ",", 
         RowBox[{"-", "2.27207539"}], ",", "34.27952800"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"15.01656818", ",", "1.65475409", ",", "36.42492791"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"17.17212415", ",", "4.03126411", ",", "33.57979580"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"20.49362367", ",", "5.76818276", ",", "35.15880853"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"19.63211695", ",", "8.61188723", ",", "32.59467673"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"19.41216880", ",", "6.65592026", ",", "31.09167359"}], "}"}],
        ",", 
       RowBox[{"{", 
        RowBox[{"21.48687096", ",", "5.64902586", ",", "30.40446845"}], 
        "}"}]}], "}"}]}], ";"}]}]}]], "Input",
 CellChangeTimes->{{3.511798320180912*^9, 3.5117983202450314`*^9}, {
   3.5123207502160463`*^9, 3.512320781172483*^9}, {3.512320852239596*^9, 
   3.512320852366061*^9}, {3.512334055642219*^9, 3.5123340571629333`*^9}, {
   3.5123444087209787`*^9, 3.512344437571576*^9}, 3.512354398396409*^9, 
   3.5127587811106873`*^9, 3.512768157736011*^9, 3.512768927501512*^9, 
   3.512926414866827*^9, {3.512932849752404*^9, 3.5129329070154543`*^9}, 
   3.512941831241247*^9, 3.512947358604451*^9, 3.512947614171072*^9, 
   3.512947906850597*^9, 3.51301905855206*^9, 3.5139710601751127`*^9, 
   3.5139714183912582`*^9, 3.5139726096755457`*^9, {3.5140578908572063`*^9, 
   3.514057907625927*^9}, 3.514058537559854*^9}],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
   "COPY", " ", "THE", " ", "FORCE", " ", "ARRAY", " ", "THAT", " ", "WAS", 
    " ", "CALCULATED", " ", "UP", " ", "TO", " ", "BUT", " ", "NOT", " ", 
    "INCLUDING", " ", "THE", " ", "POTENTIAL", " ", "THAT", " ", "YOU", " ", 
    "ARE", " ", "TRYING", " ", "TO", " ", "TEST", " ", "HERE"}], " ", "*)"}], 
  "\[IndentingNewLine]", 
  RowBox[{"(*", " ", 
   RowBox[{
    RowBox[{"DON", "'"}], "T", " ", "FORGET", " ", "TO", " ", "RENAME", " ", 
    "THE", " ", "ARRAYS", " ", "BY", " ", "PUTTING", " ", "A", " ", 
    "\"\<0\>\"", " ", "AFTER", " ", "THEM"}], " ", "*)"}], 
  "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{"fca0", "=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0.72635904", ",", "3.47315378", ",", 
         RowBox[{"-", "3.22021174"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "5.14048588"}], ",", 
         RowBox[{"-", "4.59464084"}], ",", "12.32256415"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "6.62300082"}], ",", "0.87592813", ",", 
         RowBox[{"-", "11.10550343"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "14.84444764"}], ",", "10.62895248", ",", 
         "16.70343152"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "0.80366216"}], ",", 
         RowBox[{"-", "2.37898760"}], ",", "4.66719223"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "4.78088357"}], ",", 
         RowBox[{"-", "5.77210424"}], ",", "13.50705278"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "3.03005029"}], ",", "5.72040986", ",", 
         RowBox[{"-", "1.91458523"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"5.24604929", ",", "4.13992134", ",", "0.53074736"}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "5.42142074"}], ",", "26.96401296", ",", 
         "25.40469802"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "1.40895855"}], ",", "6.87221174", ",", 
         RowBox[{"-", "2.47180534"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "7.20820959"}], ",", "1.25980582", ",", "3.71181706"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "2.59086247"}], ",", "1.55320949", ",", "0.30035776"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"6.17440929", ",", "4.64225757", ",", 
         RowBox[{"-", "4.97066942"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"1.14227485", ",", "2.23071633", ",", "7.21309485"}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "4.57689531"}], ",", "4.79447265", ",", 
         RowBox[{"-", "8.70982781"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "2.52101331"}], ",", "12.24213286", ",", "3.93987983"}],
         "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"22.45972295", ",", "4.12968545", ",", "4.40860972"}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{"0.95863215", ",", 
         RowBox[{"-", "16.01862738"}], ",", 
         RowBox[{"-", "8.29412062"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "4.35021044"}], ",", "0.05499943", ",", "2.91857704"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"4.06470818", ",", "1.27833602", ",", "10.46992159"}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "2.59522161"}], ",", 
         RowBox[{"-", "12.13316265"}], ",", 
         RowBox[{"-", "2.50616283"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"3.63392633", ",", "2.37535212", ",", 
         RowBox[{"-", "1.16007839"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"17.98345296", ",", "2.19835143", ",", 
         RowBox[{"-", "10.29599429"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "0.13192788"}], ",", "12.82278096", ",", 
         RowBox[{"-", "2.85896635"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "8.78275780"}], ",", 
         RowBox[{"-", "2.71044849"}], ",", "6.40402791"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"16.01359529", ",", 
         RowBox[{"-", "3.84323034"}], ",", 
         RowBox[{"-", "0.34305255"}]}], "}"}]}], "}"}]}], ";"}], "\n", 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"fcb0", "=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0.14316551", ",", 
         RowBox[{"-", "2.04182729"}], ",", "2.83527938"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0.32094224", ",", "1.73011160", ",", 
         RowBox[{"-", "2.46516715"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"2.47477680", ",", 
         RowBox[{"-", "5.31708672"}], ",", 
         RowBox[{"-", "6.10813141"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0.46825856", ",", "1.84188853", ",", "1.36189356"}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{"4.78206731", ",", "0.84287152", ",", 
         RowBox[{"-", "0.95611130"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "0.82512908"}], ",", 
         RowBox[{"-", "1.19236169"}], ",", "4.34714949"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "1.86591363"}], ",", 
         RowBox[{"-", "0.52119564"}], ",", "1.70272838"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"2.83682153", ",", "1.60577522", ",", "2.02601848"}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "4.05056348"}], ",", "1.33124331", ",", "11.20503234"}],
         "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"3.18500806", ",", 
         RowBox[{"-", "0.59867163"}], ",", "1.01195641"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0.72901278", ",", 
         RowBox[{"-", "0.52094070"}], ",", 
         RowBox[{"-", "0.93567087"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "0.76118792"}], ",", "5.43036450", ",", 
         RowBox[{"-", "3.49068407"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "3.25580758"}], ",", "1.89347627", ",", "10.38053382"}],
         "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "0.13535745"}], ",", "2.98809343", ",", "1.69137522"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0.26470229", ",", 
         RowBox[{"-", "2.24895633"}], ",", 
         RowBox[{"-", "0.38357273"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "0.05830197"}], ",", "1.33443373", ",", 
         RowBox[{"-", "1.20823257"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "6.95339666"}], ",", "2.95029600", ",", "5.85608887"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0.08026303", ",", "1.71047496", ",", "4.48970822"}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "0.90097616"}], ",", 
         RowBox[{"-", "6.08599397"}], ",", 
         RowBox[{"-", "1.88858144"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "0.44791888"}], ",", "10.15854054", ",", 
         RowBox[{"-", "5.45739913"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"3.25105917", ",", "1.62538790", ",", "1.80979482"}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{"2.73321100", ",", 
         RowBox[{"-", "4.25516261"}], ",", 
         RowBox[{"-", "0.80287776"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"3.57932577", ",", "1.89414391", ",", "6.03780471"}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "4.31388000"}], ",", 
         RowBox[{"-", "4.77917702"}], ",", "4.85077866"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "8.72008151"}], ",", "1.10141136", ",", 
         RowBox[{"-", "3.70968278"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"4.55090543", ",", 
         RowBox[{"-", "0.97483231"}], ",", 
         RowBox[{"-", "1.27172017"}]}], "}"}]}], "}"}]}], ";"}], "\n", 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"fo0", "=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "2.34401452"}], ",", "0.50557251", ",", 
         RowBox[{"-", "2.80502371"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"3.20583822", ",", "0.93866857", ",", 
         RowBox[{"-", "9.58165686"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "3.05039469"}], ",", 
         RowBox[{"-", "3.23712493"}], ",", "3.12088363"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"4.75346080", ",", 
         RowBox[{"-", "0.31676354"}], ",", 
         RowBox[{"-", "4.49308392"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "4.49911826"}], ",", "2.55667677", ",", 
         RowBox[{"-", "9.85112185"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"6.05468499", ",", "3.67904237", ",", 
         RowBox[{"-", "6.32786759"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"11.53494714", ",", 
         RowBox[{"-", "4.76405426"}], ",", 
         RowBox[{"-", "1.29175410"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"7.31323234", ",", 
         RowBox[{"-", "3.26688117"}], ",", 
         RowBox[{"-", "8.19810373"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "1.49087464"}], ",", 
         RowBox[{"-", "10.67560490"}], ",", 
         RowBox[{"-", "14.29567544"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"7.24056610", ",", 
         RowBox[{"-", "8.07298830"}], ",", 
         RowBox[{"-", "8.08137726"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"3.09797938", ",", 
         RowBox[{"-", "1.94151429"}], ",", 
         RowBox[{"-", "0.95189311"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "2.77047738"}], ",", 
         RowBox[{"-", "3.46547112"}], ",", 
         RowBox[{"-", "5.83395204"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "0.46616864"}], ",", "1.26608834", ",", "0.61042594"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "1.19956499"}], ",", 
         RowBox[{"-", "1.36106144"}], ",", 
         RowBox[{"-", "5.32291044"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0.11440015", ",", 
         RowBox[{"-", "10.62593487"}], ",", "0.20586232"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "15.75745117"}], ",", 
         RowBox[{"-", "12.18123265"}], ",", 
         RowBox[{"-", "7.49751621"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "4.74761404"}], ",", 
         RowBox[{"-", "5.04476810"}], ",", "2.27646846"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"2.34828286", ",", "3.59907124", ",", 
         RowBox[{"-", "5.85825121"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0.87152031", ",", 
         RowBox[{"-", "1.23541149"}], ",", "4.51443209"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "2.57246283"}], ",", "2.51182386", ",", 
         RowBox[{"-", "1.63868736"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "3.06284439"}], ",", "0.85858076", ",", 
         RowBox[{"-", "0.87875670"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "7.19428137"}], ",", 
         RowBox[{"-", "4.73375260"}], ",", 
         RowBox[{"-", "0.22348333"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "2.93193968"}], ",", 
         RowBox[{"-", "3.80029754"}], ",", "5.10496571"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"1.47320232", ",", "7.37371057", ",", 
         RowBox[{"-", "0.47403253"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0.71805753", ",", 
         RowBox[{"-", "17.45661614"}], ",", 
         RowBox[{"-", "9.48777536"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"2.65690703", ",", 
         RowBox[{"-", "1.81755340"}], ",", "1.68057979"}], "}"}]}], "}"}]}], 
    ";"}]}]}]], "Input",
 CellChangeTimes->{{3.5117983675998487`*^9, 3.511798376364098*^9}, {
   3.512320809937397*^9, 3.512320872585106*^9}, {3.512344502544957*^9, 
   3.512344504918571*^9}, 3.512354345077203*^9, {3.512354421910468*^9, 
   3.512354434568941*^9}, {3.512758729598061*^9, 3.512758742074319*^9}, 
   3.512761498935484*^9, {3.5127615296205177`*^9, 3.512761540441431*^9}, {
   3.5127681846423197`*^9, 3.512768192596807*^9}, {3.512768980350909*^9, 
   3.512768989418456*^9}, {3.5129264465569057`*^9, 3.512926503817889*^9}, {
   3.513542704308346*^9, 3.513542712466998*^9}, {3.513970982475444*^9, 
   3.513970987182385*^9}, {3.513971089456326*^9, 3.513971094790489*^9}, {
   3.513971532189207*^9, 3.513971538546219*^9}, {3.513972697840103*^9, 
   3.5139727024436502`*^9}, {3.513972819381029*^9, 3.513972824936294*^9}, {
   3.513975148106098*^9, 3.5139751570610743`*^9}, {3.513976937668439*^9, 
   3.5139769433831797`*^9}, {3.514048657984468*^9, 3.5140486715329437`*^9}, {
   3.514057361828824*^9, 3.5140573681383343`*^9}, {3.514058090504964*^9, 
   3.514058096978615*^9}, {3.514058565914123*^9, 3.514058571406705*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
   "COPY", " ", "THE", " ", "FORCE", " ", "ARRAY", " ", "THAT", " ", "WAS", 
    " ", "CALCULATED", " ", "INCLUDING", " ", "THE", " ", "POTENTIAL", " ", 
    "THAT", " ", "YOU", " ", "ARE", " ", "TRYING", " ", "TO", " ", "TEST", 
    " ", "HERE"}], " ", "*)"}], "\[IndentingNewLine]", 
  RowBox[{"(*", " ", 
   RowBox[{
    RowBox[{"DON", "'"}], "T", " ", "FORGET", " ", "TO", " ", "RENAME", " ", 
    "THE", " ", "ARRAYS", " ", "BY", " ", "PUTTING", " ", "A", " ", 
    "\"\<1\>\"", " ", "AFTER", " ", "THEM"}], " ", "*)"}], 
  "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{"fca1", "=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0.72635904", ",", "3.47315378", ",", 
         RowBox[{"-", "3.22021174"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "5.14048588"}], ",", 
         RowBox[{"-", "4.59464084"}], ",", "12.32256415"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "6.62739257"}], ",", "0.87819851", ",", 
         RowBox[{"-", "11.10920417"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "14.84444764"}], ",", "10.62895248", ",", 
         "16.70343152"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "0.80366216"}], ",", 
         RowBox[{"-", "2.37898760"}], ",", "4.66719223"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "4.77074631"}], ",", 
         RowBox[{"-", "5.77638213"}], ",", "13.51675218"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "3.03005029"}], ",", "5.72040986", ",", 
         RowBox[{"-", "1.91458523"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"2.66818631", ",", 
         RowBox[{"-", "1.05231929"}], ",", "3.39562236"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "11.93128561"}], ",", "13.71310792", ",", 
         "32.44173482"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "1.54367342"}], ",", "15.85016753", ",", 
         RowBox[{"-", "0.51222301"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "15.20369595"}], ",", "12.30564033", ",", 
         "8.78167365"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0.86300606", ",", "13.97191712", ",", "2.26919199"}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "2.54946236"}], ",", 
         RowBox[{"-", "6.03359508"}], ",", "16.44892598"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"1.69329249", ",", "2.41063265", ",", "4.92798963"}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "4.25572691"}], ",", "5.45163539", ",", 
         RowBox[{"-", "8.92679757"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "2.51980620"}], ",", "12.24488974", ",", "3.86830302"}],
         "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"22.45972295", ",", "4.12968545", ",", "4.40860972"}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{"1.19647164", ",", 
         RowBox[{"-", "18.13236373"}], ",", 
         RowBox[{"-", "7.70293410"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "4.59321654"}], ",", 
         RowBox[{"-", "0.55340080"}], ",", "3.42497320"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"4.07813642", ",", "0.16354545", ",", "12.07705639"}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "5.00340815"}], ",", 
         RowBox[{"-", "14.70381666"}], ",", 
         RowBox[{"-", "8.69873026"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"2.73743888", ",", "3.74737083", ",", 
         RowBox[{"-", "1.09384472"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"18.36528087", ",", "7.57113783", ",", 
         RowBox[{"-", "16.43078550"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "1.63245023"}], ",", "2.22826663", ",", 
         RowBox[{"-", "13.28759193"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "6.30920825"}], ",", 
         RowBox[{"-", "9.28379027"}], ",", "1.11991757"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"16.38961827", ",", 
         RowBox[{"-", "4.44582986"}], ",", 
         RowBox[{"-", "1.56382765"}]}], "}"}]}], "}"}]}], ";"}], "\n", 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"fcb1", "=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0.14316551", ",", 
         RowBox[{"-", "2.04182729"}], ",", "2.83527938"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0.32094224", ",", "1.73011160", ",", 
         RowBox[{"-", "2.46516715"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"2.46903129", ",", 
         RowBox[{"-", "5.31507921"}], ",", 
         RowBox[{"-", "6.11413007"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0.46825856", ",", "1.84188853", ",", "1.36189356"}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{"4.36275249", ",", "1.21768975", ",", 
         RowBox[{"-", "0.84588131"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "0.82512908"}], ",", 
         RowBox[{"-", "1.19236169"}], ",", "4.34714949"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "1.86591363"}], ",", 
         RowBox[{"-", "0.52119564"}], ",", "1.70272838"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0.50659735", ",", 
         RowBox[{"-", "1.65743907"}], ",", "4.38400534"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "4.05056348"}], ",", "1.33124331", ",", "11.20503234"}],
         "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"3.26885971", ",", 
         RowBox[{"-", "1.34024507"}], ",", "0.65683890"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0.72901278", ",", 
         RowBox[{"-", "0.52094070"}], ",", 
         RowBox[{"-", "0.93567087"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"8.56734854", ",", "8.43955133", ",", 
         RowBox[{"-", "6.08283863"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "3.25580758"}], ",", "1.89347627", ",", "10.38053382"}],
         "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "0.13535745"}], ",", "2.98809343", ",", "1.69137522"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"2.68502229", ",", "1.11132711", ",", "1.51314069"}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "0.05830197"}], ",", "1.33443373", ",", 
         RowBox[{"-", "1.20823257"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "6.95339666"}], ",", "2.95029600", ",", "5.85608887"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "0.33609211"}], ",", "3.41092269", ",", "4.34489110"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "0.90160413"}], ",", 
         RowBox[{"-", "6.08690952"}], ",", 
         RowBox[{"-", "1.88297305"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "1.14573327"}], ",", "13.04340549", ",", 
         RowBox[{"-", "9.45792582"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0.63755856", ",", 
         RowBox[{"-", "1.71821522"}], ",", "0.88584534"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"3.79826865", ",", 
         RowBox[{"-", "2.62582960"}], ",", "1.66574638"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"9.18136013", ",", "8.83488238", ",", "1.42231575"}], "}"}], 
       ",", 
       RowBox[{"{", 
        RowBox[{"5.95066327", ",", 
         RowBox[{"-", "4.12315829"}], ",", "0.87973115"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "8.11055740"}], ",", 
         RowBox[{"-", "2.86190709"}], ",", 
         RowBox[{"-", "7.87198262"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"4.83494835", ",", "3.05199727", ",", "1.39830866"}], "}"}]}],
       "}"}]}], ";"}], "\n", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"fo1", "=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "2.34401452"}], ",", "0.50557251", ",", 
         RowBox[{"-", "2.80502371"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"3.20583822", ",", "0.93866857", ",", 
         RowBox[{"-", "9.58165686"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "3.05039469"}], ",", 
         RowBox[{"-", "3.23712493"}], ",", "3.12088363"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"4.75346080", ",", 
         RowBox[{"-", "0.31676354"}], ",", 
         RowBox[{"-", "4.49308392"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "4.49911826"}], ",", "2.55667677", ",", 
         RowBox[{"-", "9.85112185"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"6.05468499", ",", "3.67904237", ",", 
         RowBox[{"-", "6.32786759"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"11.53494714", ",", 
         RowBox[{"-", "4.76405426"}], ",", 
         RowBox[{"-", "1.29175410"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"7.31323234", ",", 
         RowBox[{"-", "3.26688117"}], ",", 
         RowBox[{"-", "8.19810373"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "1.49087464"}], ",", 
         RowBox[{"-", "10.67560490"}], ",", 
         RowBox[{"-", "14.29567544"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"7.24056610", ",", 
         RowBox[{"-", "8.07298830"}], ",", 
         RowBox[{"-", "8.08137726"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"3.09797938", ",", 
         RowBox[{"-", "1.94151429"}], ",", 
         RowBox[{"-", "0.95189311"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "2.77047738"}], ",", 
         RowBox[{"-", "3.46547112"}], ",", 
         RowBox[{"-", "5.83395204"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "0.46616864"}], ",", "1.26608834", ",", "0.61042594"}], 
        "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "1.19956499"}], ",", 
         RowBox[{"-", "1.36106144"}], ",", 
         RowBox[{"-", "5.32291044"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0.11440015", ",", 
         RowBox[{"-", "10.62593487"}], ",", "0.20586232"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "15.75745117"}], ",", 
         RowBox[{"-", "12.18123265"}], ",", 
         RowBox[{"-", "7.49751621"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "4.74761404"}], ",", 
         RowBox[{"-", "5.04476810"}], ",", "2.27646846"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"2.34828286", ",", "3.59907124", ",", 
         RowBox[{"-", "5.85825121"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0.87152031", ",", 
         RowBox[{"-", "1.23541149"}], ",", "4.51443209"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "2.57246283"}], ",", "2.51182386", ",", 
         RowBox[{"-", "1.63868736"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "3.06284439"}], ",", "0.85858076", ",", 
         RowBox[{"-", "0.87875670"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "7.19428137"}], ",", 
         RowBox[{"-", "4.73375260"}], ",", 
         RowBox[{"-", "0.22348333"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
         RowBox[{"-", "2.93193968"}], ",", 
         RowBox[{"-", "3.80029754"}], ",", "5.10496571"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"1.47320232", ",", "7.37371057", ",", 
         RowBox[{"-", "0.47403253"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"0.71805753", ",", 
         RowBox[{"-", "17.45661614"}], ",", 
         RowBox[{"-", "9.48777536"}]}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"2.65690703", ",", 
         RowBox[{"-", "1.81755340"}], ",", "1.68057979"}], "}"}]}], "}"}]}], 
    ";"}], "\n"}]}]], "Input",
 CellChangeTimes->{{3.511798392167981*^9, 3.511798396976932*^9}, {
   3.512321201022827*^9, 3.512321233665188*^9}, {3.512344581438944*^9, 
   3.512344600547886*^9}, {3.5123544790394363`*^9, 3.51235448468552*^9}, {
   3.512758696860867*^9, 3.512758707968214*^9}, 3.5127614811263247`*^9, {
   3.51276154399411*^9, 3.512761547824424*^9}, {3.51276821187714*^9, 
   3.512768217069875*^9}, {3.5127690067473717`*^9, 3.5127690160493317`*^9}, {
   3.512926510406075*^9, 3.512926532111826*^9}, {3.512926607582354*^9, 
   3.5129266150663958`*^9}, {3.513542755673101*^9, 3.513542767577026*^9}, {
   3.5139709625096407`*^9, 3.513970966439209*^9}, {3.513971113417137*^9, 
   3.513971117540799*^9}, {3.5139715172787848`*^9, 3.513971520212365*^9}, {
   3.513972725369409*^9, 3.513972730052915*^9}, {3.513972835148752*^9, 
   3.513972839480075*^9}, {3.513975132237247*^9, 3.513975136461624*^9}, {
   3.5139769598185253`*^9, 3.513976963401061*^9}, {3.51404861186419*^9, 
   3.514048622637528*^9}, {3.514057342127749*^9, 3.514057347137878*^9}, {
   3.514058063518104*^9, 3.5140580730712357`*^9}, {3.5140585853290854`*^9, 
   3.51405858968538*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
   "THIS", " ", "DIRECTORY", " ", "SHOULD", " ", "CONTAIN", " ", "THE", " ", 
    "PARAMETER", " ", "FILES", " ", "THAT", " ", "WERE", " ", "USED", " ", 
    "IN", " ", "THE", " ", "LAMMPS", " ", "SIMULATION"}], " ", "*)"}], 
  "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
   "SetDirectory", "[", 
    "\"\</home/nschafer/lammpsamw/amnesiacmd/debugging/h4/\>\"", "]"}], 
   ";"}]}]], "Input",
 CellChangeTimes->{{3.5117197480160503`*^9, 3.5117197494880667`*^9}, {
  3.512309479799055*^9, 3.5123095136667624`*^9}, {3.51232123983422*^9, 
  3.5123212632084303`*^9}, {3.512920262916583*^9, 3.512920284558443*^9}, {
  3.5129203286558237`*^9, 3.512920332856193*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
    RowBox[{
     RowBox[{"I", "'"}], "M", " ", "NOT", " ", "SURE", " ", "WHAT", " ", 
     "THIS", " ", "IS", " ", "FOR"}], ",", " ", 
    RowBox[{
    "BUT", " ", "IT", " ", "SHOULD", " ", "NOT", " ", "BE", " ", "EXECUTED", 
     " ", "ON", " ", "A", " ", "TYPICAL", " ", "TEST"}]}], " ", "*)"}], 
  "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{"MyClear", "[", "A_", "]"}], ":=", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"B", "=", 
        RowBox[{"{", "}"}]}], "}"}], ",", 
      RowBox[{
       RowBox[{"For", "[", 
        RowBox[{
         RowBox[{"i", "=", "1"}], ",", 
         RowBox[{"i", "\[LessEqual]", 
          RowBox[{"Length", "[", "A", "]"}]}], ",", 
         RowBox[{
          RowBox[{"B", "=", 
           RowBox[{"Append", "[", 
            RowBox[{"B", ",", 
             RowBox[{"A", "[", 
              RowBox[{"[", 
               RowBox[{"i", ",", 
                RowBox[{
                 RowBox[{"-", "3"}], ";;"}]}], "]"}], "]"}]}], " ", "]"}]}], 
          ";", 
          RowBox[{"i", "++"}]}]}], "]"}], ";", 
       RowBox[{"Return", "[", "B", "]"}]}]}], "]"}]}], "\[IndentingNewLine]", 
   
   RowBox[{
    RowBox[{"rca", " ", "=", 
     RowBox[{"MyClear", "[", "rca", "]"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"rcb", " ", "=", 
     RowBox[{"MyClear", "[", "rcb", "]"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"ro", " ", "=", 
     RowBox[{"MyClear", "[", "ro", "]"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"fca", "=", 
     RowBox[{"MyClear", "[", 
      RowBox[{"fca1", "-", "fca0"}], "]"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"fcb", "=", 
     RowBox[{"MyClear", "[", 
      RowBox[{"fcb1", "-", "fcb0"}], "]"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"fo", "=", 
     RowBox[{"MyClear", "[", 
      RowBox[{"fo1", "-", "fo0"}], "]"}]}], ";"}]}]}]], "Input",
 CellOpen->False,
 CellChangeTimes->{{3.4689395345104523`*^9, 3.468939772553584*^9}, {
  3.468939885126484*^9, 3.468939950867543*^9}, {3.512321272798214*^9, 
  3.5123212954290247`*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
   "THIS", " ", "BLOCK", " ", "ISOLATES", " ", "THE", " ", "FORCES", " ", 
    "THAT", " ", "ARE", " ", "DUE", " ", "TO", " ", "THE", " ", "PARTICULAR", 
    " ", "POTENTIAL", " ", "THAT", " ", "YOU", " ", "ARE", " ", "TESTING"}], 
   " ", "*)"}], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{"fca", "=", 
    RowBox[{"fca1", "-", "fca0"}]}], " ", 
   RowBox[{"(*", " ", 
    RowBox[{
    "fca1", " ", "IS", " ", "THE", " ", "TOTAL", " ", "CUMMULATIVE", " ", 
     "FORCE", " ", "FOR", " ", "ALL", " ", "POTENTIALS", " ", "UP", " ", "TO",
      " ", "AND", " ", "INCLUDING", " ", "THE", " ", "ONE", " ", "YOU", " ", 
     "ARE", " ", "TRYING", " ", "TO", " ", "TEST"}], " ", "*)"}], 
   "\[IndentingNewLine]", 
   RowBox[{"fcb", "=", 
    RowBox[{"fcb1", "-", "fcb0"}]}], " ", 
   RowBox[{"(*", " ", 
    RowBox[{
    "fca0", " ", "IS", " ", "THE", " ", "TOTAL", " ", "CUMMULATIVE", " ", 
     "FORCE", " ", "FOR", " ", "ALL", " ", "POTENTIALS", " ", "UP", " ", "TO",
      " ", "BUT", " ", "NOT", " ", "INCLUDING", " ", "THE", " ", "ONE", " ", 
     "YOU", " ", "ARE", " ", "TRYING", " ", "TO", " ", "TEST"}], " ", "*)"}], 
   "\[IndentingNewLine]", 
   RowBox[{"fo", "=", 
    RowBox[{"fo1", "-", "fo0", " ", 
     RowBox[{"(*", " ", 
      RowBox[{
      "THEREFORE", " ", "fca", " ", "IS", " ", "THE", " ", "TOTAL", " ", 
       "FORCE", " ", "DUE", " ", "TO", " ", "THE", " ", "POTENTIAL", " ", 
       "THAT", " ", "YOU", " ", "ARE", " ", "TRYING", " ", "TO", " ", "TEST", 
       " ", "AND", " ", "LIKEWISE", " ", "FOR", " ", "THE", " ", "OTHER", " ",
        "ATOMS"}], " ", "*)"}]}]}]}]}]], "Input",
 CellChangeTimes->{{3.468954040626171*^9, 3.468954069185055*^9}, {
  3.4716984564339027`*^9, 3.471698467245936*^9}, {3.512316206321393*^9, 
  3.5123163497499866`*^9}, {3.512320148247431*^9, 3.512320154373982*^9}, {
  3.512321321857375*^9, 3.512321330960184*^9}, {3.5123536916185923`*^9, 
  3.512353694149489*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.0043917499999999166`"}], ",", "0.0022703800000000163`", 
     ",", 
     RowBox[{"-", "0.003700739999999314`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.010137260000000481`", ",", 
     RowBox[{"-", "0.004277890000000006`"}], ",", "0.009699399999998803`"}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "2.5778629800000004`"}], ",", 
     RowBox[{"-", "5.19224063`"}], ",", "2.864875`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "6.5098648699999995`"}], ",", 
     RowBox[{"-", "13.250905040000001`"}], ",", "7.037036799999999`"}], "}"}],
    ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.13471487000000004`"}], ",", "8.97795579`", ",", 
     "1.95958233`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "7.99548636`"}], ",", "11.045834509999999`", ",", 
     "5.0698565900000006`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"3.45386853`", ",", "12.41870763`", ",", "1.96883423`"}], "}"}], 
   ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "8.72387165`"}], ",", 
     RowBox[{"-", "10.67585265`"}], ",", "21.4195954`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.55101764`", ",", "0.17991632000000024`", ",", 
     RowBox[{"-", "2.28510522`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.32116840000000035`", ",", "0.6571627399999995`", ",", 
     RowBox[{"-", "0.21696975999999957`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.0012071099999997337`", ",", "0.0027568799999997395`", ",", 
     RowBox[{"-", "0.07157681000000027`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.23783948999999993`", ",", 
     RowBox[{"-", "2.11373635`"}], ",", "0.591186519999999`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.24300610000000056`"}], ",", 
     RowBox[{"-", "0.60840023`"}], ",", "0.50639616`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.013428239999999647`", ",", 
     RowBox[{"-", "1.11479057`"}], ",", "1.607134799999999`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "2.4081865400000004`"}], ",", 
     RowBox[{"-", "2.57065401`"}], ",", 
     RowBox[{"-", "6.1925674299999995`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.89648745`"}], ",", "1.3720187099999999`", ",", 
     "0.06623366999999991`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.3818279099999984`", ",", "5.372786399999999`", ",", 
     RowBox[{"-", "6.1347912099999995`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "1.5005223499999998`"}], ",", 
     RowBox[{"-", "10.594514329999999`"}], ",", 
     RowBox[{"-", "10.428625579999999`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"2.4735495500000004`", ",", 
     RowBox[{"-", "6.573341780000001`"}], ",", 
     RowBox[{"-", "5.28411034`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.37602297999999834`", ",", 
     RowBox[{"-", "0.60259952`"}], ",", 
     RowBox[{"-", "1.2207751`"}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.5140482971470737`*^9, 3.514048677991839*^9, 
  3.514051144699005*^9, 3.514057376008007*^9, 3.514058103413273*^9, 
  3.5140585955530357`*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.005745509999999676`"}], ",", "0.002007509999999435`", 
     ",", 
     RowBox[{"-", "0.005998659999999489`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.4193148200000003`"}], ",", "0.3748182299999999`", ",", 
     "0.11022999`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "2.33022418`"}], ",", 
     RowBox[{"-", "3.2632142899999996`"}], ",", "2.35798686`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.08385165000000017`", ",", 
     RowBox[{"-", "0.7415734399999999`"}], ",", 
     RowBox[{"-", "0.35511751`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"9.328536459999999`", ",", "3.009186830000001`", ",", 
     RowBox[{"-", "2.5921545600000004`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"2.4203200000000002`", ",", "3.36028344`", ",", "1.89671342`"}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.41635514`"}], ",", "1.70044773`", ",", 
     RowBox[{"-", "0.1448171199999999`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.0006279699999999222`"}], ",", 
     RowBox[{"-", "0.0009155500000002093`"}], ",", "0.00560839000000013`"}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.69781439`"}], ",", "2.884864949999999`", ",", 
     RowBox[{"-", "4.00052669`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "2.61350061`"}], ",", 
     RowBox[{"-", "3.34360312`"}], ",", 
     RowBox[{"-", "0.92394948`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"1.06505765`", ",", "1.6293330100000003`", ",", "2.46862414`"}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"5.602034359999999`", ",", "6.940738469999999`", ",", 
     RowBox[{"-", "4.6154889599999995`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"10.26454327`", ",", "0.6560187299999996`", ",", 
     RowBox[{"-", "3.9710475099999996`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.6095241100000006`", ",", 
     RowBox[{"-", "3.96331845`"}], ",", 
     RowBox[{"-", "4.162299839999999`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.2840429200000001`", ",", "4.02682958`", ",", "2.67002883`"}], 
    "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.5140482971470737`*^9, 3.514048677991839*^9, 
  3.514051144699005*^9, 3.514057376008007*^9, 3.514058103413273*^9, 
  3.514058595555526*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.5140482971470737`*^9, 3.514048677991839*^9, 
  3.514051144699005*^9, 3.514057376008007*^9, 3.514058103413273*^9, 
  3.514058595557502*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
   "THIS", " ", "MAPS", " ", "EACH", " ", "LETTER", " ", "OF", " ", "THE", 
    " ", "ALPHABET", " ", "TO", " ", "THE", " ", "ONE", " ", "LETTER", " ", 
    "AMINO", " ", "ACID", " ", "CODE"}], " ", "*)"}], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{"resmap", "=", 
     RowBox[{"{", 
      RowBox[{
      "0", ",", "0", ",", "4", ",", "3", ",", "6", ",", "13", ",", "7", ",", 
       "8", ",", "9", ",", "0", ",", "11", ",", "10", ",", "12", ",", "2", 
       ",", "0", ",", "14", ",", "5", ",", "1", ",", "15", ",", "16", ",", 
       "0", ",", "19", ",", "17", ",", "0", ",", "18", ",", "0"}], "}"}]}], 
    ";"}], "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{
    "THIS", " ", "IS", " ", "THE", " ", "ONE", " ", "LETTER", " ", "SEQUENCE",
      " ", "OF", " ", "THE", " ", "PROTEIN", " ", "THAT", " ", "YOU", " ", 
     "ARE", " ", "USING", " ", "TO", " ", "TEST"}], " ", "*)"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"seq", "=", "\"\<SGRGKGGKGLGKGGAKRARKVLRDNQ\>\""}], ";"}], " ", 
   RowBox[{"(*", " ", 
    RowBox[{
    "THIS", " ", "PARTICULAR", " ", "SEQUENCE", " ", "IS", " ", "THAT", " ", 
     "OF", " ", "H4", " ", "HISTONE", " ", "TAIL", " ", "PROTEIN"}], " ", 
    "*)"}], "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{
    "THESE", " ", "LINES", " ", "CREATE", " ", "AN", " ", "ARRAY", " ", "se", 
     " ", "THAT", " ", "CONVERTS", " ", "THE", " ", "INPUT", " ", "SEQUENCE", 
     " ", "ABOVE", " ", "USING", " ", "THE", " ", "resmap", " ", "ARRAY", " ",
      "FROM", " ", "A", " ", "ONE", " ", "LETTER", " ", "CODE", " ", "TO", 
     " ", "A", " ", "NUMBER", " ", "CODE"}], " ", "*)"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"chseq", "=", 
     RowBox[{"Characters", "[", "seq", "]"}]}], ";"}], "\[IndentingNewLine]", 
   
   RowBox[{
    RowBox[{"M", "=", 
     RowBox[{"Length", "[", "chseq", "]"}]}], ";"}], " ", 
   RowBox[{"(*", " ", 
    RowBox[{
    "CALCULATES", " ", "THE", " ", "LENGTH", " ", "OF", " ", "THE", " ", 
     "INPUT", " ", "SEQUENCE", " ", "DEFINED", " ", "ABOVE"}], " ", "*)"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"se", "=", 
     RowBox[{"Table", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"index", "=", 
         RowBox[{
          RowBox[{
           RowBox[{"ToCharacterCode", "[", "seq", "]"}], "[", 
           RowBox[{"[", "i", "]"}], "]"}], "-", 
          RowBox[{
           RowBox[{"ToCharacterCode", "[", "\"\<A\>\"", "]"}], "[", 
           RowBox[{"[", "1", "]"}], "]"}], "+", "1"}]}], ";", 
        RowBox[{
         RowBox[{"resmap", "[", 
          RowBox[{"[", "index", "]"}], "]"}], "+", "1"}]}], ",", 
       RowBox[{"{", 
        RowBox[{"i", ",", "1", ",", "M"}], "}"}]}], "]"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{
    "THESE", " ", "ARE", " ", "TWO", " ", "POSSIBLE", " ", "CHOICES", " ", 
     "OF", " ", "COEFFICIENTS", " ", "FOR", " ", "CALCULATING", " ", "THE", 
     " ", "POSITIONS", " ", "OF", " ", "THE", " ", "MISSING", " ", "ATOMS"}], 
    " ", "*)"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"an", "=", "0.4831806"}], ";", 
    RowBox[{"bn", "=", "0.7032820"}], ";", 
    RowBox[{"cn", "=", 
     RowBox[{"-", "0.1864262"}]}], ";"}], "\n", 
   RowBox[{
    RowBox[{"ap", "=", "0.4436538"}], ";", 
    RowBox[{"bp", "=", "0.2352006"}], ";", 
    RowBox[{"cp", "=", "0.3211455"}], ";"}], "\n", 
   RowBox[{
    RowBox[{"ah", "=", "0.8409657"}], ";", 
    RowBox[{"bh", "=", "0.8929599"}], ";", 
    RowBox[{"ch", "=", 
     RowBox[{"-", "0.7338894"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{
     RowBox[{
     "SINCE", " ", "THESE", " ", "ARE", " ", "EXECUTED", " ", "LAST"}], ",", 
     " ", 
     RowBox[{
     "THESE", " ", "ARE", " ", "THE", " ", "VALUES", " ", "THAT", " ", "ARE", 
      " ", "USED", " ", "AND", " ", "SHOULD", " ", "MATCH", " ", "THOSE", " ",
       "IN", " ", "LAMMPS"}]}], " ", "*)"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"an", "=", "0.483"}], ";", " ", 
    RowBox[{"bn", "=", " ", "0.703"}], ";", " ", 
    RowBox[{"cn", "=", " ", 
     RowBox[{"-", "0.186"}]}], ";"}], "\n", 
   RowBox[{
    RowBox[{"ap", "=", "0.444"}], ";", " ", 
    RowBox[{"bp", "=", " ", "0.235"}], ";", " ", 
    RowBox[{"cp", "=", " ", "0.321"}], ";"}], "\n", 
   RowBox[{
    RowBox[{"ah", "=", "0.841"}], ";", " ", 
    RowBox[{"bh", "=", " ", "0.893"}], ";", " ", 
    RowBox[{"ch", "=", 
     RowBox[{"-", "0.734"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{
    "THESE", " ", "LINES", " ", "DEFINE", " ", "PARAMETERS", " ", "EITHER", 
     " ", "DIRECTLY", " ", "OR", " ", "BY", " ", "READING", " ", "FILES", " ",
      "THAT", " ", "ARE", " ", "IN", " ", "THE", " ", "DIRECTORY", " ", 
     "SPECIFIED", " ", "ABOVE"}], " ", "*)"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"rNOzero", "=", "2.98"}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"rHOzero", "=", "2.06"}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"\[Sigma]NO", "=", "0.68"}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"\[Sigma]HO", "=", "0.76"}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"dnu", "=", "12.0"}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"pref", "=", "0.5"}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"cutoff", " ", "=", " ", "7.0"}], ";"}], "\[IndentingNewLine]", 
   RowBox[{"(*", 
    RowBox[{
     RowBox[{"hbscl", "=", 
      RowBox[{"{", 
       RowBox[{
       "1.0", ",", " ", "1.0", ",", " ", "1.0", ",", " ", "1.0", ",", " ", 
        "4.0", ",", " ", "1.0", ",", " ", "1.0", ",", " ", "1.0", ",", " ", 
        "4.0", ",", " ", "4.0", ",", " ", "1.0", ",", " ", "1.0", ",", " ", 
        "1.0", ",", " ", "1.0", ",", " ", "1.0", ",", " ", "1.0", ",", " ", 
        "1.0"}], "}"}]}], ";", "\[IndentingNewLine]", 
     RowBox[{"hbscl", "=", 
      RowBox[{"{", 
       RowBox[{
       "0.0", ",", " ", "0.0", ",", " ", "1.0", ",", " ", "0.0", ",", " ", 
        "4.0", ",", " ", "0.0", ",", " ", "1.0", ",", " ", "0.0", ",", " ", 
        "4.0", ",", " ", "4.0", ",", " ", "1.0", ",", " ", "1.0", ",", " ", 
        "1.0", ",", " ", "1.0", ",", " ", "1.0", ",", " ", "1.0", ",", " ", 
        "1.0"}], "}"}]}], ";"}], "*)"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"hbscl", "=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"0.0", ",", " ", "0.0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
        "1.37", ",", " ", "0.0", ",", " ", "3.89", ",", " ", "1.30", ",", " ",
          "1.32", ",", " ", "1.22", ",", " ", "0.0"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
        "1.36", ",", " ", "0.0", ",", " ", "3.50", ",", " ", "1.30", ",", " ",
          "1.32", ",", " ", "1.22", ",", " ", "3.47", ",", " ", "0.33", ",", 
         " ", "1.01"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{
        "1.17", ",", " ", "0.0", ",", " ", "3.52", ",", " ", "1.30", ",", " ",
          "1.32", ",", " ", "1.22", ",", " ", "3.62", ",", " ", "0.33", ",", 
         " ", "1.01"}], "}"}]}], "}"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{
    "THE", " ", "FILES", " ", "READ", " ", "HERE", " ", "SHOULD", " ", "BE", 
     " ", "IDENTICAL", " ", "TO", " ", "THOSE", " ", "USED", " ", "IN", " ", 
     "THE", " ", "LAMMPS", " ", "SIMULATIONS"}], " ", "*)"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"antiHB", "=", 
     RowBox[{"Module", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"A", ",", "A1", ",", "A2"}], "}"}], ",", 
       RowBox[{
        RowBox[{"A", "=", 
         RowBox[{"StringSplit", "[", 
          RowBox[{"ReadList", "[", 
           RowBox[{"\"\<anti_HB\>\"", ",", "String"}], "]"}], "]"}]}], ";", 
        RowBox[{"A1", "=", 
         RowBox[{"A", "[", 
          RowBox[{"[", 
           RowBox[{";;", "20"}], "]"}], "]"}]}], ";", 
        RowBox[{"A2", "=", 
         RowBox[{"A", "[", 
          RowBox[{"[", 
           RowBox[{"22", ";;"}], "]"}], "]"}]}], ";", 
        RowBox[{"ToExpression", "[", 
         RowBox[{"{", 
          RowBox[{"A1", ",", "A2"}], "}"}], "]"}]}]}], "]"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"antiNHB", "=", 
     RowBox[{"Module", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"A", ",", "A1", ",", "A2"}], "}"}], ",", 
       RowBox[{
        RowBox[{"A", "=", 
         RowBox[{"StringSplit", "[", 
          RowBox[{"ReadList", "[", 
           RowBox[{"\"\<anti_NHB\>\"", ",", "String"}], "]"}], "]"}]}], ";", 
        RowBox[{"A1", "=", 
         RowBox[{"A", "[", 
          RowBox[{"[", 
           RowBox[{";;", "20"}], "]"}], "]"}]}], ";", 
        RowBox[{"A2", "=", 
         RowBox[{"A", "[", 
          RowBox[{"[", 
           RowBox[{"22", ";;"}], "]"}], "]"}]}], ";", 
        RowBox[{"ToExpression", "[", 
         RowBox[{"{", 
          RowBox[{"A1", ",", "A2"}], "}"}], "]"}]}]}], "]"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"paraHB", "=", 
     RowBox[{"Module", "[", 
      RowBox[{
       RowBox[{"{", 
        RowBox[{"A", ",", "A1", ",", "A2"}], "}"}], ",", 
       RowBox[{
        RowBox[{"A", "=", 
         RowBox[{"StringSplit", "[", 
          RowBox[{"ReadList", "[", 
           RowBox[{"\"\<para_HB\>\"", ",", "String"}], "]"}], "]"}]}], ";", 
        RowBox[{"A1", "=", 
         RowBox[{"A", "[", 
          RowBox[{"[", 
           RowBox[{";;", "20"}], "]"}], "]"}]}], ";", 
        RowBox[{"A2", "=", 
         RowBox[{"A", "[", 
          RowBox[{"[", 
           RowBox[{"22", ";;"}], "]"}], "]"}]}], ";", 
        RowBox[{"ToExpression", "[", 
         RowBox[{"{", 
          RowBox[{"A1", ",", "A2"}], "}"}], "]"}]}]}], "]"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"paraOne", "=", 
     RowBox[{"ReadList", "[", 
      RowBox[{"\"\<para_one\>\"", ",", "Real"}], "]"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"antiOne", "=", 
     RowBox[{"ReadList", "[", 
      RowBox[{"\"\<anti_one\>\"", ",", "Real"}], "]"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{
     RowBox[{
     "WHEN", " ", "TESTING", " ", "THE", " ", "SECONDARY", " ", "STRUCTURE", 
      " ", "BIAS"}], ",", " ", 
     RowBox[{
     "IT", " ", "IS", " ", "NECESSARY", " ", "TO", " ", "INPUT", " ", "THE", 
      " ", "ssweight", " ", "FILE", " ", "HERE"}]}], " ", "*)"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"ssweight5", "=", 
     RowBox[{"Table", "[", 
      RowBox[{"0.0", ",", 
       RowBox[{"{", 
        RowBox[{"i", ",", "1", ",", "M"}], "}"}]}], "]"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"ssweight6", "=", 
     RowBox[{"Table", "[", 
      RowBox[{"0.0", ",", 
       RowBox[{"{", 
        RowBox[{"i", ",", "1", ",", "M"}], "}"}]}], "]"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"ssweight5", "=", 
     RowBox[{"{", 
      RowBox[{
      "0.0", ",", "0.0", ",", "0.0", ",", "0.0", ",", "0.0", ",", "0.0", ",", 
       "0.0", ",", "1.0", ",", "1.0", ",", "1.0", ",", "0.0", ",", "0.0", ",",
        "0.0", ",", "0.0", ",", "1.0", ",", "1.0", ",", "1.0", ",", "1.0", 
       ",", "1.0", ",", "1.0", ",", "1.0", ",", "0.0", ",", "0.0", ",", "0.0",
        ",", "0.0", ",", "0.0"}], "}"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"ssweight6", "=", 
     RowBox[{"{", 
      RowBox[{
      "0.0", ",", "0.0", ",", "1.0", ",", "0.0", ",", "0.0", ",", "1.0", ",", 
       "0.0", ",", "0.0", ",", "1.0", ",", "0.0", ",", "0.0", ",", "1.0", ",",
        "0.0", ",", "0.0", ",", "1.0", ",", "0.0", ",", "0.0", ",", "1.0", 
       ",", "0.0", ",", "0.0", ",", "1.0", ",", "0.0", ",", "0.0", ",", "1.0",
        ",", "0.0", ",", "0.0"}], "}"}]}], ";"}]}]}]], "Input",
 CellChangeTimes->{{3.468855041674871*^9, 3.468855079950027*^9}, {
   3.468855666605926*^9, 3.468855673900051*^9}, 3.468855709289034*^9, {
   3.468855745638946*^9, 3.468855829557212*^9}, {3.468855888638706*^9, 
   3.468855943278901*^9}, {3.4688576810114937`*^9, 3.468857684633127*^9}, {
   3.46885855303533*^9, 3.468858580761301*^9}, {3.468859141012412*^9, 
   3.468859145274663*^9}, {3.468859214252527*^9, 3.4688592331711807`*^9}, {
   3.468859979725029*^9, 3.46885999866171*^9}, {3.468860818095912*^9, 
   3.4688608360157146`*^9}, {3.468860933646689*^9, 3.468860942091736*^9}, {
   3.468861504985867*^9, 3.468861524727579*^9}, {3.468861960119812*^9, 
   3.468861960417213*^9}, {3.468935785699058*^9, 3.4689357869081783`*^9}, {
   3.4689361497449617`*^9, 3.468936212542829*^9}, {3.468936354506432*^9, 
   3.468936355251848*^9}, {3.46894268019945*^9, 3.4689426846410217`*^9}, 
   3.4704215988137207`*^9, {3.470422467249007*^9, 3.470422497914154*^9}, {
   3.470422776821528*^9, 3.470422801419194*^9}, {3.47042441676908*^9, 
   3.470424484888514*^9}, {3.470503754616066*^9, 3.470503755284004*^9}, {
   3.470507820837861*^9, 3.470507854574795*^9}, {3.4706652122045116`*^9, 
   3.470665213975938*^9}, {3.470704290766*^9, 3.470704300461*^9}, {
   3.471097821352539*^9, 3.471097834694333*^9}, {3.4729055995300703`*^9, 
   3.472905673556264*^9}, {3.472921774872183*^9, 3.472921775391543*^9}, {
   3.511719292462266*^9, 3.511719361249833*^9}, {3.5117277678340693`*^9, 
   3.511727800103484*^9}, 3.511727869234653*^9, {3.511727923766663*^9, 
   3.51172792509929*^9}, {3.5117280512682447`*^9, 3.511728053650504*^9}, {
   3.51173014931666*^9, 3.5117301502023907`*^9}, {3.511730197731691*^9, 
   3.511730199507243*^9}, {3.5123201646345787`*^9, 3.512320285482141*^9}, {
   3.512320320954728*^9, 3.5123204634327106`*^9}, {3.512320515247918*^9, 
   3.512320536041355*^9}, {3.512320581393486*^9, 3.5123207300306892`*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
    RowBox[{"THIS", " ", "BLOCK", " ", "\"\<SETS THE RULES\>\""}], ",", " ", 
    RowBox[{
    "WHICH", " ", "MEANS", " ", "IT", " ", "TELLS", " ", "MATHEMATICA", " ", 
     "HOW", " ", "TO", " ", "EVALUATE", " ", "ALL", " ", "THE", " ", 
     "POTENTIALS", " ", "GIVEN", " ", "THE", " ", "INPUT", " ", "SPECIFIED", 
     " ", "ABOVE"}]}], " ", "*)"}], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    SubscriptBox["r", 
     RowBox[{"u_", ",", "i_"}]], ":=", 
    RowBox[{"{", 
     RowBox[{
      SubscriptBox["x", 
       RowBox[{"u", ",", "i"}]], ",", 
      SubscriptBox["y", 
       RowBox[{"u", ",", "i"}]], ",", 
      SubscriptBox["z", 
       RowBox[{"u", ",", "i"}]]}], "}"}]}], "\[IndentingNewLine]", 
   RowBox[{
    SubscriptBox["r", 
     RowBox[{"N", ",", "i_"}]], ":=", 
    RowBox[{
     RowBox[{"an", " ", 
      SubscriptBox["r", 
       RowBox[{"Ca", ",", 
        RowBox[{"i", "-", "1"}]}]]}], "+", 
     RowBox[{"bn", " ", 
      SubscriptBox["r", 
       RowBox[{"Ca", ",", "i"}]]}], "+", 
     RowBox[{"cn", " ", 
      SubscriptBox["r", 
       RowBox[{"O", ",", 
        RowBox[{"i", "-", "1"}]}]]}]}]}], "\[IndentingNewLine]", 
   RowBox[{
    SubscriptBox["r", 
     RowBox[{"H", ",", "i_"}]], ":=", 
    RowBox[{
     RowBox[{"ah", " ", 
      SubscriptBox["r", 
       RowBox[{"Ca", ",", 
        RowBox[{"i", "-", "1"}]}]]}], "+", 
     RowBox[{"bh", " ", 
      SubscriptBox["r", 
       RowBox[{"Ca", ",", "i"}]]}], "+", 
     RowBox[{"ch", " ", 
      SubscriptBox["r", 
       RowBox[{"O", ",", 
        RowBox[{"i", "-", "1"}]}]]}]}]}], "\[IndentingNewLine]", 
   RowBox[{
    SubscriptBox["r", 
     RowBox[{"Cp", ",", "i_"}]], ":=", 
    RowBox[{
     RowBox[{"ap", " ", 
      SubscriptBox["r", 
       RowBox[{"Ca", ",", "i"}]]}], "+", 
     RowBox[{"bp", " ", 
      SubscriptBox["r", 
       RowBox[{"Ca", ",", 
        RowBox[{"i", "+", "1"}]}]]}], "+", 
     RowBox[{"cp", " ", 
      SubscriptBox["r", 
       RowBox[{"O", ",", "i"}]]}]}]}], "\[IndentingNewLine]", 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"MyNorm", "[", "a_", "]"}], ":=", 
    SqrtBox[
     RowBox[{"Total", "[", 
      RowBox[{"Table", "[", 
       RowBox[{
        SuperscriptBox["ai", "2"], ",", 
        RowBox[{"{", 
         RowBox[{"ai", ",", "a"}], "}"}]}], "]"}], "]"}]]}], 
   "\[IndentingNewLine]", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"rule", "=", 
     RowBox[{"Flatten", "[", 
      RowBox[{"Append", "[", 
       RowBox[{
        RowBox[{"Append", "[", 
         RowBox[{
          RowBox[{"Table", "[", 
           RowBox[{
            RowBox[{"{", 
             RowBox[{
              RowBox[{
               SubscriptBox["r", 
                RowBox[{"Ca", ",", "i"}]], "[", 
               RowBox[{"[", "j", "]"}], "]"}], "\[Rule]", 
              RowBox[{"rca", "[", 
               RowBox[{"[", 
                RowBox[{"i", ",", "j"}], "]"}], "]"}]}], "}"}], ",", 
            RowBox[{"{", 
             RowBox[{"i", ",", "1", ",", "M"}], "}"}], ",", 
            RowBox[{"{", 
             RowBox[{"j", ",", "1", ",", "3"}], "}"}]}], "]"}], ",", 
          RowBox[{"Table", "[", 
           RowBox[{
            RowBox[{"{", 
             RowBox[{
              RowBox[{
               SubscriptBox["r", 
                RowBox[{"Cb", ",", "i"}]], "[", 
               RowBox[{"[", "j", "]"}], "]"}], "\[Rule]", 
              RowBox[{"rcb", "[", 
               RowBox[{"[", 
                RowBox[{"i", ",", "j"}], "]"}], "]"}]}], "}"}], ",", 
            RowBox[{"{", 
             RowBox[{"i", ",", "1", ",", "M"}], "}"}], ",", 
            RowBox[{"{", 
             RowBox[{"j", ",", "1", ",", "3"}], "}"}]}], "]"}]}], "]"}], ",", 
        
        RowBox[{"Table", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{
            RowBox[{
             SubscriptBox["r", 
              RowBox[{"O", ",", "i"}]], "[", 
             RowBox[{"[", "j", "]"}], "]"}], "\[Rule]", 
            RowBox[{"ro", "[", 
             RowBox[{"[", 
              RowBox[{"i", ",", "j"}], "]"}], "]"}]}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"i", ",", "1", ",", "M"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"j", ",", "1", ",", "3"}], "}"}]}], "]"}]}], "]"}], 
      "]"}]}], ";"}], "\[IndentingNewLine]", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"rule", "=", 
     RowBox[{"Flatten", "[", 
      RowBox[{"Append", "[", 
       RowBox[{
        RowBox[{"Append", "[", 
         RowBox[{
          RowBox[{"Table", "[", 
           RowBox[{
            RowBox[{"{", 
             RowBox[{
              RowBox[{
               SubscriptBox["r", 
                RowBox[{"Ca", ",", "i"}]], "[", 
               RowBox[{"[", "j", "]"}], "]"}], "\[Rule]", 
              RowBox[{"rca", "[", 
               RowBox[{"[", 
                RowBox[{"i", ",", "j"}], "]"}], "]"}]}], "}"}], ",", 
            RowBox[{"{", 
             RowBox[{"i", ",", "1", ",", "M"}], "}"}], ",", 
            RowBox[{"{", 
             RowBox[{"j", ",", "1", ",", "3"}], "}"}]}], "]"}], ",", 
          RowBox[{"Table", "[", 
           RowBox[{
            RowBox[{"{", 
             RowBox[{
              RowBox[{
               SubscriptBox["r", 
                RowBox[{"Cb", ",", "i"}]], "[", 
               RowBox[{"[", "j", "]"}], "]"}], "\[Rule]", 
              RowBox[{"rcb", "[", 
               RowBox[{"[", 
                RowBox[{"i", ",", "j"}], "]"}], "]"}]}], "}"}], ",", 
            RowBox[{"{", 
             RowBox[{"i", ",", "1", ",", "M"}], "}"}], ",", 
            RowBox[{"{", 
             RowBox[{"j", ",", "1", ",", "3"}], "}"}]}], "]"}]}], "]"}], ",", 
        
        RowBox[{"Table", "[", 
         RowBox[{
          RowBox[{"{", 
           RowBox[{
            RowBox[{
             SubscriptBox["r", 
              RowBox[{"O", ",", "i"}]], "[", 
             RowBox[{"[", "j", "]"}], "]"}], "\[Rule]", 
            RowBox[{"ro", "[", 
             RowBox[{"[", 
              RowBox[{"i", ",", "j"}], "]"}], "]"}]}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"i", ",", "1", ",", "M"}], "}"}], ",", 
          RowBox[{"{", 
           RowBox[{"j", ",", "1", ",", "3"}], "}"}]}], "]"}]}], "]"}], 
      "]"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"rulee", "[", 
     RowBox[{"a_", ",", "i_"}], "]"}], ":=", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{"rule", "[", 
       RowBox[{"[", 
        RowBox[{
         RowBox[{"3", "M", 
          RowBox[{"(", 
           RowBox[{"a", "-", "1"}], ")"}]}], "+", 
         RowBox[{"3", 
          RowBox[{"(", 
           RowBox[{"i", "-", "1"}], ")"}]}], "+", "1"}], "]"}], "]"}], ",", 
      RowBox[{"rule", "[", 
       RowBox[{"[", 
        RowBox[{
         RowBox[{"3", "M", 
          RowBox[{"(", 
           RowBox[{"a", "-", "1"}], ")"}]}], "+", 
         RowBox[{"3", 
          RowBox[{"(", 
           RowBox[{"i", "-", "1"}], ")"}]}], "+", "2"}], "]"}], "]"}], ",", 
      RowBox[{"rule", "[", 
       RowBox[{"[", 
        RowBox[{
         RowBox[{"3", "M", 
          RowBox[{"(", 
           RowBox[{"a", "-", "1"}], ")"}]}], "+", 
         RowBox[{"3", 
          RowBox[{"(", 
           RowBox[{"i", "-", "1"}], ")"}]}], "+", "3"}], "]"}], "]"}]}], 
     "}"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"rules", "[", 
     RowBox[{"a_", ",", "i_"}], "]"}], ":=", 
    RowBox[{"Delete", "[", 
     RowBox[{"rule", ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{
          RowBox[{"3", "M", 
           RowBox[{"(", 
            RowBox[{"a", "-", "1"}], ")"}]}], "+", 
          RowBox[{"3", 
           RowBox[{"(", 
            RowBox[{"i", "-", "1"}], ")"}]}], "+", "1"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"3", "M", 
           RowBox[{"(", 
            RowBox[{"a", "-", "1"}], ")"}]}], "+", 
          RowBox[{"3", 
           RowBox[{"(", 
            RowBox[{"i", "-", "1"}], ")"}]}], "+", "2"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"3", "M", 
           RowBox[{"(", 
            RowBox[{"a", "-", "1"}], ")"}]}], "+", 
          RowBox[{"3", 
           RowBox[{"(", 
            RowBox[{"i", "-", "1"}], ")"}]}], "+", "3"}], "}"}]}], "}"}]}], 
     "]"}]}], "\[IndentingNewLine]", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"Ruless", "[", "a_", "]"}], ":=", 
     RowBox[{"Flatten", "[", 
      RowBox[{"Table", "[", 
       RowBox[{
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{"a", "\[Equal]", "1"}], ",", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{
             SubscriptBox["r", 
              RowBox[{"Ca", ",", "i"}]], "[", 
             RowBox[{"[", "j", "]"}], "]"}], "\[Rule]", 
            RowBox[{"rca", "[", 
             RowBox[{"[", 
              RowBox[{"i", ",", "j"}], "]"}], "]"}]}], "}"}], ",", 
          RowBox[{"If", "[", 
           RowBox[{
            RowBox[{"a", "\[Equal]", "2"}], ",", 
            RowBox[{"{", 
             RowBox[{
              RowBox[{
               SubscriptBox["r", 
                RowBox[{"Cb", ",", "i"}]], "[", 
               RowBox[{"[", "j", "]"}], "]"}], "\[Rule]", 
              RowBox[{"rcb", "[", 
               RowBox[{"[", 
                RowBox[{"i", ",", "j"}], "]"}], "]"}]}], "}"}], ",", 
            RowBox[{"{", 
             RowBox[{
              RowBox[{
               SubscriptBox["r", 
                RowBox[{"O", ",", "i"}]], "[", 
               RowBox[{"[", "j", "]"}], "]"}], "\[Rule]", 
              RowBox[{"ro", "[", 
               RowBox[{"[", 
                RowBox[{"i", ",", "j"}], "]"}], "]"}]}], "}"}]}], "]"}]}], 
         "]"}], ",", 
        RowBox[{"{", 
         RowBox[{"i", ",", "1", ",", "M"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{"j", ",", "1", ",", "3"}], "}"}]}], "]"}], "]"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"Rulee", "[", 
     RowBox[{"a_", ",", "i_"}], "]"}], ":=", 
    RowBox[{"{", 
     RowBox[{
      RowBox[{
       RowBox[{"Ruless", "[", "a", "]"}], "[", 
       RowBox[{"[", 
        RowBox[{
         RowBox[{"3", 
          RowBox[{"(", 
           RowBox[{"i", "-", "1"}], ")"}]}], "+", "1"}], "]"}], "]"}], ",", 
      RowBox[{
       RowBox[{"Ruless", "[", "a", "]"}], "[", 
       RowBox[{"[", 
        RowBox[{
         RowBox[{"3", 
          RowBox[{"(", 
           RowBox[{"i", "-", "1"}], ")"}]}], "+", "2"}], "]"}], "]"}], ",", 
      RowBox[{
       RowBox[{"Ruless", "[", "a", "]"}], "[", 
       RowBox[{"[", 
        RowBox[{
         RowBox[{"3", 
          RowBox[{"(", 
           RowBox[{"i", "-", "1"}], ")"}]}], "+", "3"}], "]"}], "]"}]}], 
     "}"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"Rules", "[", 
     RowBox[{"a_", ",", "i_"}], "]"}], ":=", 
    RowBox[{"Delete", "[", 
     RowBox[{
      RowBox[{"Ruless", "[", "a", "]"}], ",", 
      RowBox[{"{", 
       RowBox[{
        RowBox[{"{", 
         RowBox[{
          RowBox[{"3", 
           RowBox[{"(", 
            RowBox[{"i", "-", "1"}], ")"}]}], "+", "1"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"3", 
           RowBox[{"(", 
            RowBox[{"i", "-", "1"}], ")"}]}], "+", "2"}], "}"}], ",", 
        RowBox[{"{", 
         RowBox[{
          RowBox[{"3", 
           RowBox[{"(", 
            RowBox[{"i", "-", "1"}], ")"}]}], "+", "3"}], "}"}]}], "}"}]}], 
     "]"}]}]}]}]], "Input",
 CellChangeTimes->{{3.468854673666232*^9, 3.468854684512622*^9}, {
  3.468854902951829*^9, 3.468854932950259*^9}, {3.46885839650323*^9, 
  3.468858396977489*^9}, {3.468860186729788*^9, 3.4688601872483377`*^9}, {
  3.472411610493401*^9, 3.472411621584505*^9}, {3.512321382434227*^9, 
  3.5123214118001003`*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
   "THIS", " ", "BLOCK", " ", "DEFINES", " ", "THE", " ", "DSSP", " ", 
    "POTENTIAL"}], " ", "*)"}], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{"iRepulsive", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"If", " ", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"j", "\[Equal]", "M"}], "  ", "||", " ", 
       RowBox[{
        RowBox[{"se", "[", 
         RowBox[{"[", 
          RowBox[{"j", "+", "1"}], "]"}], "]"}], "\[Equal]", "15"}]}], ",", 
      " ", "False", ",", " ", "True"}], "]"}]}], "\n", 
   RowBox[{
    RowBox[{"iAP", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"i", "\[Equal]", "1"}], " ", "||", " ", 
       RowBox[{"j", "\[Equal]", "M"}], " ", "||", " ", 
       RowBox[{
        RowBox[{"se", "[", 
         RowBox[{"[", "i", "]"}], "]"}], "\[Equal]", "15"}]}], ",", "False", 
      ",", " ", "True"}], "]"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"iP", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"i", "\[Equal]", 
        RowBox[{"M", "-", "1"}]}], " ", "||", " ", 
       RowBox[{"j", "\[Equal]", "M"}], " ", "||", " ", 
       RowBox[{
        RowBox[{"se", "[", 
         RowBox[{"[", 
          RowBox[{"i", "+", "2"}], "]"}], "]"}], "\[Equal]", "15"}]}], ",", 
      "False", ",", " ", "True"}], "]"}]}], "\[IndentingNewLine]", 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"thetaSeqAntiHB", "[", 
     RowBox[{"i_", ",", "j_", ",", "k_"}], "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"iAP", "[", 
        RowBox[{"i", ",", "j"}], "]"}], "\[Equal]", "True"}], ",", 
      RowBox[{"0.5", "*", 
       RowBox[{"antiHB", "[", 
        RowBox[{"[", 
         RowBox[{"k", ",", 
          RowBox[{"se", "[", 
           RowBox[{"[", "i", "]"}], "]"}], ",", 
          RowBox[{"se", "[", 
           RowBox[{"[", "j", "]"}], "]"}]}], "]"}], "]"}]}], ",", "0.0"}], 
     "]"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"thetaSeqAntiNHB", "[", 
     RowBox[{"i_", ",", "j_", ",", "k_"}], "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"iAP", "[", 
        RowBox[{"i", ",", "j"}], "]"}], "\[Equal]", "True"}], ",", 
      RowBox[{"0.25", "*", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"antiNHB", "[", 
          RowBox[{"[", 
           RowBox[{"k", ",", 
            RowBox[{"se", "[", 
             RowBox[{"[", 
              RowBox[{"i", "+", "1"}], "]"}], "]"}], ",", 
            RowBox[{"se", "[", 
             RowBox[{"[", 
              RowBox[{"j", "-", "1"}], "]"}], "]"}]}], "]"}], "]"}], "+", 
         "\[IndentingNewLine]", 
         RowBox[{"antiNHB", "[", 
          RowBox[{"[", 
           RowBox[{"k", ",", 
            RowBox[{"se", "[", 
             RowBox[{"[", 
              RowBox[{"i", "-", "1"}], "]"}], "]"}], ",", 
            RowBox[{"se", "[", 
             RowBox[{"[", 
              RowBox[{"j", "+", "1"}], "]"}], "]"}]}], "]"}], "]"}]}], 
        ")"}]}], ",", "0.0"}], "]"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"thetaSeqParaHB", "[", 
     RowBox[{"i_", ",", "j_", ",", "k_"}], "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{"iP", "[", 
       RowBox[{"i", ",", "j"}], "]"}], ",", 
      RowBox[{"paraHB", "[", 
       RowBox[{"[", 
        RowBox[{"k", ",", 
         RowBox[{"se", "[", 
          RowBox[{"[", 
           RowBox[{"i", "+", "1"}], "]"}], "]"}], ",", 
         RowBox[{"se", "[", 
          RowBox[{"[", "j", "]"}], "]"}]}], "]"}], "]"}], ",", "0.0"}], 
     "]"}]}], "\[IndentingNewLine]", "\[IndentingNewLine]", 
   RowBox[{"(*", 
    RowBox[{
     RowBox[{"\[Lambda]1", "[", 
      RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
     RowBox[{
      RowBox[{
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"Abs", "[", 
           RowBox[{"j", "-", "i"}], "]"}], "<", "5"}], ",", 
         RowBox[{"-", 
          RowBox[{"hbscl", "[", 
           RowBox[{"[", "1", "]"}], "]"}]}], ",", 
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{
            RowBox[{"Abs", "[", 
             RowBox[{"j", "-", "i"}], "]"}], "<", "20"}], ",", 
           RowBox[{"-", 
            RowBox[{"hbscl", "[", 
             RowBox[{"[", "3", "]"}], "]"}]}], ",", 
           RowBox[{"-", 
            RowBox[{"hbscl", "[", 
             RowBox[{"[", "7", "]"}], "]"}]}]}], "]"}]}], "]"}], 
       "\[IndentingNewLine]", 
       RowBox[{"\[Lambda]2", "[", 
        RowBox[{"i_", ",", "j_"}], "]"}]}], ":=", 
      RowBox[{
       RowBox[{
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"Abs", "[", 
            RowBox[{"j", "-", "i"}], "]"}], "<", "5"}], ",", 
          RowBox[{"-", 
           RowBox[{"hbscl", "[", 
            RowBox[{"[", "2", "]"}], "]"}]}], ",", 
          RowBox[{"If", "[", 
           RowBox[{
            RowBox[{
             RowBox[{"Abs", "[", 
              RowBox[{"j", "-", "i"}], "]"}], "<", "20"}], ",", 
            RowBox[{"-", 
             RowBox[{"hbscl", "[", 
              RowBox[{"[", "4", "]"}], "]"}]}], ",", 
            RowBox[{"-", 
             RowBox[{"hbscl", "[", 
              RowBox[{"[", "8", "]"}], "]"}]}]}], "]"}]}], "]"}], 
        "\[IndentingNewLine]", 
        RowBox[{"\[Lambda]3", "[", 
         RowBox[{"i_", ",", "j_"}], "]"}]}], ":=", 
       RowBox[{
        RowBox[{
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{
            RowBox[{"Abs", "[", 
             RowBox[{"j", "-", "i"}], "]"}], "<", "5"}], ",", "0.0", ",", 
           RowBox[{"If", "[", 
            RowBox[{
             RowBox[{
              RowBox[{"Abs", "[", 
               RowBox[{"j", "-", "i"}], "]"}], "<", "20"}], ",", 
             RowBox[{
              RowBox[{"-", 
               RowBox[{"hbscl", "[", 
                RowBox[{"[", "5", "]"}], "]"}]}], "-", 
              RowBox[{
               RowBox[{"hbscl", "[", 
                RowBox[{"[", "11", "]"}], "]"}], 
               RowBox[{"thetaSeqAntiHB", "[", 
                RowBox[{"i", ",", "j", ",", "1"}], "]"}]}], "-", 
              RowBox[{
               RowBox[{"hbscl", "[", 
                RowBox[{"[", "12", "]"}], "]"}], 
               RowBox[{"thetaSeqAntiNHB", "[", 
                RowBox[{"i", ",", "j", ",", "1"}], "]"}]}], "-", 
              RowBox[{
               RowBox[{"hbscl", "[", 
                RowBox[{"[", "16", "]"}], "]"}], 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"antiOne", "[", 
                  RowBox[{"[", 
                   RowBox[{"se", "[", 
                    RowBox[{"[", "i", "]"}], "]"}], "]"}], "]"}], "+", 
                 RowBox[{"antiOne", "[", 
                  RowBox[{"[", 
                   RowBox[{"se", "[", 
                    RowBox[{"[", "j", "]"}], "]"}], "]"}], "]"}]}], ")"}]}]}],
              ",", 
             RowBox[{
              RowBox[{"-", 
               RowBox[{"hbscl", "[", 
                RowBox[{"[", "9", "]"}], "]"}]}], "-", 
              RowBox[{
               RowBox[{"hbscl", "[", 
                RowBox[{"[", "13", "]"}], "]"}], 
               RowBox[{"thetaSeqAntiHB", "[", 
                RowBox[{"i", ",", "j", ",", "2"}], "]"}]}], "-", 
              RowBox[{
               RowBox[{"hbscl", "[", 
                RowBox[{"[", "14", "]"}], "]"}], 
               RowBox[{"thetaSeqAntiNHB", "[", 
                RowBox[{"i", ",", "j", ",", "2"}], "]"}]}], "-", 
              RowBox[{
               RowBox[{"hbscl", "[", 
                RowBox[{"[", "16", "]"}], "]"}], 
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"antiOne", "[", 
                  RowBox[{"[", 
                   RowBox[{"se", "[", 
                    RowBox[{"[", "i", "]"}], "]"}], "]"}], "]"}], "+", 
                 RowBox[{"antiOne", "[", 
                  RowBox[{"[", 
                   RowBox[{"se", "[", 
                    RowBox[{"[", "j", "]"}], "]"}], "]"}], "]"}]}], 
                ")"}]}]}]}], "]"}]}], "]"}], "\[IndentingNewLine]", 
         RowBox[{"\[Lambda]4", "[", 
          RowBox[{"i_", ",", "j_"}], "]"}]}], ":=", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"Abs", "[", 
            RowBox[{"j", "-", "i"}], "]"}], "<", "5"}], ",", "0.0", ",", 
          RowBox[{"If", "[", 
           RowBox[{
            RowBox[{
             RowBox[{"Abs", "[", 
              RowBox[{"j", "-", "i"}], "]"}], "<", "20"}], ",", 
            RowBox[{"-", 
             RowBox[{"hbscl", "[", 
              RowBox[{"[", "6", "]"}], "]"}]}], ",", 
            RowBox[{
             RowBox[{"-", 
              RowBox[{"hbscl", "[", 
               RowBox[{"[", "10", "]"}], "]"}]}], "-", 
             RowBox[{
              RowBox[{"hbscl", "[", 
               RowBox[{"[", "15", "]"}], "]"}], 
              RowBox[{"thetaSeqParaHB", "[", 
               RowBox[{"i", ",", "j", ",", "2"}], "]"}]}], "-", 
             RowBox[{
              RowBox[{"hbscl", "[", 
               RowBox[{"[", "17", "]"}], "]"}], 
              RowBox[{"(", 
               RowBox[{
                RowBox[{"paraOne", "[", 
                 RowBox[{"[", 
                  RowBox[{"se", "[", 
                   RowBox[{"[", 
                    RowBox[{"i", "+", "1"}], "]"}], "]"}], "]"}], "]"}], "+", 
                
                RowBox[{"paraOne", "[", 
                 RowBox[{"[", 
                  RowBox[{"se", "[", 
                   RowBox[{"[", "j", "]"}], "]"}], "]"}], "]"}]}], 
               ")"}]}]}]}], "]"}]}], "]"}]}]}]}]}], "*)"}], 
   "\[IndentingNewLine]", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"\[Lambda]1", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"Abs", "[", 
        RowBox[{"j", "-", "i"}], "]"}], "<", "4"}], ",", 
      RowBox[{"-", 
       RowBox[{"hbscl", "[", 
        RowBox[{"[", 
         RowBox[{"1", ",", "1"}], "]"}], "]"}]}], ",", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"Abs", "[", 
          RowBox[{"j", "-", "i"}], "]"}], "<", "18"}], ",", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{
           RowBox[{
            RowBox[{"ssweight6", "[", 
             RowBox[{"[", "i", "]"}], "]"}], "\[Equal]", "1.0"}], " ", "&&", 
           " ", 
           RowBox[{
            RowBox[{"ssweight6", "[", 
             RowBox[{"[", "j", "]"}], "]"}], "\[Equal]", "1.0"}]}], ",", 
          RowBox[{"-", 
           RowBox[{"hbscl", "[", 
            RowBox[{"[", 
             RowBox[{"2", ",", "1"}], "]"}], "]"}]}], ",", "0.0"}], "]"}], 
        ",", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"Abs", "[", 
            RowBox[{"j", "-", "i"}], "]"}], "<", "45"}], ",", 
          RowBox[{"-", 
           RowBox[{"hbscl", "[", 
            RowBox[{"[", 
             RowBox[{"3", ",", "1"}], "]"}], "]"}]}], ",", 
          RowBox[{"-", 
           RowBox[{"hbscl", "[", 
            RowBox[{"[", 
             RowBox[{"4", ",", "1"}], "]"}], "]"}]}]}], "]"}]}], "]"}]}], 
     "]"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"\[Lambda]2", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"Abs", "[", 
        RowBox[{"j", "-", "i"}], "]"}], "<", "4"}], ",", 
      RowBox[{"-", 
       RowBox[{"hbscl", "[", 
        RowBox[{"[", 
         RowBox[{"1", ",", "2"}], "]"}], "]"}]}], ",", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"Abs", "[", 
          RowBox[{"j", "-", "i"}], "]"}], "<", "18"}], ",", 
        RowBox[{"-", 
         RowBox[{"hbscl", "[", 
          RowBox[{"[", 
           RowBox[{"2", ",", "2"}], "]"}], "]"}]}], ",", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"Abs", "[", 
            RowBox[{"j", "-", "i"}], "]"}], "<", "45"}], ",", 
          RowBox[{"-", 
           RowBox[{"hbscl", "[", 
            RowBox[{"[", 
             RowBox[{"3", ",", "2"}], "]"}], "]"}]}], ",", 
          RowBox[{"-", 
           RowBox[{"hbscl", "[", 
            RowBox[{"[", 
             RowBox[{"4", ",", "2"}], "]"}], "]"}]}]}], "]"}]}], "]"}]}], 
     "]"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"\[Lambda]3", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"Abs", "[", 
        RowBox[{"j", "-", "i"}], "]"}], "<", "4"}], ",", "0.0", ",", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"Abs", "[", 
          RowBox[{"j", "-", "i"}], "]"}], "<", "18"}], ",", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{
           RowBox[{
            RowBox[{"ssweight6", "[", 
             RowBox[{"[", "i", "]"}], "]"}], "\[Equal]", "1.0"}], " ", "&&", 
           " ", 
           RowBox[{
            RowBox[{"ssweight6", "[", 
             RowBox[{"[", "j", "]"}], "]"}], "\[Equal]", "1.0"}]}], ",", 
          RowBox[{
           RowBox[{"-", 
            RowBox[{"hbscl", "[", 
             RowBox[{"[", 
              RowBox[{"2", ",", "3"}], "]"}], "]"}]}], "-", 
           RowBox[{
            RowBox[{"hbscl", "[", 
             RowBox[{"[", 
              RowBox[{"2", ",", "4"}], "]"}], "]"}], 
            RowBox[{"thetaSeqAntiHB", "[", 
             RowBox[{"i", ",", "j", ",", "1"}], "]"}]}], "-", 
           RowBox[{
            RowBox[{"hbscl", "[", 
             RowBox[{"[", 
              RowBox[{"2", ",", "5"}], "]"}], "]"}], 
            RowBox[{"thetaSeqAntiNHB", "[", 
             RowBox[{"i", ",", "j", ",", "1"}], "]"}]}], "-", 
           RowBox[{
            RowBox[{"hbscl", "[", 
             RowBox[{"[", 
              RowBox[{"2", ",", "6"}], "]"}], "]"}], 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"antiOne", "[", 
               RowBox[{"[", 
                RowBox[{"se", "[", 
                 RowBox[{"[", "i", "]"}], "]"}], "]"}], "]"}], "+", 
              RowBox[{"antiOne", "[", 
               RowBox[{"[", 
                RowBox[{"se", "[", 
                 RowBox[{"[", "j", "]"}], "]"}], "]"}], "]"}]}], ")"}]}]}], 
          ",", "0.0"}], "]"}], ",", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"Abs", "[", 
            RowBox[{"j", "-", "i"}], "]"}], "<", "45"}], ",", 
          RowBox[{
           RowBox[{"-", 
            RowBox[{"hbscl", "[", 
             RowBox[{"[", 
              RowBox[{"3", ",", "3"}], "]"}], "]"}]}], "-", 
           RowBox[{
            RowBox[{"hbscl", "[", 
             RowBox[{"[", 
              RowBox[{"3", ",", "4"}], "]"}], "]"}], 
            RowBox[{"thetaSeqAntiHB", "[", 
             RowBox[{"i", ",", "j", ",", "2"}], "]"}]}], "-", 
           RowBox[{
            RowBox[{"hbscl", "[", 
             RowBox[{"[", 
              RowBox[{"3", ",", "5"}], "]"}], "]"}], 
            RowBox[{"thetaSeqAntiNHB", "[", 
             RowBox[{"i", ",", "j", ",", "2"}], "]"}]}], "-", 
           RowBox[{
            RowBox[{"hbscl", "[", 
             RowBox[{"[", 
              RowBox[{"3", ",", "6"}], "]"}], "]"}], 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"antiOne", "[", 
               RowBox[{"[", 
                RowBox[{"se", "[", 
                 RowBox[{"[", "i", "]"}], "]"}], "]"}], "]"}], "+", 
              RowBox[{"antiOne", "[", 
               RowBox[{"[", 
                RowBox[{"se", "[", 
                 RowBox[{"[", "j", "]"}], "]"}], "]"}], "]"}]}], ")"}]}]}], 
          ",", 
          RowBox[{
           RowBox[{"-", 
            RowBox[{"hbscl", "[", 
             RowBox[{"[", 
              RowBox[{"4", ",", "3"}], "]"}], "]"}]}], "-", 
           RowBox[{
            RowBox[{"hbscl", "[", 
             RowBox[{"[", 
              RowBox[{"4", ",", "4"}], "]"}], "]"}], 
            RowBox[{"thetaSeqAntiHB", "[", 
             RowBox[{"i", ",", "j", ",", "2"}], "]"}]}], "-", 
           RowBox[{
            RowBox[{"hbscl", "[", 
             RowBox[{"[", 
              RowBox[{"4", ",", "5"}], "]"}], "]"}], 
            RowBox[{"thetaSeqAntiNHB", "[", 
             RowBox[{"i", ",", "j", ",", "2"}], "]"}]}], "-", 
           RowBox[{
            RowBox[{"hbscl", "[", 
             RowBox[{"[", 
              RowBox[{"4", ",", "6"}], "]"}], "]"}], 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"antiOne", "[", 
               RowBox[{"[", 
                RowBox[{"se", "[", 
                 RowBox[{"[", "i", "]"}], "]"}], "]"}], "]"}], "+", 
              RowBox[{"antiOne", "[", 
               RowBox[{"[", 
                RowBox[{"se", "[", 
                 RowBox[{"[", "j", "]"}], "]"}], "]"}], "]"}]}], ")"}]}]}]}], 
         "]"}]}], "]"}]}], "]"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"\[Lambda]4", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"Abs", "[", 
        RowBox[{"j", "-", "i"}], "]"}], "<", "4"}], ",", "0.0", ",", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"Abs", "[", 
          RowBox[{"j", "-", "i"}], "]"}], "<", "18"}], ",", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{
           RowBox[{
            RowBox[{"ssweight6", "[", 
             RowBox[{"[", "i", "]"}], "]"}], "\[Equal]", "1.0"}], " ", "&&", 
           " ", 
           RowBox[{
            RowBox[{"ssweight6", "[", 
             RowBox[{"[", "j", "]"}], "]"}], "\[Equal]", "1.0"}]}], ",", 
          RowBox[{"-", 
           RowBox[{"hbscl", "[", 
            RowBox[{"[", 
             RowBox[{"2", ",", "7"}], "]"}], "]"}]}], ",", "0.0"}], "]"}], 
        ",", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"Abs", "[", 
            RowBox[{"j", "-", "i"}], "]"}], "<", "45"}], ",", 
          RowBox[{
           RowBox[{"-", 
            RowBox[{"hbscl", "[", 
             RowBox[{"[", 
              RowBox[{"3", ",", "7"}], "]"}], "]"}]}], "-", 
           RowBox[{
            RowBox[{"hbscl", "[", 
             RowBox[{"[", 
              RowBox[{"3", ",", "8"}], "]"}], "]"}], 
            RowBox[{"thetaSeqParaHB", "[", 
             RowBox[{"i", ",", "j", ",", "2"}], "]"}]}], "-", 
           RowBox[{
            RowBox[{"hbscl", "[", 
             RowBox[{"[", 
              RowBox[{"3", ",", "9"}], "]"}], "]"}], 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"paraOne", "[", 
               RowBox[{"[", 
                RowBox[{"se", "[", 
                 RowBox[{"[", 
                  RowBox[{"i", "+", "1"}], "]"}], "]"}], "]"}], "]"}], "+", 
              RowBox[{"paraOne", "[", 
               RowBox[{"[", 
                RowBox[{"se", "[", 
                 RowBox[{"[", "j", "]"}], "]"}], "]"}], "]"}]}], ")"}]}]}], 
          ",", 
          RowBox[{
           RowBox[{"-", 
            RowBox[{"hbscl", "[", 
             RowBox[{"[", 
              RowBox[{"4", ",", "7"}], "]"}], "]"}]}], "-", 
           RowBox[{
            RowBox[{"hbscl", "[", 
             RowBox[{"[", 
              RowBox[{"4", ",", "8"}], "]"}], "]"}], 
            RowBox[{"thetaSeqParaHB", "[", 
             RowBox[{"i", ",", "j", ",", "2"}], "]"}]}], "-", 
           RowBox[{
            RowBox[{"hbscl", "[", 
             RowBox[{"[", 
              RowBox[{"4", ",", "9"}], "]"}], "]"}], 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"paraOne", "[", 
               RowBox[{"[", 
                RowBox[{"se", "[", 
                 RowBox[{"[", 
                  RowBox[{"i", "+", "1"}], "]"}], "]"}], "]"}], "]"}], "+", 
              RowBox[{"paraOne", "[", 
               RowBox[{"[", 
                RowBox[{"se", "[", 
                 RowBox[{"[", "j", "]"}], "]"}], "]"}], "]"}]}], ")"}]}]}]}], 
         "]"}]}], "]"}]}], "]"}]}], "\[IndentingNewLine]", 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"rNO", "[", 
     RowBox[{"k_", ",", "i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"Switch", "[", 
     RowBox[{"k", ",", "1", ",", 
      RowBox[{"MyNorm", "[", 
       RowBox[{
        SubscriptBox["r", 
         RowBox[{"O", ",", "i"}]], "-", 
        SubscriptBox["r", 
         RowBox[{"N", ",", "j"}]]}], "]"}], ",", "2", ",", 
      RowBox[{"MyNorm", "[", 
       RowBox[{
        SubscriptBox["r", 
         RowBox[{"O", ",", "i"}]], "-", 
        SubscriptBox["r", 
         RowBox[{"N", ",", 
          RowBox[{"j", "+", "1"}]}]]}], "]"}], ",", "3", ",", 
      RowBox[{"MyNorm", "[", 
       RowBox[{
        SubscriptBox["r", 
         RowBox[{"O", ",", "j"}]], "-", 
        SubscriptBox["r", 
         RowBox[{"N", ",", "i"}]]}], "]"}], ",", "4", ",", 
      RowBox[{"MyNorm", "[", 
       RowBox[{
        SubscriptBox["r", 
         RowBox[{"O", ",", "j"}]], "-", 
        SubscriptBox["r", 
         RowBox[{"N", ",", 
          RowBox[{"i", "+", "2"}]}]]}], "]"}]}], "]"}]}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"rHO", "[", 
     RowBox[{"k_", ",", "i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"Switch", "[", 
     RowBox[{"k", ",", "1", ",", 
      RowBox[{"MyNorm", "[", 
       RowBox[{
        SubscriptBox["r", 
         RowBox[{"O", ",", "i"}]], "-", 
        SubscriptBox["r", 
         RowBox[{"H", ",", "j"}]]}], "]"}], ",", "2", ",", 
      RowBox[{"MyNorm", "[", 
       RowBox[{
        SubscriptBox["r", 
         RowBox[{"O", ",", "i"}]], "-", 
        SubscriptBox["r", 
         RowBox[{"H", ",", 
          RowBox[{"j", "+", "1"}]}]]}], "]"}], ",", "3", ",", 
      RowBox[{"MyNorm", "[", 
       RowBox[{
        SubscriptBox["r", 
         RowBox[{"O", ",", "j"}]], "-", 
        SubscriptBox["r", 
         RowBox[{"H", ",", "i"}]]}], "]"}], ",", "4", ",", 
      RowBox[{"MyNorm", "[", 
       RowBox[{
        SubscriptBox["r", 
         RowBox[{"O", ",", "j"}]], "-", 
        SubscriptBox["r", 
         RowBox[{"H", ",", 
          RowBox[{"i", "+", "2"}]}]]}], "]"}]}], "]"}]}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"\[Delta]", "[", 
     RowBox[{"k_", ",", "i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"k", "\[Equal]", "2"}], " ", "&&", "  ", 
       RowBox[{
        RowBox[{"iRepulsive", "[", 
         RowBox[{"i", ",", "j"}], "]"}], "\[Equal]", "False"}]}], ",", "0", 
      ",", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"k", "\[Equal]", "3"}], " ", "&&", " ", 
         RowBox[{
          RowBox[{"iAP", "[", 
           RowBox[{"i", ",", "j"}], "]"}], "\[Equal]", "False"}]}], ",", "0", 
        ",", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"k", "\[Equal]", "4"}], " ", "&&", " ", 
           RowBox[{
            RowBox[{"iP", "[", 
             RowBox[{"i", ",", "j"}], "]"}], "\[Equal]", "False"}]}], ",", 
          "0", ",", "1"}], "]"}]}], "]"}]}], "]"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"\[Theta]", "[", 
     RowBox[{"k_", ",", "i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"\[Delta]", "[", 
        RowBox[{"k", ",", "i", ",", "j"}], "]"}], " ", "\[Equal]", "1"}], ",", 
      RowBox[{"Exp", "[", 
       RowBox[{
        RowBox[{"-", 
         FractionBox[
          SuperscriptBox[
           RowBox[{"(", 
            RowBox[{
             RowBox[{"rNO", "[", 
              RowBox[{"k", ",", "i", ",", "j"}], "]"}], "-", "rNOzero"}], 
            ")"}], "2"], 
          RowBox[{"2.0", " ", 
           SuperscriptBox["\[Sigma]NO", "2"]}]]}], "-", 
        FractionBox[
         SuperscriptBox[
          RowBox[{"(", 
           RowBox[{
            RowBox[{"rHO", "[", 
             RowBox[{"k", ",", "i", ",", "j"}], "]"}], "-", "rHOzero"}], 
           ")"}], "2"], 
         RowBox[{"2.0", " ", 
          SuperscriptBox["\[Sigma]HO", "2"]}]]}], "]"}], ",", "0.0"}], 
     "]"}]}], "\[IndentingNewLine]", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"\[Nu]1", "[", "i_", "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"i", "-", "2"}], ">", "1"}], " ", "&&", " ", 
       RowBox[{
        RowBox[{"i", "+", "2"}], "\[LessEqual]", "M"}]}], " ", ",", 
      RowBox[{"0.5", 
       RowBox[{"(", 
        RowBox[{"1.0", "+", 
         RowBox[{"Tanh", "[", 
          RowBox[{
           RowBox[{"MyNorm", "[", 
            RowBox[{
             SubscriptBox["r", 
              RowBox[{"Ca", ",", 
               RowBox[{"i", "-", "2"}]}]], "-", 
             SubscriptBox["r", 
              RowBox[{"Ca", ",", 
               RowBox[{"i", "+", "2"}]}]]}], "]"}], "-", "dnu"}], "]"}]}], 
        ")"}]}], ",", "1.0"}], "]"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"\[Nu]2", "[", "i_", "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"i", "-", "2"}], ">", "1"}], " ", "&&", " ", 
       RowBox[{
        RowBox[{"i", "+", "2"}], "\[LessEqual]", "M"}]}], " ", ",", 
      RowBox[{"0.5", 
       RowBox[{"(", 
        RowBox[{"1.0", "+", 
         RowBox[{"Tanh", "[", 
          RowBox[{"pref", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"MyNorm", "[", 
              RowBox[{
               SubscriptBox["r", 
                RowBox[{"Ca", ",", 
                 RowBox[{"i", "-", "2"}]}]], "-", 
               SubscriptBox["r", 
                RowBox[{"Ca", ",", 
                 RowBox[{"i", "+", "2"}]}]]}], "]"}], "-", "dnu"}], ")"}]}], 
          "]"}]}], ")"}]}], ",", "1.0"}], "]"}]}], "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{
    "THIS", " ", "CALCULATES", " ", "THE", " ", "TOTAL", " ", "POTENTIAL", 
     " ", "ENERGY"}], " ", "*)"}], "\[IndentingNewLine]", 
   RowBox[{"Vdssp", ":=", 
    RowBox[{
     UnderoverscriptBox["\[Sum]", 
      RowBox[{"i", "=", "1"}], 
      RowBox[{"M", "-", "1"}]], 
     RowBox[{
      UnderoverscriptBox["\[Sum]", 
       RowBox[{"j", "=", "2"}], "M"], 
      RowBox[{
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{
          RowBox[{
           RowBox[{"Abs", "[", 
            RowBox[{"j", "-", "i"}], "]"}], "\[LessEqual]", "2"}], " ", "||", 
          " ", 
          RowBox[{
           RowBox[{"se", "[", 
            RowBox[{"[", "j", "]"}], "]"}], "\[Equal]", "15"}], " ", "||", 
          " ", 
          RowBox[{
           RowBox[{"MyNorm", "[", 
            RowBox[{
             SubscriptBox["r", 
              RowBox[{"O", ",", "i"}]], "-", 
             SubscriptBox["r", 
              RowBox[{"N", ",", "j"}]]}], "]"}], ">", "cutoff"}]}], ",", "0", 
         ",", "1"}], "]"}], " ", 
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{
          RowBox[{
           RowBox[{"Abs", "[", 
            RowBox[{"j", "-", "i"}], "]"}], "<", "18"}], " ", "&&", " ", 
          RowBox[{
           RowBox[{"Abs", "[", 
            RowBox[{"j", "-", "i"}], "]"}], "\[GreaterEqual]", "4"}]}], ",", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"\[Lambda]1", "[", 
             RowBox[{"i", ",", "j"}], "]"}], 
            RowBox[{"\[Theta]", "[", 
             RowBox[{"1", ",", "i", ",", "j"}], "]"}]}], "+", 
           RowBox[{
            RowBox[{"\[Lambda]2", "[", 
             RowBox[{"i", ",", "j"}], "]"}], 
            RowBox[{"\[Theta]", "[", 
             RowBox[{"1", ",", "i", ",", "j"}], "]"}], 
            RowBox[{"\[Theta]", "[", 
             RowBox[{"2", ",", "i", ",", "j"}], "]"}]}], "+", 
           RowBox[{
            RowBox[{"\[Lambda]3", "[", 
             RowBox[{"i", ",", "j"}], "]"}], 
            RowBox[{"\[Theta]", "[", 
             RowBox[{"1", ",", "i", ",", "j"}], "]"}], 
            RowBox[{"\[Theta]", "[", 
             RowBox[{"3", ",", "i", ",", "j"}], "]"}]}], "+", 
           RowBox[{
            RowBox[{"\[Lambda]4", "[", 
             RowBox[{"i", ",", "j"}], "]"}], 
            RowBox[{"\[Theta]", "[", 
             RowBox[{"1", ",", "i", ",", "j"}], "]"}], 
            RowBox[{"\[Theta]", "[", 
             RowBox[{"4", ",", "i", ",", "j"}], "]"}]}]}], ")"}], ",", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"\[Lambda]1", "[", 
             RowBox[{"i", ",", "j"}], "]"}], 
            RowBox[{"\[Theta]", "[", 
             RowBox[{"1", ",", "i", ",", "j"}], "]"}], 
            RowBox[{"\[Nu]1", "[", "i", "]"}], 
            RowBox[{"\[Nu]2", "[", "j", "]"}]}], "+", 
           RowBox[{
            RowBox[{"\[Lambda]2", "[", 
             RowBox[{"i", ",", "j"}], "]"}], 
            RowBox[{"\[Theta]", "[", 
             RowBox[{"1", ",", "i", ",", "j"}], "]"}], 
            RowBox[{"\[Theta]", "[", 
             RowBox[{"2", ",", "i", ",", "j"}], "]"}], 
            RowBox[{"\[Nu]1", "[", "i", "]"}], 
            RowBox[{"\[Nu]2", "[", "j", "]"}]}], "+", 
           RowBox[{
            RowBox[{"\[Lambda]3", "[", 
             RowBox[{"i", ",", "j"}], "]"}], 
            RowBox[{"\[Theta]", "[", 
             RowBox[{"1", ",", "i", ",", "j"}], "]"}], 
            RowBox[{"\[Theta]", "[", 
             RowBox[{"3", ",", "i", ",", "j"}], "]"}], 
            RowBox[{"\[Nu]1", "[", "i", "]"}], 
            RowBox[{"\[Nu]2", "[", "j", "]"}]}], "+", 
           RowBox[{
            RowBox[{"\[Lambda]4", "[", 
             RowBox[{"i", ",", "j"}], "]"}], 
            RowBox[{"\[Theta]", "[", 
             RowBox[{"1", ",", "i", ",", "j"}], "]"}], 
            RowBox[{"\[Theta]", "[", 
             RowBox[{"4", ",", "i", ",", "j"}], "]"}], 
            RowBox[{"\[Nu]1", "[", "i", "]"}], 
            RowBox[{"\[Nu]2", "[", "j", "]"}]}]}], ")"}]}], "]"}]}]}]}]}], 
   "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{
     RowBox[{
     "THIS", " ", "CALCULATES", " ", "THE", " ", "POTENTIAL", " ", "ENERGY", 
      " ", "FOR", " ", "A", " ", "PARTICULAR", " ", "PAIR", " ", "OR", " ", 
      "RESIDUES", " ", "I"}], ",", "J"}], " ", "*)"}], "\[IndentingNewLine]", 
   
   RowBox[{
    RowBox[{"VDSSP", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"i", "\[Equal]", "M"}], " ", "||", " ", 
       RowBox[{"j", "\[Equal]", "1"}], " ", "||", " ", 
       RowBox[{
        RowBox[{"Abs", "[", 
         RowBox[{"j", "-", "i"}], "]"}], "\[LessEqual]", "2"}], " ", "||", 
       " ", 
       RowBox[{
        RowBox[{"se", "[", 
         RowBox[{"[", "j", "]"}], "]"}], "\[Equal]", "15"}], " ", "||", " ", 
       RowBox[{
        RowBox[{"MyNorm", "[", 
         RowBox[{
          SubscriptBox["r", 
           RowBox[{"O", ",", "i"}]], "-", 
          SubscriptBox["r", 
           RowBox[{"N", ",", "j"}]]}], "]"}], ">", "cutoff"}]}], ",", "0.0", 
      ",", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{"Abs", "[", 
           RowBox[{"j", "-", "i"}], "]"}], "<", "18"}], " ", "&&", " ", 
         RowBox[{
          RowBox[{"Abs", "[", 
           RowBox[{"j", "-", "i"}], "]"}], "\[GreaterEqual]", "4"}]}], ",", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"\[Lambda]1", "[", 
            RowBox[{"i", ",", "j"}], "]"}], 
           RowBox[{"\[Theta]", "[", 
            RowBox[{"1", ",", "i", ",", "j"}], "]"}]}], "+", 
          RowBox[{
           RowBox[{"\[Lambda]2", "[", 
            RowBox[{"i", ",", "j"}], "]"}], 
           RowBox[{"\[Theta]", "[", 
            RowBox[{"1", ",", "i", ",", "j"}], "]"}], 
           RowBox[{"\[Theta]", "[", 
            RowBox[{"2", ",", "i", ",", "j"}], "]"}]}], "+", 
          RowBox[{
           RowBox[{"\[Lambda]3", "[", 
            RowBox[{"i", ",", "j"}], "]"}], 
           RowBox[{"\[Theta]", "[", 
            RowBox[{"1", ",", "i", ",", "j"}], "]"}], 
           RowBox[{"\[Theta]", "[", 
            RowBox[{"3", ",", "i", ",", "j"}], "]"}]}], "+", 
          RowBox[{
           RowBox[{"\[Lambda]4", "[", 
            RowBox[{"i", ",", "j"}], "]"}], 
           RowBox[{"\[Theta]", "[", 
            RowBox[{"1", ",", "i", ",", "j"}], "]"}], 
           RowBox[{"\[Theta]", "[", 
            RowBox[{"4", ",", "i", ",", "j"}], "]"}]}]}], ")"}], ",", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{
           RowBox[{"\[Lambda]1", "[", 
            RowBox[{"i", ",", "j"}], "]"}], 
           RowBox[{"\[Theta]", "[", 
            RowBox[{"1", ",", "i", ",", "j"}], "]"}], 
           RowBox[{"\[Nu]1", "[", "i", "]"}], 
           RowBox[{"\[Nu]2", "[", "j", "]"}]}], "+", 
          RowBox[{
           RowBox[{"\[Lambda]2", "[", 
            RowBox[{"i", ",", "j"}], "]"}], 
           RowBox[{"\[Theta]", "[", 
            RowBox[{"1", ",", "i", ",", "j"}], "]"}], 
           RowBox[{"\[Theta]", "[", 
            RowBox[{"2", ",", "i", ",", "j"}], "]"}], 
           RowBox[{"\[Nu]1", "[", "i", "]"}], 
           RowBox[{"\[Nu]2", "[", "j", "]"}]}], "+", 
          RowBox[{
           RowBox[{"\[Lambda]3", "[", 
            RowBox[{"i", ",", "j"}], "]"}], 
           RowBox[{"\[Theta]", "[", 
            RowBox[{"1", ",", "i", ",", "j"}], "]"}], 
           RowBox[{"\[Theta]", "[", 
            RowBox[{"3", ",", "i", ",", "j"}], "]"}], 
           RowBox[{"\[Nu]1", "[", "i", "]"}], 
           RowBox[{"\[Nu]2", "[", "j", "]"}]}], "+", 
          RowBox[{
           RowBox[{"\[Lambda]4", "[", 
            RowBox[{"i", ",", "j"}], "]"}], 
           RowBox[{"\[Theta]", "[", 
            RowBox[{"1", ",", "i", ",", "j"}], "]"}], 
           RowBox[{"\[Theta]", "[", 
            RowBox[{"4", ",", "i", ",", "j"}], "]"}], 
           RowBox[{"\[Nu]1", "[", "i", "]"}], 
           RowBox[{"\[Nu]2", "[", "j", "]"}]}]}], ")"}]}], "]"}]}], 
     "]"}]}]}]}]], "Input",
 CellOpen->False,
 CellChangeTimes->{{3.4688545231987553`*^9, 3.468854591733592*^9}, {
   3.46885469809086*^9, 3.468854712306538*^9}, {3.468854752252572*^9, 
   3.468854888823358*^9}, {3.468854934100111*^9, 3.468854938119452*^9}, {
   3.4688551412646914`*^9, 3.468855392880972*^9}, {3.468855559233973*^9, 
   3.468855559887401*^9}, {3.468855607223339*^9, 3.4688556132112923`*^9}, {
   3.468856005341827*^9, 3.468856019661236*^9}, {3.468856186726363*^9, 
   3.4688561934160547`*^9}, {3.468856275029667*^9, 3.468856325584127*^9}, {
   3.4688563834264107`*^9, 3.4688564010607567`*^9}, {3.468856472039454*^9, 
   3.468856560114645*^9}, {3.468856649460094*^9, 3.468856714871167*^9}, {
   3.468856878141852*^9, 3.46885694397437*^9}, 3.468857067921371*^9, {
   3.468858009915308*^9, 3.468858011729107*^9}, {3.468858041804687*^9, 
   3.468858106570197*^9}, {3.468858164091794*^9, 3.468858164390669*^9}, {
   3.468858195379747*^9, 3.46885832355942*^9}, {3.46885842587939*^9, 
   3.468858544400716*^9}, {3.468858608278529*^9, 3.468858650600499*^9}, {
   3.468858729321307*^9, 3.4688588790179234`*^9}, {3.468858985244705*^9, 
   3.4688590146875277`*^9}, {3.468859079936151*^9, 3.468859085113804*^9}, {
   3.468859116799193*^9, 3.468859136314533*^9}, {3.468859267094181*^9, 
   3.468859335903599*^9}, {3.468859407223928*^9, 3.468859416937151*^9}, {
   3.468859451039817*^9, 3.468859525464444*^9}, {3.468859624182054*^9, 
   3.4688596529330273`*^9}, {3.468859690038817*^9, 3.468859727512344*^9}, {
   3.468859775627851*^9, 3.4688599201589823`*^9}, {3.468860687353223*^9, 
   3.468860698488188*^9}, {3.468860744299596*^9, 3.468860757782031*^9}, {
   3.468861902865514*^9, 3.468861925163712*^9}, {3.468940498807138*^9, 
   3.468940514255422*^9}, {3.468940594671103*^9, 3.468940649034878*^9}, {
   3.468942698120571*^9, 3.4689427091209393`*^9}, {3.4689427681289473`*^9, 
   3.468942791303822*^9}, {3.468942920426601*^9, 3.4689429236360083`*^9}, {
   3.4689432558686743`*^9, 3.468943257263085*^9}, {3.4689447061034307`*^9, 
   3.468944706570022*^9}, {3.470422398922974*^9, 3.4704224291365757`*^9}, {
   3.4704225305752077`*^9, 3.470422532281509*^9}, {3.470668638299728*^9, 
   3.470668639305203*^9}, {3.472904008302815*^9, 3.47290403310181*^9}, {
   3.472904132257702*^9, 3.472904185418413*^9}, {3.472904232331568*^9, 
   3.472904246516877*^9}, {3.472904296173376*^9, 3.472904384387455*^9}, {
   3.4729044458323603`*^9, 3.4729044921841803`*^9}, {3.472904606823674*^9, 
   3.472904635566135*^9}, {3.472904685794245*^9, 3.47290471335958*^9}, {
   3.472904746313723*^9, 3.472904786318532*^9}, {3.472904829328277*^9, 
   3.4729050579230022`*^9}, {3.472905232696363*^9, 3.47290524054075*^9}, {
   3.472905339089567*^9, 3.472905341199419*^9}, {3.472905404353219*^9, 
   3.4729054629688883`*^9}, {3.4729101666063223`*^9, 3.472910189211753*^9}, {
   3.4729103191411953`*^9, 3.4729103680395327`*^9}, {3.472910405257567*^9, 
   3.4729105003996267`*^9}, {3.512321503856079*^9, 3.5123215159614964`*^9}, {
   3.5123215755892477`*^9, 3.512321619082347*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{"THIS", " ", "BLOCK", " ", "DEFINES", " ", "THE", " ", "P_AP", " ", 
    RowBox[{"(", 
     RowBox[{"LIQUID", " ", "CRYSTAL"}], ")"}], " ", "POTENTIAL"}], " ", 
   "*)"}], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{"iMin", "=", "5"}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"iMax", "=", "8"}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"iDiff", "=", "4"}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"papPref", "=", "7.0"}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"papCut", "=", "8.0"}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"W", "=", 
     RowBox[{"{", 
      RowBox[{"0.6", ",", "0.6", ",", "0.6"}], "}"}]}], ";"}], 
   "\[IndentingNewLine]", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"\[Nu]", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"0.5", " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", 
       RowBox[{"Tanh", "[", 
        RowBox[{"papPref", 
         RowBox[{"(", 
          RowBox[{"papCut", "-", 
           RowBox[{"MyNorm", "[", 
            RowBox[{
             SubscriptBox["r", 
              RowBox[{"Ca", ",", "j"}]], "-", 
             SubscriptBox["r", 
              RowBox[{"Ca", ",", "i"}]]}], "]"}]}], ")"}]}], "]"}]}], 
      ")"}]}]}], "\[IndentingNewLine]", 
   RowBox[{"VPAP", ":=", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       RowBox[{"W", "[", 
        RowBox[{"[", "1", "]"}], "]"}]}], 
      RowBox[{
       UnderoverscriptBox["\[Sum]", 
        RowBox[{"i", "=", "1"}], 
        RowBox[{"M", "-", 
         RowBox[{"(", 
          RowBox[{"iMin", "+", 
           RowBox[{"2", "iDiff"}]}], ")"}]}]], 
       RowBox[{
        UnderoverscriptBox["\[Sum]", 
         RowBox[{"j", "=", 
          RowBox[{"i", "+", 
           RowBox[{"(", 
            RowBox[{"iMin", "+", 
             RowBox[{"2", "iDiff"}]}], ")"}]}]}], 
         RowBox[{"Min", "[", 
          RowBox[{
           RowBox[{"i", "+", 
            RowBox[{"(", 
             RowBox[{"iMax", "+", 
              RowBox[{"2", "iDiff"}]}], ")"}]}], ",", "M"}], "]"}]], 
        RowBox[{
         RowBox[{"\[Nu]", "[", 
          RowBox[{"i", ",", "j"}], "]"}], 
         RowBox[{"\[Nu]", "[", 
          RowBox[{
           RowBox[{"i", "+", "iDiff"}], ",", 
           RowBox[{"j", "-", "iDiff"}]}], "]"}]}]}]}]}], "-", 
     RowBox[{
      RowBox[{"W", "[", 
       RowBox[{"[", "2", "]"}], "]"}], 
      RowBox[{
       UnderoverscriptBox["\[Sum]", 
        RowBox[{"i", "=", "1"}], 
        RowBox[{"M", "-", 
         RowBox[{"(", 
          RowBox[{"iMax", "+", 
           RowBox[{"2", "iDiff"}], "+", "1"}], ")"}]}]], 
       RowBox[{
        UnderoverscriptBox["\[Sum]", 
         RowBox[{"j", "=", 
          RowBox[{"i", "+", 
           RowBox[{"(", 
            RowBox[{"iMax", "+", 
             RowBox[{"2", "iDiff"}], "+", "1"}], ")"}]}]}], "M"], 
        RowBox[{
         RowBox[{"\[Nu]", "[", 
          RowBox[{"i", ",", "j"}], "]"}], 
         RowBox[{"\[Nu]", "[", 
          RowBox[{
           RowBox[{"i", "+", "iDiff"}], ",", 
           RowBox[{"j", "-", "iDiff"}]}], "]"}]}]}]}]}], "-", 
     RowBox[{
      RowBox[{"W", "[", 
       RowBox[{"[", "3", "]"}], "]"}], 
      RowBox[{
       UnderoverscriptBox["\[Sum]", 
        RowBox[{"i", "=", "1"}], 
        RowBox[{"M", "-", 
         RowBox[{"(", 
          RowBox[{"iMax", "+", "iDiff", "+", "1"}], ")"}]}]], 
       RowBox[{
        UnderoverscriptBox["\[Sum]", 
         RowBox[{"j", "=", 
          RowBox[{"i", "+", 
           RowBox[{"(", 
            RowBox[{"iMax", "+", "1"}], ")"}]}]}], 
         RowBox[{"M", "-", "iDiff"}]], 
        RowBox[{
         RowBox[{"\[Nu]", "[", 
          RowBox[{"i", ",", "j"}], "]"}], 
         RowBox[{"\[Nu]", "[", 
          RowBox[{
           RowBox[{"i", "+", "iDiff"}], ",", 
           RowBox[{"j", "+", "iDiff"}]}], "]"}]}]}]}]}]}]}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"VPAPij", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{
     RowBox[{
      RowBox[{"-", 
       RowBox[{"W", "[", 
        RowBox[{"[", "1", "]"}], "]"}]}], 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"i", "<=", 
          RowBox[{"M", "-", 
           RowBox[{"(", 
            RowBox[{"iMin", "+", 
             RowBox[{"2", "iDiff"}]}], ")"}]}]}], " ", "&&", " ", 
         RowBox[{"j", ">=", 
          RowBox[{"i", "+", 
           RowBox[{"(", 
            RowBox[{"iMin", "+", 
             RowBox[{"2", "iDiff"}]}], ")"}]}]}], " ", "&&", " ", 
         RowBox[{"j", "<=", 
          RowBox[{"Min", "[", 
           RowBox[{
            RowBox[{"i", "+", 
             RowBox[{"(", 
              RowBox[{"iMax", "+", 
               RowBox[{"2", "iDiff"}]}], ")"}]}], ",", "M"}], "]"}]}]}], ",", 
        
        RowBox[{
         RowBox[{"\[Nu]", "[", 
          RowBox[{"i", ",", "j"}], "]"}], 
         RowBox[{"\[Nu]", "[", 
          RowBox[{
           RowBox[{"i", "+", "iDiff"}], ",", 
           RowBox[{"j", "-", "iDiff"}]}], "]"}]}], ",", "0.0"}], "]"}]}], "-", 
     RowBox[{
      RowBox[{"W", "[", 
       RowBox[{"[", "2", "]"}], "]"}], 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"i", "<=", 
          RowBox[{"M", "-", 
           RowBox[{"(", 
            RowBox[{"iMax", "+", 
             RowBox[{"2", "iDiff"}], "+", "1"}], ")"}]}]}], " ", "&&", " ", 
         RowBox[{"j", ">=", 
          RowBox[{"i", "+", 
           RowBox[{"(", 
            RowBox[{"iMax", "+", 
             RowBox[{"2", "iDiff"}], "+", "1"}], ")"}]}]}], " ", "&&", " ", 
         RowBox[{"j", "\[LessEqual]", "M"}]}], ",", 
        RowBox[{
         RowBox[{"\[Nu]", "[", 
          RowBox[{"i", ",", "j"}], "]"}], 
         RowBox[{"\[Nu]", "[", 
          RowBox[{
           RowBox[{"i", "+", "iDiff"}], ",", 
           RowBox[{"j", "-", "iDiff"}]}], "]"}]}], ",", "0.0"}], "]"}]}], "-", 
     RowBox[{
      RowBox[{"W", "[", 
       RowBox[{"[", "3", "]"}], "]"}], 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"i", "<=", 
          RowBox[{"M", "-", 
           RowBox[{"(", 
            RowBox[{"iMax", "+", "iDiff", "+", "1"}], ")"}]}]}], " ", "&&", 
         " ", 
         RowBox[{"j", ">=", 
          RowBox[{"i", "+", 
           RowBox[{"(", 
            RowBox[{"iMax", "+", "1"}], ")"}]}]}], " ", "&&", " ", 
         RowBox[{"j", "<=", 
          RowBox[{"M", "-", "iDiff"}]}]}], ",", 
        RowBox[{
         RowBox[{"\[Nu]", "[", 
          RowBox[{"i", ",", "j"}], "]"}], 
         RowBox[{"\[Nu]", "[", 
          RowBox[{
           RowBox[{"i", "+", "iDiff"}], ",", 
           RowBox[{"j", "+", "iDiff"}]}], "]"}]}], ",", "0.0"}], 
       "]"}]}]}]}]}]}]], "Input",
 CellOpen->False,
 CellChangeTimes->{{3.470511078395874*^9, 3.4705111331393127`*^9}, {
   3.470511254251968*^9, 3.470511286660549*^9}, {3.4705113226215897`*^9, 
   3.470511410840014*^9}, {3.470511464094315*^9, 3.470511503628962*^9}, {
   3.4705116225397778`*^9, 3.470511681362752*^9}, {3.4705866895731163`*^9, 
   3.470586707438596*^9}, {3.470586762704256*^9, 3.470586884984288*^9}, {
   3.470586976761204*^9, 3.4705870756336803`*^9}, {3.470587187843832*^9, 
   3.4705872004254293`*^9}, {3.470665533641921*^9, 3.470665827695216*^9}, {
   3.4724012084561872`*^9, 3.472401209674053*^9}, {3.472410371338832*^9, 
   3.472410421748806*^9}, {3.512321631091229*^9, 3.512321646485825*^9}, {
   3.5123335807047367`*^9, 3.512333584338703*^9}, 3.5123492960506153`*^9}],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
   "THIS", " ", "BLOCK", " ", "DEFINES", " ", "THE", " ", "RAMACHANDRAN", " ",
     "POTNETIAL"}], " ", "*)"}], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{"\[Lambda]Rama", "=", "2.0"}], ";"}], "\[IndentingNewLine]", 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"phiw", "=", 
     RowBox[{"{", 
      RowBox[{
      "0.15", ",", "0.25", ",", "0.65", ",", "1.0", ",", "1.0", ",", "1.0"}], 
      "}"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"psiw", "=", 
     RowBox[{"{", 
      RowBox[{
      "0.65", ",", "0.45", ",", "0.25", ",", "1.0", ",", "1.0", ",", "1.0"}], 
      "}"}]}], ";"}], "\[IndentingNewLine]", " ", 
   RowBox[{
    RowBox[{"w", "=", 
     RowBox[{"{", 
      RowBox[{
      "1.3149", ",", "1.32016", ",", "1.0264", ",", "0.0", ",", "2.0", ",", 
       "2.0"}], "}"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"Sigma", "=", 
     RowBox[{"{", 
      RowBox[{
      "15.398", ",", "49.0521", ",", "49.0954", ",", "0.0", ",", "419.0", ",",
        "15.398"}], "}"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"phi0", "=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "1.74"}], ",", 
       RowBox[{"-", "1.265"}], ",", "1.041", ",", "0.0", ",", 
       RowBox[{"-", "0.895"}], ",", 
       RowBox[{"-", "2.25"}]}], "}"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"psi0", "=", 
     RowBox[{"{", 
      RowBox[{"2.138", ",", 
       RowBox[{"-", "0.318"}], ",", "0.78", ",", "0.0", ",", 
       RowBox[{"-", "0.820"}], ",", "2.16"}], "}"}]}], ";"}], 
   "\[IndentingNewLine]", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"psiwPr", "=", 
     RowBox[{"{", 
      RowBox[{
      "1.0", ",", "0.15", ",", "0.15", ",", "1.0", ",", "2.0", ",", "2.0"}], 
      "}"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"phiwPr", "=", 
     RowBox[{"{", 
      RowBox[{
      "1.0", ",", "1.0", ",", "1.0", ",", "1.0", ",", "1.0", ",", "1.0"}], 
      "}"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"wPr", "=", 
     RowBox[{"{", 
      RowBox[{
      "0.0", ",", "2.17", ",", "2.15", ",", "0.0", ",", "0.0", ",", "0.0"}], 
      "}"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"SigmaPr", "=", 
     RowBox[{"{", 
      RowBox[{
      "0.0", ",", "105.52", ",", "109.09", ",", "0.0", ",", "0.0", ",", 
       "0.0"}], "}"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"phi0Pr", "=", 
     RowBox[{"{", 
      RowBox[{"0.0", ",", 
       RowBox[{"-", "1.153"}], ",", 
       RowBox[{"-", "0.95"}], ",", "0.0", ",", "0.0", ",", "0.0"}], "}"}]}], 
    ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"psi0Pr", "=", 
     RowBox[{"{", 
      RowBox[{"0.0", ",", "2.4", ",", 
       RowBox[{"-", "0.218"}], ",", "0.0", ",", "0.0", ",", "0.0"}], "}"}]}], 
    ";"}], "\[IndentingNewLine]", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"aps", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"Switch", "[", 
     RowBox[{
     "j", ",", "1", ",", "1.0", ",", "2", ",", "1.0", ",", "3", ",", "1.0", 
      ",", "4", ",", "0.0", ",", "5", ",", 
      RowBox[{"ssweight5", "[", 
       RowBox[{"[", "i", "]"}], "]"}], ",", "6", ",", 
      RowBox[{"ssweight6", "[", 
       RowBox[{"[", "i", "]"}], "]"}]}], "]"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"ArcTan2", "[", 
     RowBox[{"y_", ",", "x_"}], "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{"x", "\[Equal]", "0"}], ",", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{"y", ">", "0"}], ",", 
        FractionBox["\[Pi]", "2"], ",", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{"y", "<", "0"}], ",", 
          RowBox[{"-", 
           FractionBox["\[Pi]", "2"]}], ",", "\"\<undefine\>\""}], "]"}]}], 
       "]"}], ",", 
      RowBox[{
       RowBox[{"ArcTan", "[", 
        RowBox[{"y", "/", "x"}], "]"}], "+", 
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{"x", "<", "0"}], ",", 
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{"y", "\[GreaterEqual]", "0"}], ",", "\[Pi]", ",", 
           RowBox[{"-", "\[Pi]"}]}], "]"}], ",", "0"}], "]"}]}]}], "]"}]}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"angle", "[", 
     RowBox[{"r1_", ",", "r2_", ",", "r3_", ",", "r4_"}], "]"}], ":=", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"c", "=", 
         RowBox[{"r2", "-", "r1"}]}], ",", " ", 
        RowBox[{"b", "=", 
         RowBox[{"r3", "-", "r2"}]}], ",", 
        RowBox[{"a", "=", 
         RowBox[{"r4", "-", "r3"}]}]}], "}"}], ",", 
      RowBox[{"Return", "[", 
       RowBox[{"ArcTan2", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"MyNorm", "[", "b", "]"}], " ", 
          RowBox[{"c", ".", 
           RowBox[{"Cross", "[", 
            RowBox[{"b", ",", "a"}], "]"}]}]}], ",", 
         RowBox[{
          RowBox[{"Cross", "[", 
           RowBox[{"c", ",", "b"}], "]"}], ".", 
          RowBox[{"Cross", "[", 
           RowBox[{"b", ",", "a"}], "]"}]}]}], "]"}], "]"}]}], "]"}]}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"\[Phi]", "[", "i_", "]"}], ":=", 
    RowBox[{"angle", "[", 
     RowBox[{
      SubscriptBox["r", 
       RowBox[{"Cp", ",", 
        RowBox[{"i", "-", "1"}]}]], ",", 
      SubscriptBox["r", 
       RowBox[{"N", ",", "i"}]], ",", 
      SubscriptBox["r", 
       RowBox[{"Ca", ",", "i"}]], ",", 
      SubscriptBox["r", 
       RowBox[{"Cp", ",", "i"}]]}], "]"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"\[Psi]", "[", "i_", "]"}], ":=", 
    RowBox[{"angle", "[", 
     RowBox[{
      SubscriptBox["r", 
       RowBox[{"N", ",", "i"}]], ",", 
      SubscriptBox["r", 
       RowBox[{"Ca", ",", "i"}]], ",", 
      SubscriptBox["r", 
       RowBox[{"Cp", ",", "i"}]], ",", 
      SubscriptBox["r", 
       RowBox[{"N", ",", 
        RowBox[{"i", "+", "1"}]}]]}], "]"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"VRamaI", "[", "i_", "]"}], ":=", 
    RowBox[{
     UnderoverscriptBox["\[Sum]", 
      RowBox[{"j", "=", "1"}], "6"], 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"w", "[", 
          RowBox[{"[", "j", "]"}], "]"}], "\[NotEqual]", "0.0"}], " ", "&&", 
        " ", 
        RowBox[{
         RowBox[{"aps", "[", 
          RowBox[{"i", ",", "j"}], "]"}], "\[NotEqual]", "0.0"}]}], ",", 
       RowBox[{
        RowBox[{"w", "[", 
         RowBox[{"[", "j", "]"}], "]"}], 
        RowBox[{"Exp", "[", 
         RowBox[{
          RowBox[{"-", 
           RowBox[{"Sigma", "[", 
            RowBox[{"[", "j", "]"}], "]"}]}], 
          RowBox[{"(", " ", 
           RowBox[{
            RowBox[{
             RowBox[{"phiw", "[", 
              RowBox[{"[", "j", "]"}], "]"}], " ", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                RowBox[{"Cos", "[", 
                 RowBox[{
                  RowBox[{"\[Phi]", "[", "i", "]"}], "-", 
                  RowBox[{"phi0", "[", 
                   RowBox[{"[", "j", "]"}], "]"}]}], "]"}], "-", "1.0"}], 
               ")"}], "2"]}], "+", 
            RowBox[{
             RowBox[{"psiw", "[", 
              RowBox[{"[", "j", "]"}], "]"}], " ", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                RowBox[{"Cos", "[", 
                 RowBox[{
                  RowBox[{"\[Psi]", "[", "i", "]"}], "-", 
                  RowBox[{"psi0", "[", 
                   RowBox[{"[", "j", "]"}], "]"}]}], "]"}], "-", "1.0"}], 
               ")"}], "2"]}]}], ")"}]}], "]"}]}], ",", "0.0"}], "]"}]}]}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"VRamaIPr", "[", "i_", "]"}], ":=", 
    RowBox[{
     UnderoverscriptBox["\[Sum]", 
      RowBox[{"j", "=", "1"}], "6"], 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"wPr", "[", 
          RowBox[{"[", "j", "]"}], "]"}], "\[NotEqual]", "0.0"}], " ", "&&", 
        " ", 
        RowBox[{
         RowBox[{"aps", "[", 
          RowBox[{"i", ",", "j"}], "]"}], "\[NotEqual]", "0.0"}]}], ",", 
       RowBox[{
        RowBox[{"wPr", "[", 
         RowBox[{"[", "j", "]"}], "]"}], 
        RowBox[{"Exp", "[", 
         RowBox[{
          RowBox[{"-", 
           RowBox[{"SigmaPr", "[", 
            RowBox[{"[", "j", "]"}], "]"}]}], 
          RowBox[{"(", " ", 
           RowBox[{
            RowBox[{
             RowBox[{"phiwPr", "[", 
              RowBox[{"[", "j", "]"}], "]"}], " ", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                RowBox[{"Cos", "[", 
                 RowBox[{
                  RowBox[{"\[Phi]", "[", "i", "]"}], "-", 
                  RowBox[{"phi0Pr", "[", 
                   RowBox[{"[", "j", "]"}], "]"}]}], "]"}], "-", "1.0"}], 
               ")"}], "2"]}], "+", 
            RowBox[{
             RowBox[{"psiwPr", "[", 
              RowBox[{"[", "j", "]"}], "]"}], " ", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                RowBox[{"Cos", "[", 
                 RowBox[{
                  RowBox[{"\[Psi]", "[", "i", "]"}], "-", 
                  RowBox[{"psi0Pr", "[", 
                   RowBox[{"[", "j", "]"}], "]"}]}], "]"}], "-", "1.0"}], 
               ")"}], "2"]}]}], ")"}]}], "]"}]}], ",", "0.0"}], "]"}]}]}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"Vramaij", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], " ", ":=", " ", 
    RowBox[{
     RowBox[{"-", "\[Lambda]Rama"}], " ", 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"se", "[", 
          RowBox[{"[", "i", "]"}], "]"}], "\[Equal]", "8"}], " ", "||", " ", 
        RowBox[{"i", "\[Equal]", "1"}], " ", "||", " ", 
        RowBox[{"i", "\[Equal]", "M"}], " ", "||", " ", 
        RowBox[{
         RowBox[{"aps", "[", 
          RowBox[{"i", ",", "j"}], "]"}], "\[Equal]", "0.0"}]}], ",", "0.0", 
       ",", 
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"se", "[", 
           RowBox[{"[", "i", "]"}], "]"}], "\[Equal]", "15"}], " ", ",", 
         RowBox[{
          RowBox[{"wPr", "[", 
           RowBox[{"[", "j", "]"}], "]"}], 
          RowBox[{"Exp", "[", 
           RowBox[{
            RowBox[{"-", 
             RowBox[{"SigmaPr", "[", 
              RowBox[{"[", "j", "]"}], "]"}]}], 
            RowBox[{"(", " ", 
             RowBox[{
              RowBox[{
               RowBox[{"phiwPr", "[", 
                RowBox[{"[", "j", "]"}], "]"}], " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"Cos", "[", 
                   RowBox[{
                    RowBox[{"\[Phi]", "[", "i", "]"}], "-", 
                    RowBox[{"phi0Pr", "[", 
                    RowBox[{"[", "j", "]"}], "]"}]}], "]"}], "-", "1.0"}], 
                 ")"}], "2"]}], "+", 
              RowBox[{
               RowBox[{"psiwPr", "[", 
                RowBox[{"[", "j", "]"}], "]"}], " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"Cos", "[", 
                   RowBox[{
                    RowBox[{"\[Psi]", "[", "i", "]"}], "-", 
                    RowBox[{"psi0Pr", "[", 
                    RowBox[{"[", "j", "]"}], "]"}]}], "]"}], "-", "1.0"}], 
                 ")"}], "2"]}]}], ")"}]}], "]"}]}], ",", 
         RowBox[{
          RowBox[{"w", "[", 
           RowBox[{"[", "j", "]"}], "]"}], 
          RowBox[{"Exp", "[", 
           RowBox[{
            RowBox[{"-", 
             RowBox[{"Sigma", "[", 
              RowBox[{"[", "j", "]"}], "]"}]}], 
            RowBox[{"(", " ", 
             RowBox[{
              RowBox[{
               RowBox[{"phiw", "[", 
                RowBox[{"[", "j", "]"}], "]"}], " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"Cos", "[", 
                   RowBox[{
                    RowBox[{"\[Phi]", "[", "i", "]"}], "-", 
                    RowBox[{"phi0", "[", 
                    RowBox[{"[", "j", "]"}], "]"}]}], "]"}], "-", "1.0"}], 
                 ")"}], "2"]}], "+", 
              RowBox[{
               RowBox[{"psiw", "[", 
                RowBox[{"[", "j", "]"}], "]"}], " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"Cos", "[", 
                   RowBox[{
                    RowBox[{"\[Psi]", "[", "i", "]"}], "-", 
                    RowBox[{"psi0", "[", 
                    RowBox[{"[", "j", "]"}], "]"}]}], "]"}], "-", "1.0"}], 
                 ")"}], "2"]}]}], ")"}]}], "]"}]}]}], "]"}]}], "]"}]}]}], 
   "\[IndentingNewLine]", 
   RowBox[{"Vrama", " ", ":=", " ", 
    RowBox[{
     RowBox[{"-", "\[Lambda]Rama"}], 
     RowBox[{
      UnderoverscriptBox["\[Sum]", 
       RowBox[{"i", "=", "2"}], 
       RowBox[{"M", "-", "1"}]], 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"se", "[", 
          RowBox[{"[", "i", "]"}], "]"}], "\[Equal]", "15"}], ",", 
        RowBox[{"VRamaIPr", "[", "i", "]"}], ",", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"se", "[", 
            RowBox[{"[", "i", "]"}], "]"}], "\[Equal]", "8"}], ",", "0.0", 
          ",", 
          RowBox[{"VRamaI", "[", "i", "]"}]}], "]"}]}], 
       "]"}]}]}]}]}]}]], "Input",
 CellOpen->False,
 CellChangeTimes->CompressedData["
1:eJwdzH8s1AEABfA7dSwdOl3uzNJSW5u27lAtI3Z114y5rtudEDHO5deia3GN
cRRt3W1ohZ1zrLP1y49WpC50rlYznZDIVBzSrk6UNGGn7/v+8fbZ297e7pQc
cZoThUIJIQJVIWY7c9LG047baDsIv9yP9oFDUQxfOHfgvR90L/E5BA9GvAiG
z3Ln+ZDPZUfBn70cEaySZkmhOSAgCdLfxqVB4VpsBpxyKrkAi+hqUln84iXy
v6zlCiztNt6GzYwPTdBbd/IxVKXHkA5SszvJvcGjD75KoAzAdVUK6Urx9Efo
5towAc8LKmZhR637V6jhVTC9CPvlEd7wNz1uH7y6i+oPAzWvOaT/grhwmzlc
AI3bNyKgvihSCEVLJgksHz2eCK3ZCwUwOXq9kvyN96mCPUG5tyCntqgW1tX/
0UFD/mIbbHzz2Qipq/o+mN11xwLtLn9JWdTLw3CS9W4E+slqxuETZyupIpem
ZRHml1Xr4A0l1QAjPURNsCDP0QqdmzsfwuGYtnY4Uc02QaE0cRA6TmhHYJdI
ukDuriuWoNma6YDsM/00NqFL8vIe6LUg8YfdMyYuVB77RNpetxJGduF8NDwq
k0ghV1qcAFuNKjl0DIVmwp5ALwV8+WtLHuRMBpZA+hJfDY3pokpoaS0kNbG+
VUPFNKMGCvQdOhg219sIJaXlazBB7bkBk6OWY70JV9208XBdpi+HA46d1yDL
95QGzqbKK2DOWPdTaGjgP4fOacytGYRi/l46DH8g1tgIGalB9VCilStdp2w8
9VlpIXT3a7vpSWgeGiW9e69lE5MwVqykwblHmyUw1JIlhRmS4CT4Q3VYBiNX
BFZ48ZzFDk/bRcNH9n/n2WZSx+B/GeHo8Q==
  "]],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
    RowBox[{
    "THIS", " ", "BLOCK", " ", "DEFINES", " ", "THE", " ", "WATER", " ", 
     "POTENTIAL"}], ",", " ", 
    RowBox[{
    "HELIX", " ", "POTENTIAL", " ", "AND", " ", "BURIAL", " ", 
     "POTENTIAL"}]}], " ", "*)"}], "\[IndentingNewLine]", 
  "\[IndentingNewLine]", 
  RowBox[{"(*", " ", 
   RowBox[{"WATER", " ", "POTENTIAL"}], " ", "*)"}], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{"waterkappa", " ", "=", "5.0"}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"waterkappasigma", " ", "=", " ", "7.0"}], ";"}], "\n", 
   RowBox[{
    RowBox[{"\[Rho]0", " ", "=", " ", "2.6"}], ";"}], "\n", 
   RowBox[{
    RowBox[{"contactcutoff", " ", "=", " ", "13"}], ";"}], "\n", 
   RowBox[{
    RowBox[{"rmin", "=", 
     RowBox[{"{", 
      RowBox[{"4.5", ",", "6.5"}], "}"}]}], ";"}], "\n", 
   RowBox[{
    RowBox[{"rmax", "=", 
     RowBox[{"{", 
      RowBox[{"6.5", ",", "9.5"}], "}"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"A", "=", 
     RowBox[{"ToExpression", "[", 
      RowBox[{"StringSplit", "[", 
       RowBox[{"ReadList", "[", 
        RowBox[{"\"\<gamma.dat\>\"", ",", "String"}], "]"}], "]"}], "]"}]}], 
    ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"gamma", " ", "=", " ", 
     RowBox[{"Table", "[", 
      RowBox[{
       RowBox[{"A", "[", 
        RowBox[{"[", 
         RowBox[{
          RowBox[{"210", "k"}], "+", 
          RowBox[{"If", "[", 
           RowBox[{
            RowBox[{"j", "\[GreaterEqual]", "i"}], ",", 
            RowBox[{
             RowBox[{
              RowBox[{"(", 
               RowBox[{"i", "-", "1"}], ")"}], 
              RowBox[{
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"2", " ", "20"}], " ", "-", "i"}], ")"}], "/", 
               "2"}]}], " ", "+", "j"}], ",", 
            RowBox[{
             RowBox[{
              RowBox[{"(", 
               RowBox[{"j", "-", "1"}], ")"}], 
              RowBox[{
               RowBox[{"(", 
                RowBox[{
                 RowBox[{"2", " ", "20"}], " ", "-", "j"}], ")"}], "/", 
               "2"}]}], " ", "+", "i"}]}], "]"}]}], "]"}], "]"}], ",", 
       RowBox[{"{", 
        RowBox[{"k", ",", "0", ",", "1"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"i", ",", "1", ",", "20"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"j", ",", "1", ",", "20"}], "}"}]}], "]"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"rw", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"se", "[", 
        RowBox[{"[", "i", "]"}], "]"}], "\[Equal]", "8"}], ",", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"se", "[", 
          RowBox[{"[", "j", "]"}], "]"}], "\[Equal]", "8"}], ",", 
        RowBox[{"MyNorm", "[", 
         RowBox[{
          SubscriptBox["r", 
           RowBox[{"Ca", ",", "i"}]], "-", 
          SubscriptBox["r", 
           RowBox[{"Ca", ",", "j"}]]}], "]"}], ",", 
        RowBox[{"MyNorm", "[", 
         RowBox[{
          SubscriptBox["r", 
           RowBox[{"Ca", ",", "i"}]], "-", 
          SubscriptBox["r", 
           RowBox[{"Cb", ",", "j"}]]}], "]"}]}], "]"}], ",", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"se", "[", 
          RowBox[{"[", "j", "]"}], "]"}], "\[Equal]", "8"}], ",", 
        RowBox[{"MyNorm", "[", 
         RowBox[{
          SubscriptBox["r", 
           RowBox[{"Cb", ",", "i"}]], "-", 
          SubscriptBox["r", 
           RowBox[{"Ca", ",", "j"}]]}], "]"}], ",", 
        RowBox[{"MyNorm", "[", 
         RowBox[{
          SubscriptBox["r", 
           RowBox[{"Cb", ",", "i"}]], "-", 
          SubscriptBox["r", 
           RowBox[{"Cb", ",", "j"}]]}], "]"}]}], "]"}]}], "]"}]}], " ", 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"theta", "[", 
     RowBox[{"i_", ",", "j_", ",", "iwell_"}], "]"}], ":=", 
    RowBox[{"0.25", " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", 
       RowBox[{"Tanh", "[", 
        RowBox[{"waterkappa", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"rw", "[", 
            RowBox[{"i", ",", "j"}], "]"}], "-", 
           RowBox[{"rmin", "[", 
            RowBox[{"[", "iwell", "]"}], "]"}]}], ")"}]}], "]"}]}], ")"}], 
     RowBox[{"(", 
      RowBox[{"1", "+", 
       RowBox[{"Tanh", "[", 
        RowBox[{"waterkappa", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"rmax", "[", 
            RowBox[{"[", "iwell", "]"}], "]"}], "-", 
           RowBox[{"rw", "[", 
            RowBox[{"i", ",", "j"}], "]"}]}], ")"}]}], "]"}]}], ")"}]}]}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"\[Rho]", "[", "i_", "]"}], ":=", 
    RowBox[{
     UnderoverscriptBox["\[Sum]", 
      RowBox[{"j", "=", "1"}], "M"], 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"Abs", "[", 
         RowBox[{"i", "-", "j"}], "]"}], ">", "1"}], ",", 
       RowBox[{"theta", "[", 
        RowBox[{"i", ",", "j", ",", "1"}], "]"}], ",", "0"}], "]"}]}]}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"sigma", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"0.25", " ", 
     RowBox[{"(", 
      RowBox[{"1", "-", 
       RowBox[{"Tanh", "[", 
        RowBox[{"waterkappasigma", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"\[Rho]", "[", "i", "]"}], "-", "\[Rho]0"}], ")"}]}], 
        "]"}]}], ")"}], 
     RowBox[{"(", 
      RowBox[{"1", "-", 
       RowBox[{"Tanh", "[", 
        RowBox[{"waterkappasigma", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"\[Rho]", "[", "j", "]"}], "-", "\[Rho]0"}], ")"}]}], 
        "]"}]}], ")"}]}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"Vwater", ":=", 
     RowBox[{"-", 
      RowBox[{
       UnderoverscriptBox["\[Sum]", 
        RowBox[{"i", "=", "1"}], 
        RowBox[{"M", "-", "contactcutoff"}]], 
       RowBox[{
        UnderoverscriptBox["\[Sum]", 
         RowBox[{"j", "=", 
          RowBox[{"i", "+", "contactcutoff"}]}], "M"], 
        RowBox[{
         RowBox[{"theta", "[", 
          RowBox[{"i", ",", "j", ",", "2"}], "]"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"gamma", "[", 
             RowBox[{"[", 
              RowBox[{"2", ",", 
               RowBox[{"se", "[", 
                RowBox[{"[", "i", "]"}], "]"}], ",", 
               RowBox[{"se", "[", 
                RowBox[{"[", "j", "]"}], "]"}], ",", "1"}], "]"}], "]"}], " ", 
            RowBox[{"(", 
             RowBox[{"1", "-", 
              RowBox[{"sigma", "[", 
               RowBox[{"i", ",", "j"}], "]"}]}], ")"}]}], "+", 
           RowBox[{
            RowBox[{"gamma", "[", 
             RowBox[{"[", 
              RowBox[{"2", ",", 
               RowBox[{"se", "[", 
                RowBox[{"[", "i", "]"}], "]"}], ",", 
               RowBox[{"se", "[", 
                RowBox[{"[", "j", "]"}], "]"}], ",", "2"}], "]"}], "]"}], " ", 
            RowBox[{"sigma", "[", 
             RowBox[{"i", ",", "j"}], "]"}]}]}], ")"}]}]}]}]}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"Vwaterij", "[", 
      RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
     RowBox[{"-", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"Abs", "[", 
          RowBox[{"i", "-", "j"}], "]"}], "\[GreaterEqual]", 
         "contactcutoff"}], ",", 
        RowBox[{
         RowBox[{"theta", "[", 
          RowBox[{"i", ",", "j", ",", "2"}], "]"}], " ", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{
            RowBox[{"gamma", "[", 
             RowBox[{"[", 
              RowBox[{"2", ",", 
               RowBox[{"se", "[", 
                RowBox[{"[", "i", "]"}], "]"}], ",", 
               RowBox[{"se", "[", 
                RowBox[{"[", "j", "]"}], "]"}], ",", "1"}], "]"}], "]"}], " ", 
            RowBox[{"(", 
             RowBox[{"1", "-", 
              RowBox[{"sigma", "[", 
               RowBox[{"i", ",", "j"}], "]"}]}], ")"}]}], "+", 
           RowBox[{
            RowBox[{"gamma", "[", 
             RowBox[{"[", 
              RowBox[{"2", ",", 
               RowBox[{"se", "[", 
                RowBox[{"[", "i", "]"}], "]"}], ",", 
               RowBox[{"se", "[", 
                RowBox[{"[", "j", "]"}], "]"}], ",", "2"}], "]"}], "]"}], " ", 
            RowBox[{"sigma", "[", 
             RowBox[{"i", ",", "j"}], "]"}]}]}], ")"}]}], ",", "0"}], 
       "]"}]}]}], ";"}], "\[IndentingNewLine]", "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{"BURIAL", " ", "POTENTIAL"}], " ", "*)"}], "\[IndentingNewLine]", 
   
   RowBox[{
    RowBox[{"burialkappa", "=", "4.0"}], ";"}], "\n", 
   RowBox[{
    RowBox[{"romin", " ", "=", 
     RowBox[{"{", 
      RowBox[{"0.0", ",", "3.0", ",", "6.0"}], "}"}]}], ";"}], "\n", 
   RowBox[{
    RowBox[{"romax", " ", "=", 
     RowBox[{"{", 
      RowBox[{"3.0", ",", "6.0", ",", "9.0"}], "}"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"A", "=", 
     RowBox[{"ReadList", "[", 
      RowBox[{"\"\<burial_gamma.dat\>\"", ",", "Real"}], "]"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"burialgamma", "=", 
     RowBox[{"Table", "[", 
      RowBox[{
       RowBox[{"A", "[", 
        RowBox[{"[", 
         RowBox[{
          RowBox[{"3", 
           RowBox[{"(", 
            RowBox[{"i", "-", "1"}], ")"}]}], "+", "j"}], "]"}], "]"}], ",", 
       RowBox[{"{", 
        RowBox[{"i", ",", "1", ",", "20"}], "}"}], ",", 
       RowBox[{"{", 
        RowBox[{"j", ",", "1", ",", "3"}], "}"}]}], "]"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"S", "[", 
     RowBox[{"i_", ",", "\[Mu]_"}], "]"}], ":=", 
    RowBox[{
     RowBox[{"Tanh", "[", 
      RowBox[{"burialkappa", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"\[Rho]", "[", "i", "]"}], "-", 
         RowBox[{"romin", "[", 
          RowBox[{"[", "\[Mu]", "]"}], "]"}]}], ")"}]}], "]"}], "+", 
     RowBox[{"Tanh", "[", 
      RowBox[{"burialkappa", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"romax", "[", 
          RowBox[{"[", "\[Mu]", "]"}], "]"}], "-", 
         RowBox[{"\[Rho]", "[", "i", "]"}]}], ")"}]}], "]"}]}]}], 
   "\[IndentingNewLine]", 
   RowBox[{"Vburial", ":=", 
    RowBox[{
     RowBox[{"-", "0.5"}], 
     RowBox[{
      UnderoverscriptBox["\[Sum]", 
       RowBox[{"i", "=", "1"}], "M"], 
      RowBox[{
       UnderoverscriptBox["\[Sum]", 
        RowBox[{"\[Mu]", "=", "1"}], "3"], " ", 
       RowBox[{
        RowBox[{"burialgamma", "[", 
         RowBox[{"[", 
          RowBox[{
           RowBox[{"se", "[", 
            RowBox[{"[", "i", "]"}], "]"}], ",", "\[Mu]"}], "]"}], "]"}], 
        RowBox[{"S", "[", 
         RowBox[{"i", ",", "\[Mu]"}], "]"}]}]}]}]}]}], "\[IndentingNewLine]", 
   "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{"HELIX", " ", "POTENTIAL"}], " ", "*)"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"\[Lambda]Helix", "=", "1.5"}], ";"}], "\n", 
   RowBox[{
    RowBox[{"helixgammap", "=", "2.0"}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"helixgammaw", "=", 
     RowBox[{"-", "1.0"}]}], ";"}], "\n", 
   RowBox[{
    RowBox[{"helixkappa", "=", "7.0"}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"helixkappasigma", "=", "7.0"}], ";"}], "\n", 
   RowBox[{
    RowBox[{"helixtreshold", "=", "3.0"}], ";"}], "\n", 
   RowBox[{
    RowBox[{"helixidiff", "=", "4"}], ";"}], "\n", 
   RowBox[{
    RowBox[{"helixcutoff", "=", "15.0"}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"h4prob", "=", 
     RowBox[{"{", 
      RowBox[{
       RowBox[{"-", "0.77"}], ",", 
       RowBox[{"-", "0.68"}], ",", 
       RowBox[{"-", "0.07"}], ",", 
       RowBox[{"-", "0.15"}], ",", 
       RowBox[{"-", "0.23"}], ",", 
       RowBox[{"-", "0.33"}], ",", 
       RowBox[{"-", "0.27"}], ",", " ", "0.0", ",", 
       RowBox[{"-", "0.06"}], ",", 
       RowBox[{"-", "0.23"}], ",", 
       RowBox[{"-", "0.62"}], ",", 
       RowBox[{"-", "0.65"}], ",", 
       RowBox[{"-", "0.50"}], ",", 
       RowBox[{"-", "0.41"}], ",", " ", "3.0", ",", 
       RowBox[{"-", "0.35"}], ",", 
       RowBox[{"-", "0.11"}], ",", 
       RowBox[{"-", "0.45"}], ",", 
       RowBox[{"-", "0.17"}], ",", 
       RowBox[{"-", "0.14"}]}], "}"}]}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"\[Theta]pair", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"Exp", "[", 
     RowBox[{
      RowBox[{"-", 
       FractionBox[
        SuperscriptBox[
         RowBox[{"(", 
          RowBox[{
           RowBox[{"MyNorm", "[", 
            RowBox[{
             SubscriptBox["r", 
              RowBox[{"O", ",", "i"}]], "-", 
             SubscriptBox["r", 
              RowBox[{"N", ",", "j"}]]}], "]"}], "-", "rNOzero"}], ")"}], 
         "2"], 
        RowBox[{"2.0", " ", 
         SuperscriptBox["\[Sigma]NO", "2"]}]]}], "-", 
      FractionBox[
       SuperscriptBox[
        RowBox[{"(", 
         RowBox[{
          RowBox[{"MyNorm", "[", 
           RowBox[{
            SubscriptBox["r", 
             RowBox[{"O", ",", "i"}]], "-", 
            SubscriptBox["r", 
             RowBox[{"H", ",", "j"}]]}], "]"}], "-", "rHOzero"}], ")"}], "2"], 
       RowBox[{"2.0", " ", 
        SuperscriptBox["\[Sigma]HO", "2"]}]]}], "]"}]}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"thetaHelix", "[", 
     RowBox[{"i_", ",", "j_", ",", "iwell_"}], "]"}], ":=", 
    RowBox[{"0.25", " ", 
     RowBox[{"(", 
      RowBox[{"1", "+", 
       RowBox[{"Tanh", "[", 
        RowBox[{"helixkappa", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"rw", "[", 
            RowBox[{"i", ",", "j"}], "]"}], "-", 
           RowBox[{"rmin", "[", 
            RowBox[{"[", "iwell", "]"}], "]"}]}], ")"}]}], "]"}]}], ")"}], 
     RowBox[{"(", 
      RowBox[{"1", "+", 
       RowBox[{"Tanh", "[", 
        RowBox[{"helixkappa", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"rmax", "[", 
            RowBox[{"[", "iwell", "]"}], "]"}], "-", 
           RowBox[{"rw", "[", 
            RowBox[{"i", ",", "j"}], "]"}]}], ")"}]}], "]"}]}], ")"}]}]}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"\[Rho]Helix", "[", "i_", "]"}], ":=", 
    RowBox[{
     UnderoverscriptBox["\[Sum]", 
      RowBox[{"j", "=", "1"}], "M"], 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"Abs", "[", 
         RowBox[{"i", "-", "j"}], "]"}], ">", "1"}], ",", 
       RowBox[{"thetaHelix", "[", 
        RowBox[{"i", ",", "j", ",", "1"}], "]"}], ",", "0"}], "]"}]}]}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"sigmaHelix", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"0.25", " ", 
     RowBox[{"(", 
      RowBox[{"1", "-", 
       RowBox[{"Tanh", "[", 
        RowBox[{"helixkappasigma", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"\[Rho]Helix", "[", "i", "]"}], "-", "helixtreshold"}], 
          ")"}]}], "]"}]}], ")"}], 
     RowBox[{"(", 
      RowBox[{"1", "-", 
       RowBox[{"Tanh", "[", 
        RowBox[{"helixkappasigma", 
         RowBox[{"(", 
          RowBox[{
           RowBox[{"\[Rho]Helix", "[", "j", "]"}], "-", "helixtreshold"}], 
          ")"}]}], "]"}]}], ")"}]}]}], "\[IndentingNewLine]", 
   RowBox[{"VHelix", ":=", 
    RowBox[{"-", 
     RowBox[{
      UnderoverscriptBox["\[Sum]", 
       RowBox[{"i", "=", "1"}], 
       RowBox[{"M", "-", "helixidiff", "-", "1"}]], 
      RowBox[{"\[Lambda]Helix", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"helixgammap", " ", 
          RowBox[{"(", 
           RowBox[{"1", "-", 
            RowBox[{"sigmaHelix", "[", 
             RowBox[{"i", ",", 
              RowBox[{"i", "+", "4"}]}], "]"}]}], ")"}]}], "+", " ", 
         RowBox[{"helixgammaw", " ", 
          RowBox[{"sigmaHelix", "[", 
           RowBox[{"i", ",", 
            RowBox[{"i", "+", "4"}]}], "]"}]}]}], ")"}], " ", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"h4prob", "[", 
          RowBox[{"[", 
           RowBox[{"se", "[", 
            RowBox[{"[", "i", "]"}], "]"}], "]"}], "]"}], "+", 
         RowBox[{"h4prob", "[", 
          RowBox[{"[", 
           RowBox[{"se", "[", 
            RowBox[{"[", 
             RowBox[{"i", "+", "4"}], "]"}], "]"}], "]"}], "]"}]}], ")"}], 
       " ", 
       RowBox[{"\[Theta]pair", "[", 
        RowBox[{"i", ",", 
         RowBox[{"i", "+", "4"}]}], "]"}]}]}]}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"VHelixi", "[", "i_", "]"}], ":=", 
    RowBox[{"-", 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{"i", ">", 
        RowBox[{"M", "-", "helixidiff"}]}], ",", "0.0", ",", 
       RowBox[{"\[Lambda]Helix", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"helixgammap", " ", 
           RowBox[{"(", 
            RowBox[{"1", "-", 
             RowBox[{"sigmaHelix", "[", 
              RowBox[{"i", ",", 
               RowBox[{"i", "+", "4"}]}], "]"}]}], ")"}]}], "+", " ", 
          RowBox[{"helixgammaw", " ", 
           RowBox[{"sigmaHelix", "[", 
            RowBox[{"i", ",", 
             RowBox[{"i", "+", "4"}]}], "]"}]}]}], ")"}], " ", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"h4prob", "[", 
           RowBox[{"[", 
            RowBox[{"se", "[", 
             RowBox[{"[", "i", "]"}], "]"}], "]"}], "]"}], "+", 
          RowBox[{"h4prob", "[", 
           RowBox[{"[", 
            RowBox[{"se", "[", 
             RowBox[{"[", 
              RowBox[{"i", "+", "4"}], "]"}], "]"}], "]"}], "]"}]}], ")"}], 
        " ", 
        RowBox[{"\[Theta]pair", "[", 
         RowBox[{"i", ",", 
          RowBox[{"i", "+", "4"}]}], "]"}]}]}], "]"}]}]}]}]}]], "Input",
 CellOpen->False,
 CellChangeTimes->{{3.4912367820456*^9, 3.4912368171236*^9}, {
   3.4912369024076*^9, 3.4912369101745996`*^9}, {3.4912369558136*^9, 
   3.4912370787656*^9}, {3.4912371149556*^9, 3.4912371272746*^9}, {
   3.4912387389346*^9, 3.4912387450866003`*^9}, {3.4912387930916*^9, 
   3.4912388112595997`*^9}, {3.4912388414736*^9, 3.4912388661916*^9}, {
   3.4912389047506*^9, 3.4912389141336*^9}, {3.4912390996046*^9, 
   3.4912391338176003`*^9}, {3.4912391741436*^9, 3.4912391795186*^9}, {
   3.4912404447326*^9, 3.4912404519666*^9}, {3.4912405329056*^9, 
   3.4912405336176*^9}, {3.4912405806156*^9, 3.4912406427956*^9}, {
   3.4912406728886003`*^9, 3.4912407571126003`*^9}, {3.491403767951*^9, 
   3.49140381875*^9}, {3.491403856148*^9, 3.491403921064*^9}, {
   3.4914042334709997`*^9, 3.491404254223*^9}, {3.4914043921029997`*^9, 
   3.491404394158*^9}, {3.4914075806*^9, 3.491407580881*^9}, {
   3.491407747216*^9, 3.4914077838129997`*^9}, {3.491411718149*^9, 
   3.491411733428*^9}, {3.491412204328*^9, 3.491412298362*^9}, {
   3.491412382873*^9, 3.491412487744*^9}, {3.491412532894*^9, 
   3.491412542645*^9}, {3.4914129028570004`*^9, 3.49141296115*^9}, {
   3.491414420347*^9, 3.491414442722*^9}, {3.4914144803599997`*^9, 
   3.491414496704*^9}, {3.4914145513310003`*^9, 3.491414666727*^9}, {
   3.491414700868*^9, 3.491414874782*^9}, {3.491430772488*^9, 
   3.491430778663*^9}, {3.491430818593*^9, 3.491430908502*^9}, {
   3.491430963159*^9, 3.491430968212*^9}, {3.491431055601*^9, 
   3.491431056152*^9}, {3.491431097566*^9, 3.4914312532939997`*^9}, {
   3.491431305645*^9, 3.49143131901*^9}, {3.491431424668*^9, 
   3.4914314427*^9}, {3.491431645451*^9, 3.491431689572*^9}, {
   3.491431729738*^9, 3.4914317581280003`*^9}, {3.4914318069040003`*^9, 
   3.491432018637*^9}, {3.491435150863*^9, 3.491435151727*^9}, {
   3.491435614043*^9, 3.491435614741*^9}, {3.491505941273*^9, 
   3.4915059721140003`*^9}, {3.4915074560880003`*^9, 3.49150753391*^9}, 
   3.49152942466028*^9, {3.4915302969262805`*^9, 3.4915302997042804`*^9}, {
   3.4915303324072804`*^9, 3.4915303326242805`*^9}, 3.4915317476862803`*^9, {
   3.4961593656041183`*^9, 3.496159431985414*^9}, {3.496159463935855*^9, 
   3.496159567355567*^9}, {3.496159936336125*^9, 3.496160001744379*^9}, {
   3.496160057518712*^9, 3.496160080317877*^9}, {3.496160346018867*^9, 
   3.496160380145644*^9}, {3.496160671530918*^9, 3.49616067559326*^9}, {
   3.496160713815338*^9, 3.496160768902717*^9}, {3.496160806069759*^9, 
   3.4961608181974583`*^9}, 3.496168649590753*^9, {3.496168704145046*^9, 
   3.4961687277469673`*^9}, {3.4961689029546824`*^9, 3.496168964599119*^9}, {
   3.496169050387465*^9, 3.496169060426627*^9}, {3.496169097128862*^9, 
   3.496169127336895*^9}, {3.496169296898683*^9, 3.496169304947937*^9}, {
   3.511724198753471*^9, 3.511724199209133*^9}, 3.5117242298679457`*^9, 
   3.511724264288126*^9, {3.512315837151211*^9, 3.512315889330798*^9}, {
   3.5123217310053864`*^9, 3.512321734984325*^9}, {3.512321775810339*^9, 
   3.512321800598653*^9}}],

Cell[BoxData[
 RowBox[{"\[IndentingNewLine]", 
  RowBox[{"(*", " ", 
   RowBox[{
   "THIS", " ", "BLOCK", " ", "DEFINES", " ", "THE", " ", "RAMACHANDRAN", " ",
     "POTENTIAL"}], " ", "*)"}], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{"aps", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"Switch", "[", 
     RowBox[{
     "j", ",", "1", ",", "1.0", ",", "2", ",", "1.0", ",", "3", ",", "1.0", 
      ",", "4", ",", "0.0", ",", "5", ",", 
      RowBox[{"ssweight5", "[", 
       RowBox[{"[", "i", "]"}], "]"}], ",", "6", ",", 
      RowBox[{"ssweight6", "[", 
       RowBox[{"[", "i", "]"}], "]"}]}], "]"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"ArcTan2", "[", 
     RowBox[{"y_", ",", "x_"}], "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{"x", "\[Equal]", "0"}], ",", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{"y", ">", "0"}], ",", 
        FractionBox["\[Pi]", "2"], ",", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{"y", "<", "0"}], ",", 
          RowBox[{"-", 
           FractionBox["\[Pi]", "2"]}], ",", "\"\<undefine\>\""}], "]"}]}], 
       "]"}], ",", 
      RowBox[{
       RowBox[{"ArcTan", "[", 
        RowBox[{"y", "/", "x"}], "]"}], "+", 
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{"x", "<", "0"}], ",", 
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{"y", "\[GreaterEqual]", "0"}], ",", "\[Pi]", ",", 
           RowBox[{"-", "\[Pi]"}]}], "]"}], ",", "0"}], "]"}]}]}], "]"}]}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"angle", "[", 
     RowBox[{"r1_", ",", "r2_", ",", "r3_", ",", "r4_"}], "]"}], ":=", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"c", "=", 
         RowBox[{"r2", "-", "r1"}]}], ",", " ", 
        RowBox[{"b", "=", 
         RowBox[{"r3", "-", "r2"}]}], ",", 
        RowBox[{"a", "=", 
         RowBox[{"r4", "-", "r3"}]}]}], "}"}], ",", 
      RowBox[{"Return", "[", 
       RowBox[{"ArcTan2", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"MyNorm", "[", "b", "]"}], " ", 
          RowBox[{"c", ".", 
           RowBox[{"Cross", "[", 
            RowBox[{"b", ",", "a"}], "]"}]}]}], ",", 
         RowBox[{
          RowBox[{"Cross", "[", 
           RowBox[{"c", ",", "b"}], "]"}], ".", 
          RowBox[{"Cross", "[", 
           RowBox[{"b", ",", "a"}], "]"}]}]}], "]"}], "]"}]}], "]"}]}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"\[Phi]", "[", "i_", "]"}], ":=", 
    RowBox[{"angle", "[", 
     RowBox[{
      SubscriptBox["r", 
       RowBox[{"Cp", ",", 
        RowBox[{"i", "-", "1"}]}]], ",", 
      SubscriptBox["r", 
       RowBox[{"N", ",", "i"}]], ",", 
      SubscriptBox["r", 
       RowBox[{"Ca", ",", "i"}]], ",", 
      SubscriptBox["r", 
       RowBox[{"Cp", ",", "i"}]]}], "]"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"\[Psi]", "[", "i_", "]"}], ":=", 
    RowBox[{"angle", "[", 
     RowBox[{
      SubscriptBox["r", 
       RowBox[{"N", ",", "i"}]], ",", 
      SubscriptBox["r", 
       RowBox[{"Ca", ",", "i"}]], ",", 
      SubscriptBox["r", 
       RowBox[{"Cp", ",", "i"}]], ",", 
      SubscriptBox["r", 
       RowBox[{"N", ",", 
        RowBox[{"i", "+", "1"}]}]]}], "]"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"VRamaI", "[", "i_", "]"}], ":=", 
    RowBox[{
     UnderoverscriptBox["\[Sum]", 
      RowBox[{"j", "=", "1"}], "6"], 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"w", "[", 
          RowBox[{"[", "j", "]"}], "]"}], "\[NotEqual]", "0.0"}], " ", "&&", 
        " ", 
        RowBox[{
         RowBox[{"aps", "[", 
          RowBox[{"i", ",", "j"}], "]"}], "\[NotEqual]", "0.0"}]}], ",", 
       RowBox[{
        RowBox[{"w", "[", 
         RowBox[{"[", "j", "]"}], "]"}], 
        RowBox[{"Exp", "[", 
         RowBox[{
          RowBox[{"-", 
           RowBox[{"Sigma", "[", 
            RowBox[{"[", "j", "]"}], "]"}]}], 
          RowBox[{"(", " ", 
           RowBox[{
            RowBox[{
             RowBox[{"phiw", "[", 
              RowBox[{"[", "j", "]"}], "]"}], " ", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                RowBox[{"Cos", "[", 
                 RowBox[{
                  RowBox[{"\[Phi]", "[", "i", "]"}], "-", 
                  RowBox[{"phi0", "[", 
                   RowBox[{"[", "j", "]"}], "]"}]}], "]"}], "-", "1.0"}], 
               ")"}], "2"]}], "+", 
            RowBox[{
             RowBox[{"psiw", "[", 
              RowBox[{"[", "j", "]"}], "]"}], " ", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                RowBox[{"Cos", "[", 
                 RowBox[{
                  RowBox[{"\[Psi]", "[", "i", "]"}], "-", 
                  RowBox[{"psi0", "[", 
                   RowBox[{"[", "j", "]"}], "]"}]}], "]"}], "-", "1.0"}], 
               ")"}], "2"]}]}], ")"}]}], "]"}]}], ",", "0.0"}], "]"}]}]}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"VRamaIPr", "[", "i_", "]"}], ":=", 
    RowBox[{
     UnderoverscriptBox["\[Sum]", 
      RowBox[{"j", "=", "1"}], "6"], 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"wPr", "[", 
          RowBox[{"[", "j", "]"}], "]"}], "\[NotEqual]", "0.0"}], " ", "&&", 
        " ", 
        RowBox[{
         RowBox[{"aps", "[", 
          RowBox[{"i", ",", "j"}], "]"}], "\[NotEqual]", "0.0"}]}], ",", 
       RowBox[{
        RowBox[{"wPr", "[", 
         RowBox[{"[", "j", "]"}], "]"}], 
        RowBox[{"Exp", "[", 
         RowBox[{
          RowBox[{"-", 
           RowBox[{"SigmaPr", "[", 
            RowBox[{"[", "j", "]"}], "]"}]}], 
          RowBox[{"(", " ", 
           RowBox[{
            RowBox[{
             RowBox[{"phiwPr", "[", 
              RowBox[{"[", "j", "]"}], "]"}], " ", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                RowBox[{"Cos", "[", 
                 RowBox[{
                  RowBox[{"\[Phi]", "[", "i", "]"}], "-", 
                  RowBox[{"phi0Pr", "[", 
                   RowBox[{"[", "j", "]"}], "]"}]}], "]"}], "-", "1.0"}], 
               ")"}], "2"]}], "+", 
            RowBox[{
             RowBox[{"psiwPr", "[", 
              RowBox[{"[", "j", "]"}], "]"}], " ", 
             SuperscriptBox[
              RowBox[{"(", 
               RowBox[{
                RowBox[{"Cos", "[", 
                 RowBox[{
                  RowBox[{"\[Psi]", "[", "i", "]"}], "-", 
                  RowBox[{"psi0Pr", "[", 
                   RowBox[{"[", "j", "]"}], "]"}]}], "]"}], "-", "1.0"}], 
               ")"}], "2"]}]}], ")"}]}], "]"}]}], ",", "0.0"}], "]"}]}]}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"Vramaij", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], " ", ":=", " ", 
    RowBox[{
     RowBox[{"-", "\[Lambda]Rama"}], " ", 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"se", "[", 
          RowBox[{"[", "i", "]"}], "]"}], "\[Equal]", "8"}], " ", "||", " ", 
        RowBox[{"i", "\[Equal]", "1"}], " ", "||", " ", 
        RowBox[{"i", "\[Equal]", "M"}], " ", "||", " ", 
        RowBox[{
         RowBox[{"aps", "[", 
          RowBox[{"i", ",", "j"}], "]"}], "\[Equal]", "0.0"}]}], ",", "0.0", 
       ",", 
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"se", "[", 
           RowBox[{"[", "i", "]"}], "]"}], "\[Equal]", "15"}], " ", ",", 
         RowBox[{
          RowBox[{"wPr", "[", 
           RowBox[{"[", "j", "]"}], "]"}], 
          RowBox[{"Exp", "[", 
           RowBox[{
            RowBox[{"-", 
             RowBox[{"SigmaPr", "[", 
              RowBox[{"[", "j", "]"}], "]"}]}], 
            RowBox[{"(", " ", 
             RowBox[{
              RowBox[{
               RowBox[{"phiwPr", "[", 
                RowBox[{"[", "j", "]"}], "]"}], " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"Cos", "[", 
                   RowBox[{
                    RowBox[{"\[Phi]", "[", "i", "]"}], "-", 
                    RowBox[{"phi0Pr", "[", 
                    RowBox[{"[", "j", "]"}], "]"}]}], "]"}], "-", "1.0"}], 
                 ")"}], "2"]}], "+", 
              RowBox[{
               RowBox[{"psiwPr", "[", 
                RowBox[{"[", "j", "]"}], "]"}], " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"Cos", "[", 
                   RowBox[{
                    RowBox[{"\[Psi]", "[", "i", "]"}], "-", 
                    RowBox[{"psi0Pr", "[", 
                    RowBox[{"[", "j", "]"}], "]"}]}], "]"}], "-", "1.0"}], 
                 ")"}], "2"]}]}], ")"}]}], "]"}]}], ",", 
         RowBox[{
          RowBox[{"w", "[", 
           RowBox[{"[", "j", "]"}], "]"}], 
          RowBox[{"Exp", "[", 
           RowBox[{
            RowBox[{"-", 
             RowBox[{"Sigma", "[", 
              RowBox[{"[", "j", "]"}], "]"}]}], 
            RowBox[{"(", " ", 
             RowBox[{
              RowBox[{
               RowBox[{"phiw", "[", 
                RowBox[{"[", "j", "]"}], "]"}], " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"Cos", "[", 
                   RowBox[{
                    RowBox[{"\[Phi]", "[", "i", "]"}], "-", 
                    RowBox[{"phi0", "[", 
                    RowBox[{"[", "j", "]"}], "]"}]}], "]"}], "-", "1.0"}], 
                 ")"}], "2"]}], "+", 
              RowBox[{
               RowBox[{"psiw", "[", 
                RowBox[{"[", "j", "]"}], "]"}], " ", 
               SuperscriptBox[
                RowBox[{"(", 
                 RowBox[{
                  RowBox[{"Cos", "[", 
                   RowBox[{
                    RowBox[{"\[Psi]", "[", "i", "]"}], "-", 
                    RowBox[{"psi0", "[", 
                    RowBox[{"[", "j", "]"}], "]"}]}], "]"}], "-", "1.0"}], 
                 ")"}], "2"]}]}], ")"}]}], "]"}]}]}], "]"}]}], "]"}]}]}], 
   "\[IndentingNewLine]", 
   RowBox[{"Vrama", " ", ":=", " ", 
    RowBox[{
     RowBox[{"-", "\[Lambda]Rama"}], 
     RowBox[{
      UnderoverscriptBox["\[Sum]", 
       RowBox[{"i", "=", "2"}], 
       RowBox[{"M", "-", "1"}]], 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"se", "[", 
          RowBox[{"[", "i", "]"}], "]"}], "\[Equal]", "15"}], ",", 
        RowBox[{"VRamaIPr", "[", "i", "]"}], ",", 
        RowBox[{"If", "[", 
         RowBox[{
          RowBox[{
           RowBox[{"se", "[", 
            RowBox[{"[", "i", "]"}], "]"}], "\[Equal]", "8"}], ",", "0.0", 
          ",", 
          RowBox[{"VRamaI", "[", "i", "]"}]}], "]"}]}], 
       "]"}]}]}]}]}]}]], "Input",
 CellOpen->False,
 CellChangeTimes->{{3.5118051191327744`*^9, 3.511805119425357*^9}, {
  3.512316114456573*^9, 3.512316124828285*^9}, {3.51232202244118*^9, 
  3.512322028577326*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
   "THIS", " ", "BLOCK", " ", "DEFINES", " ", "THE", " ", "SOLVENT", " ", 
    "SEPARATED", " ", "BARRIER", " ", "POTENTIAL"}], " ", "*)"}], 
  "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{"kssb", "=", "1.0"}], ";"}], "\n", 
   RowBox[{
    RowBox[{"kappassb", "=", "5.0"}], ";"}], "\n", 
   RowBox[{
    RowBox[{"rssbmin", "=", "3.0"}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"rssbmax", "=", "8.0"}], ";"}], "\n", 
   RowBox[{
    RowBox[{"minsep", "=", "13"}], ";"}], "\n", 
   RowBox[{
    RowBox[{"cbcorrection", "=", "1"}], ";"}], "\n", 
   RowBox[{
    RowBox[{"residueradii", "=", 
     RowBox[{"{", 
      RowBox[{
      "0.00", ",", "2.04", ",", "0.57", ",", "0.57", ",", "0.36", ",", "1.11",
        ",", "1.17", ",", 
       RowBox[{"-", "1.52"}], ",", "0.87", ",", "0.67", ",", "0.79", ",", 
       "1.47", ",", "1.03", ",", "1.00", ",", 
       RowBox[{"-", "0.10"}], ",", "0.26", ",", "0.37", ",", "1.21", ",", 
       "1.15", ",", "0.39"}], "}"}]}], ";"}], "\n", 
   RowBox[{"(*", " ", 
    RowBox[{
     RowBox[{
      RowBox[{
      "DEFINE", " ", "THE", " ", "SEPARATION", " ", "AS", " ", "THE", " ", 
       "CB"}], "-", 
      RowBox[{
      "CB", " ", "DISTANCE", " ", "UNLESS", " ", "ONE", " ", "OR", " ", 
       "BOTH", " ", "OF", " ", "THE", " ", "RESIDUES", " ", "ARE", " ", 
       "GLYCINES"}]}], ",", " ", 
     RowBox[{
     "IN", " ", "WHICH", " ", "CASE", " ", "THE", " ", "CA", " ", 
      "COORDINATES", " ", "ARE", " ", "USED"}]}], " ", "*)"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"rssb", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"se", "[", 
        RowBox[{"[", "i", "]"}], "]"}], "\[Equal]", "8"}], ",", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"se", "[", 
          RowBox[{"[", "j", "]"}], "]"}], "\[Equal]", "8"}], ",", 
        RowBox[{"MyNorm", "[", 
         RowBox[{
          SubscriptBox["r", 
           RowBox[{"Ca", ",", "i"}]], "-", 
          SubscriptBox["r", 
           RowBox[{"Ca", ",", "j"}]]}], "]"}], ",", 
        RowBox[{"MyNorm", "[", 
         RowBox[{
          SubscriptBox["r", 
           RowBox[{"Ca", ",", "i"}]], "-", 
          SubscriptBox["r", 
           RowBox[{"Cb", ",", "j"}]]}], "]"}]}], "]"}], ",", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"se", "[", 
          RowBox[{"[", "j", "]"}], "]"}], "\[Equal]", "8"}], ",", 
        RowBox[{"MyNorm", "[", 
         RowBox[{
          SubscriptBox["r", 
           RowBox[{"Cb", ",", "i"}]], "-", 
          SubscriptBox["r", 
           RowBox[{"Ca", ",", "j"}]]}], "]"}], ",", 
        RowBox[{"MyNorm", "[", 
         RowBox[{
          SubscriptBox["r", 
           RowBox[{"Cb", ",", "i"}]], "-", 
          SubscriptBox["r", 
           RowBox[{"Cb", ",", "j"}]]}], "]"}]}], "]"}]}], "]"}]}], 
   "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{
     RowBox[{
     "DEFINE", " ", "SHIFTED", " ", "rssbmin", " ", "AND", " ", "rssbmax", 
      " ", "WHICH", " ", "ARE", " ", "USED", " ", "IN", " ", "THE", " ", 
      "CASE", " ", "THAT", " ", "cbcorrection"}], "\[Equal]", "1"}], " ", 
    "*)"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"rssbminshifted", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"rssbmin", "+", 
     RowBox[{"residueradii", "[", 
      RowBox[{"[", 
       RowBox[{"se", "[", 
        RowBox[{"[", "i", "]"}], "]"}], "]"}], "]"}], "+", 
     RowBox[{"residueradii", "[", 
      RowBox[{"[", 
       RowBox[{"se", "[", 
        RowBox[{"[", "j", "]"}], "]"}], "]"}], "]"}]}]}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"rssbmaxshifted", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"rssbmax", "+", 
     RowBox[{"residueradii", "[", 
      RowBox[{"[", 
       RowBox[{"se", "[", 
        RowBox[{"[", "i", "]"}], "]"}], "]"}], "]"}], "+", 
     RowBox[{"residueradii", "[", 
      RowBox[{"[", 
       RowBox[{"se", "[", 
        RowBox[{"[", "j", "]"}], "]"}], "]"}], "]"}]}]}], 
   "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{"DEFINE", " ", "CUTOFF", " ", "DISTANCE"}], " ", "*)"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"rssbmaxcutoff", "[", 
      RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
     RowBox[{
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{"cbcorrection", "\[Equal]", "1"}], ",", 
        RowBox[{"rssbmaxshifted", "[", 
         RowBox[{"i", ",", "j"}], "]"}], ",", "rssbmax"}], "]"}], "+", 
      RowBox[{"(", 
       RowBox[{"10.0", "/", "kappassb"}], ")"}]}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{"(*", 
    RowBox[{
    "CALCULATE", " ", "THE", " ", "SOLVENT", " ", "SEPARATED", " ", "BARRIER",
      " ", "ENERGY", " ", "FOR", " ", "A", " ", "PAIR", " ", "OF", " ", 
     "RESIDUES"}], " ", "*)"}], "\[IndentingNewLine]", 
   RowBox[{"(*", 
    RowBox[{
     RowBox[{"Vssbij", "[", 
      RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{"Abs", "[", 
          RowBox[{"i", "-", "j"}], "]"}], "<", "minsep"}], " ", "||", " ", 
        RowBox[{
         RowBox[{"rssb", "[", 
          RowBox[{"i", ",", "j"}], "]"}], ">", 
         RowBox[{"rssbmaxcutoff", "[", 
          RowBox[{"i", ",", "j"}], "]"}]}]}], ",", "0.0", ",", 
       RowBox[{"kssb", "*", "0.5", "*", 
        RowBox[{"(", 
         RowBox[{
          RowBox[{"Tanh", "[", 
           RowBox[{"kappassb", "*", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"rssb", "[", 
               RowBox[{"i", ",", "j"}], "]"}], "-", 
              RowBox[{"If", "[", 
               RowBox[{
                RowBox[{"cbcorrection", "\[Equal]", "1"}], ",", 
                RowBox[{"rssbminshifted", "[", 
                 RowBox[{"i", ",", "j"}], "]"}], ",", "rssbmin"}], "]"}]}], 
             ")"}]}], "]"}], "+", 
          RowBox[{"Tanh", "[", 
           RowBox[{"kappassb", "*", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"If", "[", 
               RowBox[{
                RowBox[{"cbcorrection", "\[Equal]", "1"}], ",", 
                RowBox[{"rssbmaxshifted", "[", 
                 RowBox[{"i", ",", "j"}], "]"}], ",", "rssbmax"}], "]"}], "-", 
              RowBox[{"rssb", "[", 
               RowBox[{"i", ",", "j"}], "]"}]}], ")"}]}], "]"}]}], ")"}]}]}], 
      "]"}]}], "*)"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"Vssbij", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"Abs", "[", 
        RowBox[{"i", "-", "j"}], "]"}], "<", "minsep"}], ",", "0.0", ",", 
      RowBox[{"kssb", "*", "0.5", "*", 
       RowBox[{"(", 
        RowBox[{
         RowBox[{"Tanh", "[", 
          RowBox[{"kappassb", "*", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"rssb", "[", 
              RowBox[{"i", ",", "j"}], "]"}], "-", 
             RowBox[{"If", "[", 
              RowBox[{
               RowBox[{"cbcorrection", "\[Equal]", "1"}], ",", 
               RowBox[{"rssbminshifted", "[", 
                RowBox[{"i", ",", "j"}], "]"}], ",", "rssbmin"}], "]"}]}], 
            ")"}]}], "]"}], "+", 
         RowBox[{"Tanh", "[", 
          RowBox[{"kappassb", "*", 
           RowBox[{"(", 
            RowBox[{
             RowBox[{"If", "[", 
              RowBox[{
               RowBox[{"cbcorrection", "\[Equal]", "1"}], ",", 
               RowBox[{"rssbmaxshifted", "[", 
                RowBox[{"i", ",", "j"}], "]"}], ",", "rssbmax"}], "]"}], "-", 
             
             RowBox[{"rssb", "[", 
              RowBox[{"i", ",", "j"}], "]"}]}], ")"}]}], "]"}]}], ")"}]}]}], 
     "]"}]}], "\[IndentingNewLine]", "\[IndentingNewLine]", 
   RowBox[{"(*", " ", 
    RowBox[{
    "CALCULATE", " ", "THE", " ", "TOTAL", " ", "SOLVENT", " ", "SEPARATED", 
     " ", "BARRIER", " ", "ENERGY"}], " ", "*)"}], "\[IndentingNewLine]", 
   RowBox[{"Vssb", ":=", 
    RowBox[{
     UnderoverscriptBox["\[Sum]", 
      RowBox[{"i", "=", "1"}], 
      RowBox[{"M", "-", "1"}]], 
     RowBox[{"(", 
      RowBox[{
       UnderoverscriptBox["\[Sum]", 
        RowBox[{"j", "=", 
         RowBox[{"i", "+", "1"}]}], "M"], 
       RowBox[{"Vssbij", "[", 
        RowBox[{"i", ",", "j"}], "]"}]}], ")"}]}]}]}]}]], "Input",
 CellOpen->False,
 CellChangeTimes->{{3.512334374316305*^9, 3.512334561973007*^9}, {
   3.5123345975303783`*^9, 3.5123346394852047`*^9}, {3.5123346800437937`*^9, 
   3.512334702297639*^9}, {3.512334804848846*^9, 3.512334858189114*^9}, {
   3.512334929735024*^9, 3.5123349548524017`*^9}, 3.5123350158082323`*^9, {
   3.5123350808639183`*^9, 3.5123351543303223`*^9}, {3.512335427131501*^9, 
   3.5123354896958303`*^9}, {3.512340910822135*^9, 3.512341063069209*^9}, {
   3.5123411958281116`*^9, 3.512341360456814*^9}, {3.512341419962578*^9, 
   3.512341433075323*^9}, {3.512341469787331*^9, 3.5123414897850657`*^9}, {
   3.512341530113044*^9, 3.5123415324164457`*^9}, {3.512341562473892*^9, 
   3.5123416153786077`*^9}, {3.512341653632106*^9, 3.512341669363619*^9}, {
   3.512341716832548*^9, 3.5123418833052273`*^9}, {3.512341936361721*^9, 
   3.51234201869131*^9}, {3.512342051993537*^9, 3.512342157437056*^9}, {
   3.5123423094506598`*^9, 3.512342339274049*^9}, {3.512342382254984*^9, 
   3.512342382523411*^9}, {3.512342712247609*^9, 3.512342717333022*^9}, {
   3.51234291312582*^9, 3.5123429661412907`*^9}, {3.512343125923366*^9, 
   3.512343126144085*^9}, {3.512343159541482*^9, 3.512343180409527*^9}, {
   3.512343210983222*^9, 3.5123432541630774`*^9}, {3.5123457598095713`*^9, 
   3.512345764591309*^9}, 3.5123458020234737`*^9, {3.512345864588251*^9, 
   3.5123458668123217`*^9}, {3.512345903134356*^9, 3.512346001229871*^9}, {
   3.512349801985283*^9, 3.5123499951308947`*^9}, {3.512350126229039*^9, 
   3.5123501301930113`*^9}, {3.512350443793626*^9, 3.512350482408517*^9}, {
   3.512350879308119*^9, 3.512350888961907*^9}, {3.5123510850982924`*^9, 
   3.512351085280636*^9}, {3.512351539256563*^9, 3.5123515653785667`*^9}, {
   3.512351676650198*^9, 3.512351710406589*^9}, {3.512351768445006*^9, 
   3.512351778216229*^9}, {3.512353090641622*^9, 3.512353165722602*^9}, {
   3.5123532502716637`*^9, 3.512353284722406*^9}, {3.512353369908868*^9, 
   3.5123533784254847`*^9}, {3.5123534122009687`*^9, 
   3.5123534432903223`*^9}, {3.5123557768439913`*^9, 3.512355866261181*^9}, {
   3.512761250748369*^9, 3.512761258843855*^9}, {3.5127680658554688`*^9, 
   3.512768070057308*^9}, {3.512768248823433*^9, 3.512768252580122*^9}, {
   3.51276829937626*^9, 3.5127683530882053`*^9}, {3.5127686629539824`*^9, 
   3.512768672407872*^9}, {3.512920714797585*^9, 3.5129207441776*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
    RowBox[{"THIS", " ", "BLOCK", " ", "DEFINES", " ", "THE", " ", "AMH"}], 
    "-", 
    RowBox[{"GO", " ", "POTENTIAL"}]}], " ", "*)"}], "\[IndentingNewLine]", 
  RowBox[{"(*", " ", 
   RowBox[{"LOAD", " ", "NATIVE", " ", 
    RowBox[{"DISTANCES", ":", " ", 
     RowBox[{
      RowBox[{
      "THE", " ", "INPUT", " ", "FILE", " ", "CAN", " ", "BE", " ", 
       "GENERATED", " ", "USING", " ", 
       RowBox[{"GetAMHGoDistances", ".", "py"}], " ", "with", " ", "the"}], 
      " ", "-", 
      RowBox[{"s", " ", "option"}]}]}]}], " ", "*)"}], "\[IndentingNewLine]", 
  
  RowBox[{
   RowBox[{
    RowBox[{"rnativeijk", "=", 
     RowBox[{"Import", "[", 
      RowBox[{"\"\<nativedistances_A.data\>\"", ",", "\"\<Table\>\""}], 
      "]"}]}], ";"}], " ", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"rnative", "[", 
      RowBox[{"i_", ",", "j_", ",", "k_"}], "]"}], ":=", 
     RowBox[{
      RowBox[{"rnativeijk", "[", 
       RowBox[{"[", "i", "]"}], "]"}], "[", 
      RowBox[{"[", 
       RowBox[{
        RowBox[{"4", 
         RowBox[{"(", 
          RowBox[{"j", "-", "1"}], ")"}]}], "+", "k"}], "]"}], "]"}]}], ";"}],
    "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"minSep", "=", "3"}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"kamhgo", "=", "1.0"}], ";", " ", 
    RowBox[{"(*", " ", 
     RowBox[{"DEFINE", " ", "PARAMETERS"}], " ", "*)"}], "\n", 
    RowBox[{"pamhgo", "=", "3"}], ";"}], "\n", 
   RowBox[{
    RowBox[{"nativedistancecutoff", "=", "8.0"}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"amhgogamma", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"Abs", "[", 
        RowBox[{"i", "-", "j"}], "]"}], "<", "minSep"}], ",", "0.0", ",", 
      RowBox[{"If", "[", 
       RowBox[{
        RowBox[{
         RowBox[{"Abs", "[", 
          RowBox[{"i", "-", "j"}], "]"}], "<", "5"}], ",", "0.125", ",", 
        "0.5"}], "]"}]}], "]"}]}], " ", 
   RowBox[{"(*", " ", 
    RowBox[{
     RowBox[{
     "MAKE", " ", "SURE", " ", "THAT", " ", "THIS", " ", "IS", " ", "WHAT", 
      " ", "YOU", " ", "ARE", " ", "USING", " ", "IN", " ", "amh"}], "-", 
     RowBox[{"go", ".", "gamma"}]}], " ", "*)"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"amhgosigma", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{
     RowBox[{"Abs", "[", 
      RowBox[{"i", "-", "j"}], "]"}], "^", 
     RowBox[{"(", "0.15", ")"}]}]}], " ", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"rcacaamhgo", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"MyNorm", "[", 
     RowBox[{
      SubscriptBox["r", 
       RowBox[{"Ca", ",", "i"}]], "-", 
      SubscriptBox["r", 
       RowBox[{"Ca", ",", "j"}]]}], "]"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"rcacbamhgo", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"se", "[", 
        RowBox[{"[", "j", "]"}], "]"}], "\[Equal]", "8"}], ",", "\[Infinity]",
       ",", 
      RowBox[{"MyNorm", "[", 
       RowBox[{
        SubscriptBox["r", 
         RowBox[{"Ca", ",", "i"}]], "-", 
        SubscriptBox["r", 
         RowBox[{"Cb", ",", "j"}]]}], "]"}]}], "]"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"rcbcaamhgo", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"se", "[", 
        RowBox[{"[", "i", "]"}], "]"}], "\[Equal]", "8"}], ",", "\[Infinity]",
       ",", 
      RowBox[{"MyNorm", "[", 
       RowBox[{
        SubscriptBox["r", 
         RowBox[{"Cb", ",", "i"}]], "-", 
        SubscriptBox["r", 
         RowBox[{"Ca", ",", "j"}]]}], "]"}]}], "]"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"rcbcbamhgo", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"se", "[", 
         RowBox[{"[", "i", "]"}], "]"}], "\[Equal]", "8"}], " ", "||", " ", 
       RowBox[{
        RowBox[{"se", "[", 
         RowBox[{"[", "j", "]"}], "]"}], "\[Equal]", "8"}]}], ",", 
      "\[Infinity]", ",", 
      RowBox[{"MyNorm", "[", 
       RowBox[{
        SubscriptBox["r", 
         RowBox[{"Cb", ",", "i"}]], "-", 
        SubscriptBox["r", 
         RowBox[{"Cb", ",", "j"}]]}], "]"}]}], "]"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"ramhgoijk", "[", 
      RowBox[{"i_", ",", "j_", ",", "k_"}], "]"}], ":=", 
     RowBox[{"Switch", "[", 
      RowBox[{"k", ",", "1", ",", 
       RowBox[{"rcacaamhgo", "[", 
        RowBox[{"i", ",", "j"}], "]"}], ",", "2", ",", 
       RowBox[{"rcacbamhgo", "[", 
        RowBox[{"i", ",", "j"}], "]"}], ",", "3", ",", 
       RowBox[{"rcbcaamhgo", "[", 
        RowBox[{"i", ",", "j"}], "]"}], ",", "4", ",", 
       RowBox[{"rcbcbamhgo", "[", 
        RowBox[{"i", ",", "j"}], "]"}]}], "]"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"dr", "[", 
      RowBox[{"i_", ",", "j_", ",", "k_"}], "]"}], ":=", 
     RowBox[{
      RowBox[{"ramhgoijk", "[", 
       RowBox[{"i", ",", "j", ",", "k"}], "]"}], "-", 
      RowBox[{"rnative", "[", 
       RowBox[{"i", ",", "j", ",", "k"}], "]"}]}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{"(*", 
    RowBox[{
     RowBox[{"Vamhgoijk", "[", 
      RowBox[{"i_", ",", "j_", ",", "k_"}], "]"}], ":=", 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{
        RowBox[{
         RowBox[{
          RowBox[{"dr", "[", 
           RowBox[{"i", ",", "j", ",", "k"}], "]"}], "^", "2"}], ">", 
         RowBox[{"27.6", "*", 
          RowBox[{
           RowBox[{"amhgosigma", "[", 
            RowBox[{"i", ",", "j"}], "]"}], "^", "2"}]}]}], " ", "||", " ", 
        RowBox[{
         RowBox[{"rnative", "[", 
          RowBox[{"i", ",", "j", ",", "k"}], "]"}], ">", 
         "nativedistancecutoff"}]}], ",", "0.0", ",", 
       RowBox[{
        RowBox[{"-", "kamhgo"}], "*", 
        RowBox[{"amhgogamma", "[", 
         RowBox[{"i", ",", "j"}], "]"}], "*", 
        RowBox[{"Exp", "[", 
         RowBox[{
          RowBox[{"-", 
           RowBox[{
            RowBox[{"dr", "[", 
             RowBox[{"i", ",", "j", ",", "k"}], "]"}], "^", "2"}]}], "/", 
          RowBox[{"(", 
           RowBox[{"2", "*", 
            RowBox[{
             RowBox[{"amhgosigma", "[", 
              RowBox[{"i", ",", "j"}], "]"}], "^", "2"}]}], ")"}]}], 
         "]"}]}]}], "]"}]}], "*)"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"Vamhgoijk", "[", 
     RowBox[{"i_", ",", "j_", ",", "k_"}], "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"Abs", "[", 
         RowBox[{"i", "-", "j"}], "]"}], "<", "minSep"}], "||", 
       RowBox[{
        RowBox[{"rnative", "[", 
         RowBox[{"i", ",", "j", ",", "k"}], "]"}], ">", 
        "nativedistancecutoff"}]}], ",", "0.0", ",", 
      RowBox[{
       RowBox[{"-", "kamhgo"}], "*", 
       RowBox[{"amhgogamma", "[", 
        RowBox[{"i", ",", "j"}], "]"}], "*", 
       RowBox[{"Exp", "[", 
        RowBox[{
         RowBox[{"-", 
          RowBox[{
           RowBox[{"dr", "[", 
            RowBox[{"i", ",", "j", ",", "k"}], "]"}], "^", "2"}]}], "/", 
         RowBox[{"(", 
          RowBox[{"2", "*", 
           RowBox[{
            RowBox[{"amhgosigma", "[", 
             RowBox[{"i", ",", "j"}], "]"}], "^", "2"}]}], ")"}]}], "]"}]}]}],
      "]"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"Vamhgoca", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{
     RowBox[{"Vamhgoijk", "[", 
      RowBox[{"i", ",", "j", ",", "1"}], "]"}], "+", 
     RowBox[{"Vamhgoijk", "[", 
      RowBox[{"i", ",", "j", ",", "2"}], "]"}]}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"Vamhgocb", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{
     RowBox[{"Vamhgoijk", "[", 
      RowBox[{"i", ",", "j", ",", "3"}], "]"}], "+", 
     RowBox[{"Vamhgoijk", "[", 
      RowBox[{"i", ",", "j", ",", "4"}], "]"}]}]}], "\[IndentingNewLine]", 
   RowBox[{"Vamhgo", ":=", 
    RowBox[{"0.5", "*", 
     RowBox[{
      UnderoverscriptBox["\[Sum]", 
       RowBox[{"i", "=", "1"}], "M"], 
      RowBox[{"(", 
       RowBox[{
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           UnderoverscriptBox["\[Sum]", 
            RowBox[{"j", "=", "1"}], "M"], 
           RowBox[{"Vamhgoca", "[", 
            RowBox[{"i", ",", "j"}], "]"}]}], ")"}], "^", "pamhgo"}], "+", 
        RowBox[{
         RowBox[{"(", 
          RowBox[{
           UnderoverscriptBox["\[Sum]", 
            RowBox[{"j", "=", "1"}], "M"], 
           RowBox[{"Vamhgocb", "[", 
            RowBox[{"i", ",", "j"}], "]"}]}], ")"}], "^", "pamhgo"}]}], 
       ")"}]}]}]}], 
   RowBox[{"(*", " ", 
    RowBox[{
     RowBox[{
     "FOR", " ", "EVEN", " ", "VALUES", " ", "OF", " ", "pamhgo", " ", 
      "THERE", " ", "SHOULD", " ", "BE", " ", "A", " ", "NEGATIVE", " ", 
      "SIGN", " ", "IN", " ", "FRONT"}], ",", " ", 
     RowBox[{"OTHERWISE", " ", "NOT"}]}], " ", "*)"}], 
   "\[IndentingNewLine]"}]}]], "Input",
 CellChangeTimes->{{3.5127834280716753`*^9, 3.512783449970538*^9}, 
   3.512783483981803*^9, {3.5127835638653097`*^9, 3.512783621757844*^9}, {
   3.512783767315929*^9, 3.512783781624098*^9}, {3.512784423143238*^9, 
   3.512784429741341*^9}, {3.5127845477953053`*^9, 3.512784557778577*^9}, {
   3.5127845889754763`*^9, 3.512784644748103*^9}, {3.512784716966358*^9, 
   3.512784735295278*^9}, 3.5127848166684847`*^9, {3.512784939533637*^9, 
   3.512784954766465*^9}, {3.5127851055654793`*^9, 3.512785139033927*^9}, {
   3.512785696217544*^9, 3.512785717042426*^9}, {3.512785979885357*^9, 
   3.5127859938340273`*^9}, {3.5127860859413147`*^9, 3.51278611882526*^9}, {
   3.512786243845785*^9, 3.512786244616432*^9}, {3.512786287779727*^9, 
   3.512786288290999*^9}, 3.5127863354357758`*^9, 3.5127865542540083`*^9, {
   3.512786615463518*^9, 3.512786622357679*^9}, {3.512786668919016*^9, 
   3.512786727482285*^9}, {3.512786806754056*^9, 3.5127868322051373`*^9}, {
   3.512786922044744*^9, 3.5127869507059927`*^9}, 3.5127871856592493`*^9, {
   3.5127874542790947`*^9, 3.512787457618371*^9}, {3.512788337520791*^9, 
   3.512788466286641*^9}, {3.512788527843295*^9, 3.512788529846381*^9}, {
   3.5127885625281363`*^9, 3.5127886133285913`*^9}, {3.512788659048429*^9, 
   3.512788688306876*^9}, {3.5127887193849373`*^9, 3.512788720111474*^9}, {
   3.512788774925132*^9, 3.51278881118931*^9}, {3.512788860195463*^9, 
   3.512788907908037*^9}, {3.512912309860483*^9, 3.512912322603016*^9}, {
   3.512916609660343*^9, 3.512916638123743*^9}, {3.5129166737850018`*^9, 
   3.5129168265317097`*^9}, {3.512917036624991*^9, 3.512917087550788*^9}, {
   3.512920753884132*^9, 3.5129208245159063`*^9}, {3.512920886677244*^9, 
   3.5129208919288673`*^9}, {3.51292093667829*^9, 3.512920972928206*^9}, {
   3.5129210384678593`*^9, 3.512921067652079*^9}, {3.512921106099114*^9, 
   3.512921131913453*^9}, {3.512921672454811*^9, 3.5129216896357107`*^9}, {
   3.512921805896039*^9, 3.512921810655373*^9}, {3.512921885244454*^9, 
   3.5129218883037024`*^9}, {3.512921967930595*^9, 3.5129219693583*^9}, {
   3.512922021425391*^9, 3.5129220258098373`*^9}, {3.5129249603275557`*^9, 
   3.5129250802669897`*^9}, {3.512925297128887*^9, 3.512925316550137*^9}, {
   3.512925351063332*^9, 3.5129254357839603`*^9}, {3.512925523940778*^9, 
   3.512925545987041*^9}, {3.51292586598794*^9, 3.51292586929941*^9}, {
   3.512925913786984*^9, 3.5129259438550587`*^9}, {3.512925996279181*^9, 
   3.512926005443447*^9}, {3.512926082504704*^9, 3.512926118021494*^9}, {
   3.512926171438538*^9, 3.51292619728224*^9}, {3.512926322026973*^9, 
   3.512926324838036*^9}, {3.512926748155814*^9, 3.5129267590553493`*^9}, {
   3.512927976768454*^9, 3.5129279862213593`*^9}, {3.51292804971054*^9, 
   3.5129280790349092`*^9}, {3.512933188512272*^9, 3.512933207204545*^9}, {
   3.512933354060191*^9, 3.5129333542209787`*^9}, {3.512933580257387*^9, 
   3.512933581257339*^9}, {3.512936033756816*^9, 3.51293635695278*^9}, {
   3.512936395866111*^9, 3.512936561448762*^9}, {3.5129371447324266`*^9, 
   3.512937157941387*^9}, {3.512937267301261*^9, 3.51293735594*^9}, {
   3.512937573343522*^9, 3.512937585762814*^9}, {3.51293762337728*^9, 
   3.5129376236305017`*^9}, {3.5129376552452583`*^9, 3.512937659648745*^9}, 
   3.5129390038645*^9, {3.5129390505539083`*^9, 3.5129390555708723`*^9}, {
   3.512939086492906*^9, 3.512939109746567*^9}, {3.512939185615861*^9, 
   3.512939257694776*^9}, {3.512939371706649*^9, 3.5129394037335*^9}, {
   3.512939456674449*^9, 3.512939459542713*^9}, {3.5129395075411463`*^9, 
   3.512939519661355*^9}, {3.5129395516655283`*^9, 3.512939650000963*^9}, {
   3.512939858730277*^9, 3.512939859108571*^9}, {3.51294022364287*^9, 
   3.512940298727282*^9}, {3.512940478887858*^9, 3.512940480007552*^9}, {
   3.512941060675259*^9, 3.512941085332076*^9}, {3.512941408197279*^9, 
   3.512941453065196*^9}, {3.512941556566399*^9, 3.5129415573415422`*^9}, {
   3.5129418489579277`*^9, 3.512941849129842*^9}, {3.512941899902018*^9, 
   3.512941900073368*^9}, {3.512942288621201*^9, 3.512942304840971*^9}, {
   3.5129476543837976`*^9, 3.512947654601367*^9}, {3.512947936754839*^9, 
   3.512947936982238*^9}, {3.512948170379945*^9, 3.5129481705988503`*^9}, {
   3.513019146719233*^9, 3.513019146914679*^9}, {3.513023247746817*^9, 
   3.51302327578684*^9}, {3.513369031169059*^9, 3.513369039680031*^9}, {
   3.5135317913876677`*^9, 3.513531791514947*^9}, {3.513531849691679*^9, 
   3.513531849769435*^9}, {3.51353228138027*^9, 3.513532321571122*^9}, {
   3.51353301723657*^9, 3.513533066440051*^9}, {3.513540649669231*^9, 
   3.513540683623698*^9}, {3.513541298200078*^9, 3.513541300009598*^9}, {
   3.513541345132114*^9, 3.513541388812215*^9}, {3.513541510031004*^9, 
   3.513541511888959*^9}, {3.5135417244388323`*^9, 3.513541727688305*^9}, {
   3.5135420631551867`*^9, 3.513542161996929*^9}, {3.513542253721485*^9, 
   3.5135423421737633`*^9}, {3.513542406833084*^9, 3.513542426161009*^9}, {
   3.51354246689123*^9, 3.513542468131795*^9}, {3.5135429073287563`*^9, 
   3.513542912477518*^9}, {3.513543175234632*^9, 3.5135431753174067`*^9}, {
   3.513545422669837*^9, 3.5135454421307383`*^9}, {3.51354549756924*^9, 
   3.513545497719509*^9}, 3.513709721210884*^9, {3.513709812239777*^9, 
   3.513709827647686*^9}, {3.5139716658795147`*^9, 3.5139716814165373`*^9}, 
   3.513972421276412*^9, {3.5139724663023243`*^9, 3.5139725262960577`*^9}, {
   3.513973353635577*^9, 3.5139733669343853`*^9}, {3.513973592570792*^9, 
   3.513973593806649*^9}, {3.513973694213789*^9, 3.5139737233422956`*^9}, {
   3.5139737569525633`*^9, 3.513973764544489*^9}, {3.513973796103071*^9, 
   3.5139738017814827`*^9}, 3.513973851905484*^9, {3.513973884436392*^9, 
   3.513973957927483*^9}, {3.513973990290502*^9, 3.513973998190855*^9}, 
   3.513974032272547*^9, {3.513974153364016*^9, 3.51397420238592*^9}, {
   3.513974310982374*^9, 3.513974327142131*^9}, {3.513976042959944*^9, 
   3.513976064740404*^9}, 3.5139768452772093`*^9, {3.5140489922475367`*^9, 
   3.514048992581759*^9}, {3.514049268971786*^9, 3.5140492826009293`*^9}, {
   3.5140493599954367`*^9, 3.5140493931422167`*^9}, {3.514051997206333*^9, 
   3.5140519990842543`*^9}, {3.514057055758922*^9, 3.514057066629827*^9}, {
   3.5140571439689827`*^9, 3.5140571445044947`*^9}}],

Cell[BoxData[
 RowBox[{"\[IndentingNewLine]", 
  RowBox[{"(*", " ", 
   RowBox[{
   "THIS", " ", "BLOCK", " ", "DEFINES", " ", "THE", " ", "FRAGMENT", " ", 
    "MEMORY", " ", "POTENTIAL"}], " ", "*)"}], "\[IndentingNewLine]", 
  RowBox[{"(*", " ", 
   RowBox[{"LOAD", " ", "NATIVE", " ", 
    RowBox[{"DISTANCES", ":", " ", 
     RowBox[{
      RowBox[{
      "THE", " ", "INPUT", " ", "FILE", " ", "CAN", " ", "BE", " ", 
       "GENERATED", " ", "USING", " ", 
       RowBox[{"GetAMHGoDistances", ".", "py"}], " ", "with", " ", "the"}], 
      " ", "-", 
      RowBox[{"s", " ", "option"}]}]}]}], " ", "*)"}], "\[IndentingNewLine]", 
  
  RowBox[{
   RowBox[{
    RowBox[{"rnativeijk", "=", 
     RowBox[{"Import", "[", 
      RowBox[{"\"\<nativedistances_A.data\>\"", ",", "\"\<Table\>\""}], 
      "]"}]}], ";"}], " ", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"rnative", "[", 
      RowBox[{"i_", ",", "j_", ",", "k_"}], "]"}], ":=", 
     RowBox[{
      RowBox[{"rnativeijk", "[", 
       RowBox[{"[", "i", "]"}], "]"}], "[", 
      RowBox[{"[", 
       RowBox[{
        RowBox[{"4", 
         RowBox[{"(", 
          RowBox[{"j", "-", "1"}], ")"}]}], "+", "k"}], "]"}], "]"}]}], ";"}],
    " ", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"minSep", "=", "3"}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"kfragmem", "=", "1.0"}], ";", " ", 
    RowBox[{"(*", " ", 
     RowBox[{"DEFINE", " ", "PARAMETERS"}], " ", "*)"}], "\n", 
    RowBox[{
     RowBox[{"fragmemgamma", "[", 
      RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
     RowBox[{"If", "[", 
      RowBox[{
       RowBox[{
        RowBox[{"Abs", "[", 
         RowBox[{"i", "-", "j"}], "]"}], "<", "minSep"}], ",", "0.0", ",", 
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"Abs", "[", 
           RowBox[{"i", "-", "j"}], "]"}], ">", "12"}], ",", "0.0", ",", 
         "1.0"}], "]"}]}], "]"}]}]}], " ", 
   RowBox[{"(*", " ", 
    RowBox[{
    "MAKE", " ", "SURE", " ", "THAT", " ", "THIS", " ", "IS", " ", "WHAT", 
     " ", "YOU", " ", "ARE", " ", "USING", " ", "IN", " ", "YOUR", " ", 
     "GAMMA", " ", "FILE"}], " ", "*)"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"fragmemsigma", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{
     RowBox[{"Abs", "[", 
      RowBox[{"i", "-", "j"}], "]"}], "^", 
     RowBox[{"(", "0.15", ")"}]}]}], " ", "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"rcacafragmem", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"MyNorm", "[", 
     RowBox[{
      SubscriptBox["r", 
       RowBox[{"Ca", ",", "i"}]], "-", 
      SubscriptBox["r", 
       RowBox[{"Ca", ",", "j"}]]}], "]"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"rcacbfragmem", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"se", "[", 
        RowBox[{"[", "j", "]"}], "]"}], "\[Equal]", "8"}], ",", "\[Infinity]",
       ",", 
      RowBox[{"MyNorm", "[", 
       RowBox[{
        SubscriptBox["r", 
         RowBox[{"Ca", ",", "i"}]], "-", 
        SubscriptBox["r", 
         RowBox[{"Cb", ",", "j"}]]}], "]"}]}], "]"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"rcbcafragmem", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{"se", "[", 
        RowBox[{"[", "i", "]"}], "]"}], "\[Equal]", "8"}], ",", "\[Infinity]",
       ",", 
      RowBox[{"MyNorm", "[", 
       RowBox[{
        SubscriptBox["r", 
         RowBox[{"Cb", ",", "i"}]], "-", 
        SubscriptBox["r", 
         RowBox[{"Ca", ",", "j"}]]}], "]"}]}], "]"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"rcbcbfragmem", "[", 
     RowBox[{"i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"If", "[", 
     RowBox[{
      RowBox[{
       RowBox[{
        RowBox[{"se", "[", 
         RowBox[{"[", "i", "]"}], "]"}], "\[Equal]", "8"}], " ", "||", " ", 
       RowBox[{
        RowBox[{"se", "[", 
         RowBox[{"[", "j", "]"}], "]"}], "\[Equal]", "8"}]}], ",", 
      "\[Infinity]", ",", 
      RowBox[{"MyNorm", "[", 
       RowBox[{
        SubscriptBox["r", 
         RowBox[{"Cb", ",", "i"}]], "-", 
        SubscriptBox["r", 
         RowBox[{"Cb", ",", "j"}]]}], "]"}]}], "]"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"rfragmemijk", "[", 
      RowBox[{"i_", ",", "j_", ",", "k_"}], "]"}], ":=", 
     RowBox[{"Switch", "[", 
      RowBox[{"k", ",", "1", ",", 
       RowBox[{"rcacafragmem", "[", 
        RowBox[{"i", ",", "j"}], "]"}], ",", "2", ",", 
       RowBox[{"rcacbfragmem", "[", 
        RowBox[{"i", ",", "j"}], "]"}], ",", "3", ",", 
       RowBox[{"rcbcafragmem", "[", 
        RowBox[{"i", ",", "j"}], "]"}], ",", "4", ",", 
       RowBox[{"rcbcbfragmem", "[", 
        RowBox[{"i", ",", "j"}], "]"}]}], "]"}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{
     RowBox[{"dr", "[", 
      RowBox[{"i_", ",", "j_", ",", "k_"}], "]"}], ":=", 
     RowBox[{
      RowBox[{"rfragmemijk", "[", 
       RowBox[{"i", ",", "j", ",", "k"}], "]"}], "-", 
      RowBox[{"rnative", "[", 
       RowBox[{"i", ",", "j", ",", "k"}], "]"}]}]}], ";"}], 
   "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"Vfragmemijk", "[", 
     RowBox[{"i_", ",", "j_", ",", "k_"}], "]"}], ":=", 
    RowBox[{
     RowBox[{"-", "kfragmem"}], "*", 
     RowBox[{"fragmemgamma", "[", 
      RowBox[{"i", ",", "j"}], "]"}], "*", 
     RowBox[{"Exp", "[", 
      RowBox[{
       RowBox[{"-", 
        RowBox[{
         RowBox[{"dr", "[", 
          RowBox[{"i", ",", "j", ",", "k"}], "]"}], "^", "2"}]}], "/", 
       RowBox[{"(", 
        RowBox[{"2", "*", 
         RowBox[{
          RowBox[{"fragmemsigma", "[", 
           RowBox[{"i", ",", "j"}], "]"}], "^", "2"}]}], ")"}]}], "]"}]}]}], 
   "\[IndentingNewLine]", 
   RowBox[{"Vfragmem", ":=", 
    RowBox[{
     UnderoverscriptBox["\[Sum]", 
      RowBox[{"i", "=", "1"}], "M"], 
     RowBox[{
      UnderoverscriptBox["\[Sum]", 
       RowBox[{"j", "=", 
        RowBox[{"i", "+", "1"}]}], "M"], 
      RowBox[{
       UnderoverscriptBox["\[Sum]", 
        RowBox[{"k", "=", "1"}], "4"], 
       RowBox[{"Vfragmemijk", "[", 
        RowBox[{"i", ",", "j", ",", "k"}], "]"}]}]}]}]}]}]}]], "Input",
 CellChangeTimes->{{3.513714924511195*^9, 3.513714965991281*^9}, {
   3.5137150793283043`*^9, 3.513715095838957*^9}, {3.513715212859891*^9, 
   3.513715367882272*^9}, 3.513715480853549*^9, {3.513715528958846*^9, 
   3.513715632928996*^9}, {3.5139685766891813`*^9, 3.513968603968403*^9}, {
   3.513968723743802*^9, 3.513968738860376*^9}, {3.51396884216463*^9, 
   3.513968864175507*^9}, {3.513968904261001*^9, 3.513968916238957*^9}, {
   3.513968952937399*^9, 3.5139689684153337`*^9}, {3.513969073721258*^9, 
   3.513969080229*^9}, {3.513970539171433*^9, 3.513970565893683*^9}, {
   3.513970606161374*^9, 3.5139706229784613`*^9}, {3.5139706546527863`*^9, 
   3.5139706966260157`*^9}, {3.5139707367866707`*^9, 3.513970888939725*^9}, {
   3.5139740882202187`*^9, 3.5139741040935163`*^9}, 3.513975014534733*^9, {
   3.513975046003183*^9, 3.51397506107309*^9}, {3.513975108339973*^9, 
   3.5139751084677343`*^9}, {3.5139755130277042`*^9, 
   3.5139755131049347`*^9}, {3.513975548384933*^9, 3.513975548522696*^9}, {
   3.514048239951634*^9, 3.514048271546607*^9}, {3.5140484773671303`*^9, 
   3.514048503127616*^9}, {3.514048773272977*^9, 3.514048807126499*^9}, {
   3.514049291025495*^9, 3.5140493482434683`*^9}, {3.5140511570277557`*^9, 
   3.514051159699375*^9}, {3.514051211687931*^9, 3.51405121302848*^9}, {
   3.514051561741046*^9, 3.5140515843788977`*^9}, {3.51405703644624*^9, 
   3.514057050572613*^9}, {3.514057138097354*^9, 3.514057138566289*^9}}],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
   "THIS", " ", "BLOCK", " ", "INITIALIZES", " ", "THE", " ", "VARIABLE", " ",
     "VTotal", " ", "WHICH", " ", "IS", " ", "USED", " ", "IN", " ", "THE", 
    " ", "SUBSEQUENT", " ", "BLOCK", " ", "THAT", " ", "EVALUATES", " ", 
    "THE", " ", "FORCES"}], " ", "*)"}], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{"VTotal", " ", "=", 
    RowBox[{"2", " ", "Vamhgo"}]}], ";", " ", 
   RowBox[{"(*", " ", 
    RowBox[{
     RowBox[{
     "THIS", " ", "ASSIGNMENT", " ", "CARRIES", " ", "WITH", " ", "IT", " ", 
      "ALL", " ", "THE", " ", "DETAILS", " ", "OF", " ", "THE", " ", 
      "FUNCTION"}], ",", " ", 
     RowBox[{
      RowBox[{"I", ".", "E", ".", " ", "IT"}], " ", "IS", " ", "NOT", " ", 
      "JUST", " ", "ASSIGNING", " ", "A", " ", "NUMBER", " ", "TO", " ", 
      RowBox[{"VTotal", ".", " ", "IN"}], " ", "THIS", " ", "WAY"}], ",", " ", 
     RowBox[{
     "IT", " ", "CAN", " ", "BE", " ", "USED", " ", "BELOW", " ", "AS", " ", 
      "INPUT", " ", "FOR", " ", "CALCULATING", " ", "THE", " ", "FORCES", " ",
       "AND", " ", "THAT", " ", "BLOCK", " ", 
      RowBox[{"DOESN", "'"}], "T", " ", "HAVE", " ", "TO", " ", "BE", " ", 
      "CHANGED", " ", "WHEN", " ", "CHANGING", " ", "THE", " ", "POTENTIAL", 
      " ", "THAT", " ", "YOU", " ", "ARE", " ", "TESTING"}]}], " ", "*)"}], 
   " "}]}]], "Input",
 CellChangeTimes->{{3.4706685671253777`*^9, 3.470668568923956*^9}, 
   3.470669140687944*^9, {3.472568135764303*^9, 3.472568136077662*^9}, {
   3.49153042070728*^9, 3.49153043000228*^9}, 3.5062799804243402`*^9, {
   3.5117206113578653`*^9, 3.5117206143005743`*^9}, {3.511720931183508*^9, 
   3.511720945895186*^9}, {3.511721091663517*^9, 3.511721093624007*^9}, 
   3.511730339573325*^9, 3.5118034247967377`*^9, {3.5123310006221437`*^9, 
   3.512331009104251*^9}, {3.512331066833764*^9, 3.51233109174583*^9}, {
   3.5123336302879763`*^9, 3.512333743640379*^9}, {3.5123432922590218`*^9, 
   3.512343294794918*^9}, {3.512345020526559*^9, 3.512345042630878*^9}, {
   3.5129251139698277`*^9, 3.512925116279237*^9}, {3.51354544579733*^9, 
   3.513545470430064*^9}, {3.513715485994145*^9, 3.513715487528349*^9}, {
   3.513972435683196*^9, 3.513972438121892*^9}, {3.513972746780676*^9, 
   3.513972748574991*^9}, {3.513972865034998*^9, 3.5139728681854753`*^9}, {
   3.5139750692766123`*^9, 3.5139750711381807`*^9}, {3.513976905758066*^9, 
   3.513976907731024*^9}, {3.514048333178842*^9, 3.514048334856587*^9}, {
   3.514057388588689*^9, 3.514057391146385*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{"THIS", " ", "BLOCK", " ", "USES", " ", 
    RowBox[{"MATHEMATICA", "'"}], "S", " ", "CAPABILITIES", " ", "TO", " ", 
    "ANALYTICALLY", " ", "DIFFERENTIATE", " ", "THE", " ", "POTENTIAL", " ", 
    "FUNCTION", " ", "TO", " ", "GIVE", " ", "FORCES"}], " ", "*)"}], 
  "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{"ProgressIndicator", "[", 
    RowBox[{
     RowBox[{"Dynamic", "[", "ipr", "]"}], ",", 
     RowBox[{"{", 
      RowBox[{"0", ",", "M"}], "}"}]}], "]"}], "\[IndentingNewLine]", 
   RowBox[{"FCA", "=", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"V0", ",", "V"}], "}"}], ",", 
      RowBox[{
       RowBox[{"ipr", "=", "0"}], ";", 
       RowBox[{"V0", "=", 
        RowBox[{
         RowBox[{"VTotal", "/.", 
          RowBox[{"Ruless", "[", "2", "]"}]}], "/.", 
         RowBox[{"Ruless", "[", "3", "]"}]}]}], ";", 
       RowBox[{"Table", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"ipr", "=", "i"}], ";", 
          RowBox[{"V", "=", 
           RowBox[{"V0", "/.", 
            RowBox[{"Rules", "[", 
             RowBox[{"1", ",", "i"}], "]"}]}]}], ";", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              RowBox[{"D", "[", 
               RowBox[{"V", ",", " ", 
                SubscriptBox["x", 
                 RowBox[{"Ca", ",", "i"}]]}], "]"}], "/.", 
              RowBox[{"Rulee", "[", 
               RowBox[{"1", ",", "i"}], "]"}]}], ")"}], ",", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"D", "[", 
               RowBox[{"V", ",", " ", 
                SubscriptBox["y", 
                 RowBox[{"Ca", ",", "i"}]]}], "]"}], "/.", 
              RowBox[{"Rulee", "[", 
               RowBox[{"1", ",", "i"}], "]"}]}], ")"}], ",", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"D", "[", 
               RowBox[{"V", ",", " ", 
                SubscriptBox["z", 
                 RowBox[{"Ca", ",", "i"}]]}], "]"}], "/.", 
              RowBox[{"Rulee", "[", 
               RowBox[{"1", ",", "i"}], "]"}]}], ")"}]}], "}"}]}], ",", 
         RowBox[{"{", 
          RowBox[{"i", ",", "1", ",", "M"}], "}"}]}], "]"}]}]}], "]"}]}], 
   "\[IndentingNewLine]", 
   RowBox[{"FCB", "=", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"V0", ",", "V"}], "}"}], ",", 
      RowBox[{
       RowBox[{"ipr", "=", "0"}], ";", 
       RowBox[{"V0", "=", 
        RowBox[{
         RowBox[{"VTotal", "/.", 
          RowBox[{"Ruless", "[", "1", "]"}]}], "/.", 
         RowBox[{"Ruless", "[", "3", "]"}]}]}], ";", 
       RowBox[{"Table", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"ipr", "=", "i"}], ";", 
          RowBox[{"V", "=", 
           RowBox[{"V0", "/.", 
            RowBox[{"Rules", "[", 
             RowBox[{"2", ",", "i"}], "]"}]}]}], ";", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              RowBox[{"D", "[", 
               RowBox[{"V", ",", " ", 
                SubscriptBox["x", 
                 RowBox[{"Cb", ",", "i"}]]}], "]"}], "/.", 
              RowBox[{"Rulee", "[", 
               RowBox[{"2", ",", "i"}], "]"}]}], ")"}], ",", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"D", "[", 
               RowBox[{"V", ",", " ", 
                SubscriptBox["y", 
                 RowBox[{"Cb", ",", "i"}]]}], "]"}], "/.", 
              RowBox[{"Rulee", "[", 
               RowBox[{"2", ",", "i"}], "]"}]}], ")"}], ",", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"D", "[", 
               RowBox[{"V", ",", " ", 
                SubscriptBox["z", 
                 RowBox[{"Cb", ",", "i"}]]}], "]"}], "/.", 
              RowBox[{"Rulee", "[", 
               RowBox[{"2", ",", "i"}], "]"}]}], ")"}]}], "}"}]}], ",", 
         RowBox[{"{", 
          RowBox[{"i", ",", "1", ",", "M"}], "}"}]}], "]"}]}]}], "]"}]}], 
   "\[IndentingNewLine]", 
   RowBox[{"FO", "=", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{"V0", ",", "V"}], "}"}], ",", 
      RowBox[{
       RowBox[{"ipr", "=", "0"}], ";", 
       RowBox[{"V0", "=", 
        RowBox[{
         RowBox[{"VTotal", "/.", 
          RowBox[{"Ruless", "[", "1", "]"}]}], "/.", 
         RowBox[{"Ruless", "[", "2", "]"}]}]}], ";", 
       RowBox[{"Table", "[", 
        RowBox[{
         RowBox[{
          RowBox[{"ipr", "=", "i"}], ";", 
          RowBox[{"V", "=", 
           RowBox[{"V0", "/.", 
            RowBox[{"Rules", "[", 
             RowBox[{"3", ",", "i"}], "]"}]}]}], ";", 
          RowBox[{"{", 
           RowBox[{
            RowBox[{"(", 
             RowBox[{
              RowBox[{"D", "[", 
               RowBox[{"V", ",", " ", 
                SubscriptBox["x", 
                 RowBox[{"O", ",", "i"}]]}], "]"}], "/.", 
              RowBox[{"Rulee", "[", 
               RowBox[{"3", ",", "i"}], "]"}]}], ")"}], ",", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"D", "[", 
               RowBox[{"V", ",", " ", 
                SubscriptBox["y", 
                 RowBox[{"O", ",", "i"}]]}], "]"}], "/.", 
              RowBox[{"Rulee", "[", 
               RowBox[{"3", ",", "i"}], "]"}]}], ")"}], ",", 
            RowBox[{"(", 
             RowBox[{
              RowBox[{"D", "[", 
               RowBox[{"V", ",", " ", 
                SubscriptBox["z", 
                 RowBox[{"O", ",", "i"}]]}], "]"}], "/.", 
              RowBox[{"Rulee", "[", 
               RowBox[{"3", ",", "i"}], "]"}]}], ")"}]}], "}"}]}], ",", 
         RowBox[{"{", 
          RowBox[{"i", ",", "1", ",", "M"}], "}"}]}], "]"}]}]}], 
     "]"}]}]}]}]], "Input",
 CellChangeTimes->{{3.471701217544476*^9, 3.471701246546933*^9}, {
   3.472406747727654*^9, 3.47240675002771*^9}, {3.472574328849304*^9, 
   3.472574329475623*^9}, 3.4725743767652073`*^9, {3.472574432137315*^9, 
   3.472574434319613*^9}, {3.472574739790389*^9, 3.472574750794717*^9}, {
   3.472574799687457*^9, 3.472574800128894*^9}, {3.47257485955485*^9, 
   3.4725748713780193`*^9}, {3.512331096871563*^9, 3.51233115208499*^9}, {
   3.512349578871891*^9, 3.512349580132258*^9}, {3.512353658151764*^9, 
   3.512353659727635*^9}}],

Cell[BoxData[
 ProgressIndicatorBox[Dynamic[$CellContext`ipr], {0, 26}]], "Output",
 CellChangeTimes->{
  3.512926645175634*^9, 3.512927400451806*^9, 3.5130191786048603`*^9, 
   3.513542825623665*^9, 3.5135454491013002`*^9, 3.513709776317152*^9, 
   3.513968632695541*^9, 3.5139687469142733`*^9, 3.5139688710009604`*^9, 
   3.51396892261096*^9, 3.513969087602244*^9, {3.5139705529159327`*^9, 
   3.513970570827422*^9}, 3.5139706297644987`*^9, {3.513970679567185*^9, 
   3.513970702245462*^9}, 3.513970766739275*^9, 3.513970817972329*^9, 
   3.513970862743157*^9, 3.513970897761904*^9, 3.51397113787652*^9, 
   3.513971549609227*^9, 3.513972440940618*^9, 3.5139724939300117`*^9, 
   3.51397253344867*^9, 3.513972751002678*^9, 3.513972871073107*^9, 
   3.513973861791725*^9, 3.5139738939642773`*^9, {3.513973946589707*^9, 
   3.513973970046949*^9}, 3.513974005575858*^9, 3.513974040724448*^9, 
   3.51397410866325*^9, 3.5139742083006268`*^9, 3.513975073889505*^9, 
   3.5139755214141006`*^9, 3.5139755535858793`*^9, 3.513976860397132*^9, 
   3.513976911491637*^9, 3.514048517737574*^9, 3.514048783186898*^9, 
   3.514048815235408*^9, 3.514051165479944*^9, 3.514051218238743*^9, {
   3.51405156927391*^9, 3.514051590336379*^9}, 3.514057078809155*^9, 
   3.514057393723371*^9, 3.5140581219290323`*^9, 3.514058612852304*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.004391762673812387`", ",", 
     RowBox[{"-", "0.0022703898833777627`"}], ",", "0.0037007534868690305`"}],
     "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.01013726991904716`"}], ",", "0.004277904870035948`", ",", 
     RowBox[{"-", "0.009699413149563563`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"2.577850185177091`", ",", "5.192234043112508`", ",", 
     RowBox[{"-", "2.864870600007472`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"6.509862862981013`", ",", "13.250891891662906`", ",", 
     RowBox[{"-", "7.037045805553218`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.13471653419569588`", ",", 
     RowBox[{"-", "8.977994942132462`"}], ",", 
     RowBox[{"-", "1.9595894881450042`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"7.995505832033409`", ",", 
     RowBox[{"-", "11.045866131509811`"}], ",", 
     RowBox[{"-", "5.0698677182183545`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "3.4539338886292676`"}], ",", 
     RowBox[{"-", "12.418700245891202`"}], ",", 
     RowBox[{"-", "1.9688179093467988`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"8.723930054366738`", ",", "10.675909713940595`", ",", 
     RowBox[{"-", "21.419696865413986`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.5510067287115433`"}], ",", 
     RowBox[{"-", "0.17989525542140758`"}], ",", "2.2849645350622727`"}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.32116046872879606`"}], ",", 
     RowBox[{"-", "0.6571543712115583`"}], ",", "0.2169730992806818`"}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.0012071187120477034`"}], ",", 
     RowBox[{"-", "0.0027568899312588237`"}], ",", "0.07157683182762929`"}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.237839265168653`"}], ",", "2.1137349319673837`", ",", 
     RowBox[{"-", "0.5911869297319583`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.24300231081440338`", ",", "0.6083898850096154`", ",", 
     RowBox[{"-", "0.5063901200002342`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.013428067428998264`"}], ",", "1.1147767178748824`", ",", 
     
     RowBox[{"-", "1.6071148176742092`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "2.408193306634816`", ",", "2.5706174108777944`", ",", 
     "6.192656702918571`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.8965267684796421`", ",", 
     RowBox[{"-", "1.3720343058796343`"}], ",", 
     RowBox[{"-", "0.0661771920074039`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.38187039499190534`"}], ",", 
     RowBox[{"-", "5.372770315400479`"}], ",", "6.134779573671662`"}], "}"}], 
   ",", 
   RowBox[{"{", 
    RowBox[{
    "1.5004523153587617`", ",", "10.594528793727173`", ",", 
     "10.428682866821335`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "2.4735497025735644`"}], ",", "6.573327068998836`", ",", 
     "5.284077837140413`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.37602443955282344`"}], ",", "0.6026058579924872`", ",", 
     "1.2207840909021312`"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{
  3.512926645175634*^9, 3.512927400451806*^9, 3.5130191786048603`*^9, 
   3.513542825623665*^9, 3.5135454491013002`*^9, 3.513709776317152*^9, 
   3.513968632695541*^9, 3.5139687469142733`*^9, 3.5139688710009604`*^9, 
   3.51396892261096*^9, 3.513969087602244*^9, {3.5139705529159327`*^9, 
   3.513970570827422*^9}, 3.5139706297644987`*^9, {3.513970679567185*^9, 
   3.513970702245462*^9}, 3.513970766739275*^9, 3.513970817972329*^9, 
   3.513970862743157*^9, 3.513970897761904*^9, 3.51397113787652*^9, 
   3.513971549609227*^9, 3.513972440940618*^9, 3.5139724939300117`*^9, 
   3.51397253344867*^9, 3.513972751002678*^9, 3.513972871073107*^9, 
   3.513973861791725*^9, 3.5139738939642773`*^9, {3.513973946589707*^9, 
   3.513973970046949*^9}, 3.513974005575858*^9, 3.513974040724448*^9, 
   3.51397410866325*^9, 3.5139742083006268`*^9, 3.513975073889505*^9, 
   3.5139755214141006`*^9, 3.5139755535858793`*^9, 3.513976860397132*^9, 
   3.513976911491637*^9, 3.514048517737574*^9, 3.514048783186898*^9, 
   3.514048815235408*^9, 3.514051165479944*^9, 3.514051218238743*^9, {
   3.51405156927391*^9, 3.514051590336379*^9}, 3.514057078809155*^9, 
   3.514057393723371*^9, 3.5140581219290323`*^9, 3.514058613414693*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.005745507245234772`", ",", 
     RowBox[{"-", "0.002007514986658185`"}], ",", "0.005998659662694531`"}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.41931638804793947`", ",", 
     RowBox[{"-", "0.37481963391676015`"}], ",", 
     RowBox[{"-", "0.11023040174737865`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"2.3302431198453792`", ",", "3.2632249860340052`", ",", 
     RowBox[{"-", "2.357978375434541`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.0838527156540494`"}], ",", "0.7415828324140846`", ",", 
     "0.3551220071854123`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "9.328429202460159`"}], ",", 
     RowBox[{"-", "3.009240262650126`"}], ",", "2.592041689553665`"}], "}"}], 
   ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "2.420316169782868`"}], ",", 
     RowBox[{"-", "3.3602790835864393`"}], ",", 
     RowBox[{"-", "1.8967097312995087`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.4163556537109051`", ",", 
     RowBox[{"-", "1.700449351993746`"}], ",", "0.14481795773487774`"}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.0006279763437249586`", ",", "0.0009155639585823265`", ",", 
     RowBox[{"-", "0.005608420648313366`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.6978128619378736`", ",", 
     RowBox[{"-", "2.8848586447200364`"}], ",", "4.00051794881204`"}], "}"}], 
   ",", 
   RowBox[{"{", 
    RowBox[{
    "2.613496614234736`", ",", "3.343604470760576`", ",", 
     "0.9240364203721116`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "1.0650787438187759`"}], ",", 
     RowBox[{"-", "1.6293147632558924`"}], ",", 
     RowBox[{"-", "2.4686299160541822`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "5.602024019352909`"}], ",", 
     RowBox[{"-", "6.9407274404573585`"}], ",", "4.6154693443285755`"}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "10.264529339898177`"}], ",", 
     RowBox[{"-", "0.6560064715404849`"}], ",", "3.9710514468360936`"}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.609593865713735`"}], ",", "3.9633464165932097`", ",", 
     "4.162380465763881`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.2840486529838561`"}], ",", 
     RowBox[{"-", "4.026822475425993`"}], ",", 
     RowBox[{"-", "2.6700185269288026`"}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{
  3.512926645175634*^9, 3.512927400451806*^9, 3.5130191786048603`*^9, 
   3.513542825623665*^9, 3.5135454491013002`*^9, 3.513709776317152*^9, 
   3.513968632695541*^9, 3.5139687469142733`*^9, 3.5139688710009604`*^9, 
   3.51396892261096*^9, 3.513969087602244*^9, {3.5139705529159327`*^9, 
   3.513970570827422*^9}, 3.5139706297644987`*^9, {3.513970679567185*^9, 
   3.513970702245462*^9}, 3.513970766739275*^9, 3.513970817972329*^9, 
   3.513970862743157*^9, 3.513970897761904*^9, 3.51397113787652*^9, 
   3.513971549609227*^9, 3.513972440940618*^9, 3.5139724939300117`*^9, 
   3.51397253344867*^9, 3.513972751002678*^9, 3.513972871073107*^9, 
   3.513973861791725*^9, 3.5139738939642773`*^9, {3.513973946589707*^9, 
   3.513973970046949*^9}, 3.513974005575858*^9, 3.513974040724448*^9, 
   3.51397410866325*^9, 3.5139742083006268`*^9, 3.513975073889505*^9, 
   3.5139755214141006`*^9, 3.5139755535858793`*^9, 3.513976860397132*^9, 
   3.513976911491637*^9, 3.514048517737574*^9, 3.514048783186898*^9, 
   3.514048815235408*^9, 3.514051165479944*^9, 3.514051218238743*^9, {
   3.51405156927391*^9, 3.514051590336379*^9}, 3.514057078809155*^9, 
   3.514057393723371*^9, 3.5140581219290323`*^9, 3.5140586139550133`*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{
  3.512926645175634*^9, 3.512927400451806*^9, 3.5130191786048603`*^9, 
   3.513542825623665*^9, 3.5135454491013002`*^9, 3.513709776317152*^9, 
   3.513968632695541*^9, 3.5139687469142733`*^9, 3.5139688710009604`*^9, 
   3.51396892261096*^9, 3.513969087602244*^9, {3.5139705529159327`*^9, 
   3.513970570827422*^9}, 3.5139706297644987`*^9, {3.513970679567185*^9, 
   3.513970702245462*^9}, 3.513970766739275*^9, 3.513970817972329*^9, 
   3.513970862743157*^9, 3.513970897761904*^9, 3.51397113787652*^9, 
   3.513971549609227*^9, 3.513972440940618*^9, 3.5139724939300117`*^9, 
   3.51397253344867*^9, 3.513972751002678*^9, 3.513972871073107*^9, 
   3.513973861791725*^9, 3.5139738939642773`*^9, {3.513973946589707*^9, 
   3.513973970046949*^9}, 3.513974005575858*^9, 3.513974040724448*^9, 
   3.51397410866325*^9, 3.5139742083006268`*^9, 3.513975073889505*^9, 
   3.5139755214141006`*^9, 3.5139755535858793`*^9, 3.513976860397132*^9, 
   3.513976911491637*^9, 3.514048517737574*^9, 3.514048783186898*^9, 
   3.514048815235408*^9, 3.514051165479944*^9, 3.514051218238743*^9, {
   3.51405156927391*^9, 3.514051590336379*^9}, 3.514057078809155*^9, 
   3.514057393723371*^9, 3.5140581219290323`*^9, 3.5140586140771923`*^9}]
}, Open  ]],

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
    RowBox[{
    "THIS", " ", "CALCULATES", " ", "FORCES", " ", "FROM", " ", "POTENTIALS", 
     " ", "USING", " ", "A", " ", "NUMERICAL", " ", "DIFFERENTIATION", " ", 
     "SCHEME"}], " ", "-", " ", 
    RowBox[{
    "IT", " ", "IS", " ", "NOT", " ", "THE", " ", "RECOMMENDED", " ", 
     "METHOD"}]}], " ", "*)"}], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{"h", "=", "0.0001"}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"rdh", "[", 
     RowBox[{"r_", ",", "h_", ",", "i_", ",", "j_"}], "]"}], ":=", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", "rr", "}"}], ",", 
      RowBox[{
       RowBox[{"rr", "=", "r"}], ";", 
       RowBox[{
        RowBox[{"rr", "[", 
         RowBox[{"[", 
          RowBox[{"i", ",", "j"}], "]"}], "]"}], "+=", "h"}], ";", 
       RowBox[{"Return", "[", "rr", "]"}]}]}], "]"}]}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"i", "=", "5"}], ";", 
    RowBox[{"j", "=", "2"}], ";"}], "\[IndentingNewLine]", 
   RowBox[{
    RowBox[{"Vwaterhij", "[", 
     RowBox[{"h_", ",", "ii_", ",", "jj_", ",", "kk_"}], "]"}], ":=", 
    RowBox[{"Module", "[", 
     RowBox[{
      RowBox[{"{", 
       RowBox[{
        RowBox[{"rcah", "=", "rca"}], ",", 
        RowBox[{"rcbh", "=", "rcb"}], ",", 
        RowBox[{"roh", "=", "ro"}], ",", "ruleh"}], "}"}], ",", 
      RowBox[{
       RowBox[{"If", "[", 
        RowBox[{
         RowBox[{"kk", "\[Equal]", "1"}], ",", 
         RowBox[{"rcah", "=", 
          RowBox[{"rdh", "[", 
           RowBox[{"rca", ",", "h", ",", "ii", ",", "jj"}], "]"}]}], ",", 
         RowBox[{"If", "[", 
          RowBox[{
           RowBox[{"kk", "\[Equal]", "2"}], ",", 
           RowBox[{"rcbh", "=", 
            RowBox[{"rdh", "[", 
             RowBox[{"rcb", ",", "h", ",", "ii", ",", "jj"}], "]"}]}], ",", 
           RowBox[{"roh", "=", 
            RowBox[{"rdh", "[", 
             RowBox[{"ro", ",", "h", ",", "ii", ",", "jj"}], "]"}]}]}], 
          "]"}]}], "]"}], ";", "\[IndentingNewLine]", 
       RowBox[{"ruleh", "=", 
        RowBox[{"Flatten", "[", 
         RowBox[{"Append", "[", 
          RowBox[{
           RowBox[{"Append", "[", 
            RowBox[{
             RowBox[{"Table", "[", 
              RowBox[{
               RowBox[{"{", 
                RowBox[{
                 RowBox[{
                  SubscriptBox["r", 
                   RowBox[{"Ca", ",", "i"}]], "[", 
                  RowBox[{"[", "j", "]"}], "]"}], "\[Rule]", 
                 RowBox[{"rcah", "[", 
                  RowBox[{"[", 
                   RowBox[{"i", ",", "j"}], "]"}], "]"}]}], "}"}], ",", 
               RowBox[{"{", 
                RowBox[{"i", ",", "1", ",", "M"}], "}"}], ",", 
               RowBox[{"{", 
                RowBox[{"j", ",", "1", ",", "3"}], "}"}]}], "]"}], ",", 
             RowBox[{"Table", "[", 
              RowBox[{
               RowBox[{"{", 
                RowBox[{
                 RowBox[{
                  SubscriptBox["r", 
                   RowBox[{"Cb", ",", "i"}]], "[", 
                  RowBox[{"[", "j", "]"}], "]"}], "\[Rule]", 
                 RowBox[{"rcbh", "[", 
                  RowBox[{"[", 
                   RowBox[{"i", ",", "j"}], "]"}], "]"}]}], "}"}], ",", 
               RowBox[{"{", 
                RowBox[{"i", ",", "1", ",", "M"}], "}"}], ",", 
               RowBox[{"{", 
                RowBox[{"j", ",", "1", ",", "3"}], "}"}]}], "]"}]}], "]"}], 
           ",", 
           RowBox[{"Table", "[", 
            RowBox[{
             RowBox[{"{", 
              RowBox[{
               RowBox[{
                SubscriptBox["r", 
                 RowBox[{"O", ",", "i"}]], "[", 
                RowBox[{"[", "j", "]"}], "]"}], "\[Rule]", 
               RowBox[{"roh", "[", 
                RowBox[{"[", 
                 RowBox[{"i", ",", "j"}], "]"}], "]"}]}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{"i", ",", "1", ",", "M"}], "}"}], ",", 
             RowBox[{"{", 
              RowBox[{"j", ",", "1", ",", "3"}], "}"}]}], "]"}]}], "]"}], 
         "]"}]}], ";", 
       RowBox[{"Return", "[", 
        RowBox[{"Vwater", "/.", "ruleh"}], "]"}]}]}], "]"}]}], 
   "\[IndentingNewLine]", 
   RowBox[{"ProgressIndicator", "[", 
    RowBox[{
     RowBox[{"Dynamic", "[", "kt", "]"}], ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", "3"}], "}"}]}], "]"}], "\[IndentingNewLine]", 
   RowBox[{"ProgressIndicator", "[", 
    RowBox[{
     RowBox[{"Dynamic", "[", "it", "]"}], ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", "M"}], "}"}]}], "]"}], "\[IndentingNewLine]", 
   RowBox[{"ProgressIndicator", "[", 
    RowBox[{
     RowBox[{"Dynamic", "[", "jt", "]"}], ",", 
     RowBox[{"{", 
      RowBox[{"1", ",", "3"}], "}"}]}], "]"}], "\[IndentingNewLine]", 
   RowBox[{"FF", "=", 
    RowBox[{"Table", "[", 
     RowBox[{
      RowBox[{"0.5", 
       RowBox[{
        RowBox[{"(", 
         RowBox[{
          RowBox[{"Vwaterhij", "[", 
           RowBox[{"h", ",", "it", ",", "jt", ",", "kt"}], "]"}], "-", 
          RowBox[{"Vwaterhij", "[", 
           RowBox[{
            RowBox[{"-", "h"}], ",", "it", ",", "jt", ",", "kt"}], "]"}]}], 
         ")"}], "/", "h"}]}], ",", 
      RowBox[{"{", 
       RowBox[{"kt", ",", "1", ",", "3"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"it", ",", "1", ",", "M"}], "}"}], ",", 
      RowBox[{"{", 
       RowBox[{"jt", ",", "1", ",", "3"}], "}"}]}], "]"}]}], 
   "\[IndentingNewLine]"}]}]], "Input",
 CellOpen->False,
 CellChangeTimes->{{3.5062785899733057`*^9, 3.506278831999267*^9}, {
  3.5062788718752813`*^9, 3.506279041983946*^9}, {3.5062791047284107`*^9, 
  3.5062793944440813`*^9}, {3.50627942466101*^9, 3.50627947534096*^9}, {
  3.50627950604035*^9, 3.5062795625999727`*^9}, {3.5062796037022867`*^9, 
  3.506279793908305*^9}, {3.50627982401138*^9, 3.5062799104062243`*^9}, {
  3.506280017732071*^9, 3.5062800212929363`*^9}, {3.5062800998437862`*^9, 
  3.506280271146153*^9}, {3.506282150400591*^9, 3.5062821577074633`*^9}, {
  3.506343452089678*^9, 3.506343459765299*^9}, {3.512323114881507*^9, 
  3.512323142960685*^9}}],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
   "TEST", " ", "IF", " ", "FORCES", " ", "CALCULATED", " ", "BY", " ", 
    "MATHEMATICA", " ", "AGREE", " ", "WITH", " ", "FORCES", " ", "INPUT", 
    " ", "FROM", " ", "LAMMPS"}], " ", "*)"}], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{"Max", "[", 
    RowBox[{"Abs", "[", 
     RowBox[{"Flatten", "[", 
      RowBox[{"FCA", "+", "fca"}], "]"}], "]"}], "]"}], " ", 
   RowBox[{"(*", " ", 
    RowBox[{
     RowBox[{
     "THE", " ", "REASON", " ", "WHY", " ", "IT", " ", "IS", " ", "PLUS", " ",
       "INSTEAD", " ", "OF", " ", "MINUS", " ", "IS", " ", "BECAUSE", " ", 
      "WHAT", " ", "MATHEMATICA", " ", "CALCULATES", " ", "IS", " ", 
      "ACTUALLY"}], "..."}], " ", "*)"}], "\[IndentingNewLine]", 
   RowBox[{"Max", "[", 
    RowBox[{"Abs", "[", 
     RowBox[{"Flatten", "[", 
      RowBox[{"FCB", "+", "fcb"}], "]"}], "]"}], "]"}], " ", 
   RowBox[{"(*", " ", 
    RowBox[{
     RowBox[{
      RowBox[{"dV", "/", "dX"}], " ", "AND", " ", "THE", " ", "FORCE", " ", 
      "IS", " ", "ACTUALLY", " ", "F"}], "=", 
     RowBox[{
      RowBox[{"-", "dV"}], "/", "dX"}]}], " ", "*)"}], "\[IndentingNewLine]", 
   
   RowBox[{"Max", "[", 
    RowBox[{"Abs", "[", 
     RowBox[{"Flatten", "[", 
      RowBox[{"FO", "+", "fo"}], "]"}], "]"}], "]"}]}]}]], "Input",
 CellChangeTimes->{{3.5117206881029053`*^9, 3.5117206995741243`*^9}, {
  3.512316013182979*^9, 3.512316014174736*^9}, {3.512331686034169*^9, 
  3.512331742949005*^9}}],

Cell[BoxData["0.00014068493772745327`"], "Output",
 CellChangeTimes->{
  3.513971021261215*^9, 3.5139711467375174`*^9, 3.513971572881699*^9, 
   3.513972755892272*^9, 3.513972875279571*^9, 3.5139751713521*^9, 
   3.513975528177947*^9, 3.513975558264139*^9, 3.513976864549876*^9, 
   3.513976974058248*^9, 3.514048524535432*^9, 3.514048688393271*^9, 
   3.514048789227511*^9, 3.514048823971246*^9, 3.514051170741516*^9, 
   3.514051222938801*^9, {3.5140515754132233`*^9, 3.514051593805303*^9}, 
   3.51405708345573*^9, 3.514057397834463*^9, 3.5140581259411182`*^9, 
   3.514058616449513*^9}],

Cell[BoxData["0.00011287044633556675`"], "Output",
 CellChangeTimes->{
  3.513971021261215*^9, 3.5139711467375174`*^9, 3.513971572881699*^9, 
   3.513972755892272*^9, 3.513972875279571*^9, 3.5139751713521*^9, 
   3.513975528177947*^9, 3.513975558264139*^9, 3.513976864549876*^9, 
   3.513976974058248*^9, 3.514048524535432*^9, 3.514048688393271*^9, 
   3.514048789227511*^9, 3.514048823971246*^9, 3.514051170741516*^9, 
   3.514051222938801*^9, {3.5140515754132233`*^9, 3.514051593805303*^9}, 
   3.51405708345573*^9, 3.514057397834463*^9, 3.5140581259411182`*^9, 
   3.514058616450798*^9}],

Cell[BoxData["0.`"], "Output",
 CellChangeTimes->{
  3.513971021261215*^9, 3.5139711467375174`*^9, 3.513971572881699*^9, 
   3.513972755892272*^9, 3.513972875279571*^9, 3.5139751713521*^9, 
   3.513975528177947*^9, 3.513975558264139*^9, 3.513976864549876*^9, 
   3.513976974058248*^9, 3.514048524535432*^9, 3.514048688393271*^9, 
   3.514048789227511*^9, 3.514048823971246*^9, 3.514051170741516*^9, 
   3.514051222938801*^9, {3.5140515754132233`*^9, 3.514051593805303*^9}, 
   3.51405708345573*^9, 3.514057397834463*^9, 3.5140581259411182`*^9, 
   3.514058616451825*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
   "THIS", " ", "IS", " ", "AN", " ", "EXAMPLE", " ", "OF", " ", "HOW", " ", 
    "TO", " ", "GET", " ", "THE", " ", "NUMERICAL", " ", "TOTAL", " ", "FOR", 
    " ", "A", " ", "PARTICULAR", " ", "POTENTIAL"}], " ", "*)"}], 
  "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{
    RowBox[{"2", " ", "*", "Vamhgo"}], "/.", "rule"}], "/.", "rule", " ", 
   RowBox[{"(*", " ", 
    RowBox[{
    "THE", " ", "2", " ", "HERE", " ", "IS", " ", "SET", " ", "TO", " ", 
     "MATCH", " ", "THE", " ", "\"\<[EPSILON]\>\"", " ", "VALUE", " ", "IN", 
     " ", "LAMMPS"}], " ", "*)"}]}]}]], "Input",
 CellChangeTimes->{{3.5123160223112993`*^9, 3.512316072711299*^9}, {
  3.512333763072858*^9, 3.512333784509334*^9}, {3.5123434025408363`*^9, 
  3.5123434035251617`*^9}, {3.5123449229128637`*^9, 3.5123449278072243`*^9}, {
  3.512351074117487*^9, 3.5123510741929903`*^9}, {3.512351327148741*^9, 
  3.512351328312166*^9}, {3.5123513596639147`*^9, 3.512351361284752*^9}, {
  3.512351488344685*^9, 3.5123514919008904`*^9}, {3.512354590125286*^9, 
  3.512354590480981*^9}, {3.512761564094194*^9, 3.512761564951641*^9}, {
  3.512925129763092*^9, 3.5129251309705553`*^9}, {3.5137154933215*^9, 
  3.5137154953931313`*^9}, {3.5139724281774063`*^9, 3.513972429919568*^9}, {
  3.513972657169777*^9, 3.513972658971281*^9}, {3.513972886624353*^9, 
  3.513972888583193*^9}, {3.513975080322307*^9, 3.513975081856345*^9}, {
  3.514057410244487*^9, 3.5140574126436367`*^9}}],

Cell[BoxData[
 RowBox[{"-", "996.431695606642`"}]], "Output",
 CellChangeTimes->{
  3.512925131398477*^9, 3.512926207466957*^9, 3.512926336451729*^9, {
   3.512926653911696*^9, 3.512926664559507*^9}, 3.512926766282247*^9, 
   3.51292763648991*^9, 3.512927991363587*^9, {3.5129329276701508`*^9, 
   3.512932949339422*^9}, 3.5129330050909777`*^9, {3.512933194232483*^9, 
   3.512933218672188*^9}, 3.512933372836646*^9, 3.51293908115009*^9, 
   3.512939803532214*^9, 3.5129398642066393`*^9, 3.5129402836570873`*^9, 
   3.512940485740037*^9, {3.512941068543447*^9, 3.512941092716151*^9}, 
   3.512941563236416*^9, 3.512941872775342*^9, 3.512941904739791*^9, 
   3.512942293286613*^9, 3.512947455326808*^9, {3.512947640156352*^9, 
   3.5129476591929502`*^9}, 3.51294794064501*^9, 3.512948038942526*^9, 
   3.512948175687625*^9, 3.513019082340271*^9, 3.5130191513790483`*^9, 
   3.513023283438846*^9, 3.5133689932212753`*^9, 3.51353167358022*^9, 
   3.513531798010551*^9, 3.5135318552837133`*^9, {3.513532294545507*^9, 
   3.513532327680807*^9}, 3.513540692024377*^9, 3.513541306593033*^9, 
   3.513541363949274*^9, 3.513541407469784*^9, 3.513541518825338*^9, 
   3.513541733755251*^9, 3.513542474546967*^9, 3.513543181129487*^9, 
   3.513709784448352*^9, 3.513709816842909*^9, 3.513715497112624*^9, {
   3.513715624599185*^9, 3.513715637318431*^9}, 3.5139709062384453`*^9, 
   3.5139711529072247`*^9, 3.5139714621616173`*^9, 3.5139724304546432`*^9, 
   3.513972487443947*^9, {3.5139726345718317`*^9, 3.513972659622793*^9}, 
   3.513972889089961*^9, 3.513973378188882*^9, 3.513973809446363*^9, 
   3.5139742160676813`*^9, {3.5139743163099213`*^9, 3.513974331910974*^9}, 
   3.513975082325986*^9, {3.513975533807454*^9, 3.5139755627745953`*^9}, 
   3.514048530802775*^9, 3.514057105853405*^9, 3.5140574130105963`*^9, 
   3.514058138777413*^9, 3.5140586289200706`*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
   "SHOW", " ", "DIFFERENCES", " ", "IN", " ", "CALCULATED", " ", "VERSUS", 
    " ", "LAMMPS", " ", "FORCES"}], " ", "*)"}], "\[IndentingNewLine]", 
  RowBox[{"FCA", " ", 
   RowBox[{"(*", " ", "MATHEMATICA", " ", "*)"}], "\[IndentingNewLine]", 
   "fca", " ", 
   RowBox[{"(*", " ", "LAMMPS", " ", "*)"}], "\[IndentingNewLine]", 
   RowBox[{"FCA", "+", "fca"}], " ", 
   RowBox[{"(*", " ", "DIFFERENCE", " ", "*)"}], "\[IndentingNewLine]", "FCB",
    "\[IndentingNewLine]", "fcb", "\[IndentingNewLine]", 
   RowBox[{"FCB", "+", "fcb"}], "\[IndentingNewLine]", "FO", 
   "\[IndentingNewLine]", "fo", "\[IndentingNewLine]", 
   RowBox[{"FO", "+", "fo"}]}]}]], "Input",
 CellChangeTimes->{{3.512759229606512*^9, 3.512759267400701*^9}, {
  3.513972936875342*^9, 3.513972953183226*^9}, {3.5139754177718277`*^9, 
  3.513975442932308*^9}}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.004391762673812387`", ",", 
     RowBox[{"-", "0.0022703898833777627`"}], ",", "0.0037007534868690305`"}],
     "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.01013726991904716`"}], ",", "0.004277904870035948`", ",", 
     RowBox[{"-", "0.009699413149563563`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"2.577850185177091`", ",", "5.192234043112508`", ",", 
     RowBox[{"-", "2.864870600007472`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"6.509862862981013`", ",", "13.250891891662906`", ",", 
     RowBox[{"-", "7.037045805553218`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.13471653419569588`", ",", 
     RowBox[{"-", "8.977994942132462`"}], ",", 
     RowBox[{"-", "1.9595894881450042`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"7.995505832033409`", ",", 
     RowBox[{"-", "11.045866131509811`"}], ",", 
     RowBox[{"-", "5.0698677182183545`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "3.4539338886292676`"}], ",", 
     RowBox[{"-", "12.418700245891202`"}], ",", 
     RowBox[{"-", "1.9688179093467988`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"8.723930054366738`", ",", "10.675909713940595`", ",", 
     RowBox[{"-", "21.419696865413986`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.5510067287115433`"}], ",", 
     RowBox[{"-", "0.17989525542140758`"}], ",", "2.2849645350622727`"}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.32116046872879606`"}], ",", 
     RowBox[{"-", "0.6571543712115583`"}], ",", "0.2169730992806818`"}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.0012071187120477034`"}], ",", 
     RowBox[{"-", "0.0027568899312588237`"}], ",", "0.07157683182762929`"}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.237839265168653`"}], ",", "2.1137349319673837`", ",", 
     RowBox[{"-", "0.5911869297319583`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.24300231081440338`", ",", "0.6083898850096154`", ",", 
     RowBox[{"-", "0.5063901200002342`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.013428067428998264`"}], ",", "1.1147767178748824`", ",", 
     
     RowBox[{"-", "1.6071148176742092`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "2.408193306634816`", ",", "2.5706174108777944`", ",", 
     "6.192656702918571`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.8965267684796421`", ",", 
     RowBox[{"-", "1.3720343058796343`"}], ",", 
     RowBox[{"-", "0.0661771920074039`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.38187039499190534`"}], ",", 
     RowBox[{"-", "5.372770315400479`"}], ",", "6.134779573671662`"}], "}"}], 
   ",", 
   RowBox[{"{", 
    RowBox[{
    "1.5004523153587617`", ",", "10.594528793727173`", ",", 
     "10.428682866821335`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "2.4735497025735644`"}], ",", "6.573327068998836`", ",", 
     "5.284077837140413`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.37602443955282344`"}], ",", "0.6026058579924872`", ",", 
     "1.2207840909021312`"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.5139715992840033`*^9, 3.513972760739259*^9, 
  3.513972879589386*^9, 3.5139729538315487`*^9, 3.513975175655528*^9, 
  3.513976990334423*^9, 3.5140488769013453`*^9, 3.514057111026916*^9, 
  3.514058147068818*^9, 3.514058654389997*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.0043917499999999166`"}], ",", "0.0022703800000000163`", 
     ",", 
     RowBox[{"-", "0.003700739999999314`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.010137260000000481`", ",", 
     RowBox[{"-", "0.004277890000000006`"}], ",", "0.009699399999998803`"}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "2.5778629800000004`"}], ",", 
     RowBox[{"-", "5.19224063`"}], ",", "2.864875`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "6.5098648699999995`"}], ",", 
     RowBox[{"-", "13.250905040000001`"}], ",", "7.037036799999999`"}], "}"}],
    ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.13471487000000004`"}], ",", "8.97795579`", ",", 
     "1.95958233`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "7.99548636`"}], ",", "11.045834509999999`", ",", 
     "5.0698565900000006`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"3.45386853`", ",", "12.41870763`", ",", "1.96883423`"}], "}"}], 
   ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "8.72387165`"}], ",", 
     RowBox[{"-", "10.67585265`"}], ",", "21.4195954`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.55101764`", ",", "0.17991632000000024`", ",", 
     RowBox[{"-", "2.28510522`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.32116840000000035`", ",", "0.6571627399999995`", ",", 
     RowBox[{"-", "0.21696975999999957`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.0012071099999997337`", ",", "0.0027568799999997395`", ",", 
     RowBox[{"-", "0.07157681000000027`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.23783948999999993`", ",", 
     RowBox[{"-", "2.11373635`"}], ",", "0.591186519999999`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.24300610000000056`"}], ",", 
     RowBox[{"-", "0.60840023`"}], ",", "0.50639616`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.013428239999999647`", ",", 
     RowBox[{"-", "1.11479057`"}], ",", "1.607134799999999`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "2.4081865400000004`"}], ",", 
     RowBox[{"-", "2.57065401`"}], ",", 
     RowBox[{"-", "6.1925674299999995`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.89648745`"}], ",", "1.3720187099999999`", ",", 
     "0.06623366999999991`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.3818279099999984`", ",", "5.372786399999999`", ",", 
     RowBox[{"-", "6.1347912099999995`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "1.5005223499999998`"}], ",", 
     RowBox[{"-", "10.594514329999999`"}], ",", 
     RowBox[{"-", "10.428625579999999`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"2.4735495500000004`", ",", 
     RowBox[{"-", "6.573341780000001`"}], ",", 
     RowBox[{"-", "5.28411034`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.37602297999999834`", ",", 
     RowBox[{"-", "0.60259952`"}], ",", 
     RowBox[{"-", "1.2207751`"}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.5139715992840033`*^9, 3.513972760739259*^9, 
  3.513972879589386*^9, 3.5139729538315487`*^9, 3.513975175655528*^9, 
  3.513976990334423*^9, 3.5140488769013453`*^9, 3.514057111026916*^9, 
  3.514058147068818*^9, 3.514058654392581*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"1.2673812470852253`*^-8", ",", 
     RowBox[{"-", "9.883377746419675`*^-9"}], ",", 
     "1.3486869716514138`*^-8"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "9.919046679504606`*^-9"}], ",", "1.487003594129871`*^-8", 
     ",", 
     RowBox[{"-", "1.314956475997886`*^-8"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.000012794822909167891`"}], ",", 
     RowBox[{"-", "6.586887491444315`*^-6"}], ",", "4.399992528103525`*^-6"}],
     "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "2.007018986560638`*^-6"}], ",", 
     RowBox[{"-", "0.000013148337094648355`"}], ",", 
     RowBox[{"-", "9.005553218699447`*^-6"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"1.6641956958329818`*^-6", ",", 
     RowBox[{"-", "0.000039152132462305644`"}], ",", 
     RowBox[{"-", "7.158145004293459`*^-6"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.000019472033408796108`", ",", 
     RowBox[{"-", "0.0000316215098123962`"}], ",", 
     RowBox[{"-", "0.000011128218353917418`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.00006535862926781988`"}], ",", "7.384108798191846`*^-6", 
     ",", "0.000016320653201118418`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.0000584043667384293`", ",", "0.00005706394059501463`", ",", 
     RowBox[{"-", "0.00010146541398725617`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "0.000010911288456649793`", ",", "0.000021064578592661087`", ",", 
     RowBox[{"-", "0.00014068493772745327`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "7.931271204297818`*^-6", ",", "8.368788441237562`*^-6", ",", 
     "3.339280682246315`*^-6"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "8.712047969757636`*^-9"}], ",", 
     RowBox[{"-", "9.931259084203226`*^-9"}], ",", 
     "2.1827629023696282`*^-8"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"2.2483134692885542`*^-7", ",", 
     RowBox[{"-", "1.4180326162183121`*^-6"}], ",", 
     RowBox[{"-", "4.097319592499815`*^-7"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "3.7891855971805555`*^-6"}], ",", 
     RowBox[{"-", "0.000010344990384547614`"}], ",", 
     "6.039999765783399`*^-6"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"1.725710013830012`*^-7", ",", 
     RowBox[{"-", "0.000013852125117619352`"}], ",", 
     "0.000019982325789813515`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"6.766634815580375`*^-6", ",", 
     RowBox[{"-", "0.00003659912220577155`"}], ",", 
     "0.0000892729185713037`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.00003931847964211421`", ",", 
     RowBox[{"-", "0.000015595879634400234`"}], ",", 
     "0.00005647799259600461`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.00004248499190695476`"}], ",", 
     "0.000016084599519849974`", ",", 
     RowBox[{"-", "0.000011636328337338853`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.00007003464123811476`"}], ",", 
     "0.000014463727174174323`", ",", "0.000057286821336433036`"}], "}"}], 
   ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "1.5257356400510957`*^-7"}], ",", 
     RowBox[{"-", "0.000014711001164791071`"}], ",", 
     RowBox[{"-", "0.000032502859586891475`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "1.459552825100463`*^-6"}], ",", "6.337992487170929`*^-6", 
     ",", "8.990902131245093`*^-6"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.5139715992840033`*^9, 3.513972760739259*^9, 
  3.513972879589386*^9, 3.5139729538315487`*^9, 3.513975175655528*^9, 
  3.513976990334423*^9, 3.5140488769013453`*^9, 3.514057111026916*^9, 
  3.514058147068818*^9, 3.51405865439469*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.005745507245234772`", ",", 
     RowBox[{"-", "0.002007514986658185`"}], ",", "0.005998659662694531`"}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.41931638804793947`", ",", 
     RowBox[{"-", "0.37481963391676015`"}], ",", 
     RowBox[{"-", "0.11023040174737865`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"2.3302431198453792`", ",", "3.2632249860340052`", ",", 
     RowBox[{"-", "2.357978375434541`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.0838527156540494`"}], ",", "0.7415828324140846`", ",", 
     "0.3551220071854123`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "9.328429202460159`"}], ",", 
     RowBox[{"-", "3.009240262650126`"}], ",", "2.592041689553665`"}], "}"}], 
   ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "2.420316169782868`"}], ",", 
     RowBox[{"-", "3.3602790835864393`"}], ",", 
     RowBox[{"-", "1.8967097312995087`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.4163556537109051`", ",", 
     RowBox[{"-", "1.700449351993746`"}], ",", "0.14481795773487774`"}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.0006279763437249586`", ",", "0.0009155639585823265`", ",", 
     RowBox[{"-", "0.005608420648313366`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.6978128619378736`", ",", 
     RowBox[{"-", "2.8848586447200364`"}], ",", "4.00051794881204`"}], "}"}], 
   ",", 
   RowBox[{"{", 
    RowBox[{
    "2.613496614234736`", ",", "3.343604470760576`", ",", 
     "0.9240364203721116`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "1.0650787438187759`"}], ",", 
     RowBox[{"-", "1.6293147632558924`"}], ",", 
     RowBox[{"-", "2.4686299160541822`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "5.602024019352909`"}], ",", 
     RowBox[{"-", "6.9407274404573585`"}], ",", "4.6154693443285755`"}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "10.264529339898177`"}], ",", 
     RowBox[{"-", "0.6560064715404849`"}], ",", "3.9710514468360936`"}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.609593865713735`"}], ",", "3.9633464165932097`", ",", 
     "4.162380465763881`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.2840486529838561`"}], ",", 
     RowBox[{"-", "4.026822475425993`"}], ",", 
     RowBox[{"-", "2.6700185269288026`"}]}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.5139715992840033`*^9, 3.513972760739259*^9, 
  3.513972879589386*^9, 3.5139729538315487`*^9, 3.513975175655528*^9, 
  3.513976990334423*^9, 3.5140488769013453`*^9, 3.514057111026916*^9, 
  3.514058147068818*^9, 3.514058654398443*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.005745509999999676`"}], ",", "0.002007509999999435`", 
     ",", 
     RowBox[{"-", "0.005998659999999489`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.4193148200000003`"}], ",", "0.3748182299999999`", ",", 
     "0.11022999`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "2.33022418`"}], ",", 
     RowBox[{"-", "3.2632142899999996`"}], ",", "2.35798686`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.08385165000000017`", ",", 
     RowBox[{"-", "0.7415734399999999`"}], ",", 
     RowBox[{"-", "0.35511751`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"9.328536459999999`", ",", "3.009186830000001`", ",", 
     RowBox[{"-", "2.5921545600000004`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"2.4203200000000002`", ",", "3.36028344`", ",", "1.89671342`"}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.41635514`"}], ",", "1.70044773`", ",", 
     RowBox[{"-", "0.1448171199999999`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.0006279699999999222`"}], ",", 
     RowBox[{"-", "0.0009155500000002093`"}], ",", "0.00560839000000013`"}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.69781439`"}], ",", "2.884864949999999`", ",", 
     RowBox[{"-", "4.00052669`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "2.61350061`"}], ",", 
     RowBox[{"-", "3.34360312`"}], ",", 
     RowBox[{"-", "0.92394948`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"1.06505765`", ",", "1.6293330100000003`", ",", "2.46862414`"}], 
    "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"5.602034359999999`", ",", "6.940738469999999`", ",", 
     RowBox[{"-", "4.6154889599999995`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"10.26454327`", ",", "0.6560187299999996`", ",", 
     RowBox[{"-", "3.9710475099999996`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.6095241100000006`", ",", 
     RowBox[{"-", "3.96331845`"}], ",", 
     RowBox[{"-", "4.162299839999999`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.2840429200000001`", ",", "4.02682958`", ",", "2.67002883`"}], 
    "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.5139715992840033`*^9, 3.513972760739259*^9, 
  3.513972879589386*^9, 3.5139729538315487`*^9, 3.513975175655528*^9, 
  3.513976990334423*^9, 3.5140488769013453`*^9, 3.514057111026916*^9, 
  3.514058147068818*^9, 3.514058654400281*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "2.7547649040365885`*^-9"}], ",", 
     RowBox[{"-", "4.986658749990547`*^-9"}], ",", 
     RowBox[{"-", "3.373049582700016`*^-10"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"1.5680479391599889`*^-6", ",", 
     RowBox[{"-", "1.403916760234214`*^-6"}], ",", 
     RowBox[{"-", "4.1174737865079525`*^-7"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "0.000018939845379151166`", ",", "0.000010696034005608368`", ",", 
     "8.484565459099258`*^-6"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "1.0656540492398614`*^-6"}], ",", "9.392414084707923`*^-6", 
     ",", "4.497185412288829`*^-6"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.00010725753983997777`", ",", 
     RowBox[{"-", "0.00005343265012491827`"}], ",", 
     RowBox[{"-", "0.00011287044633556675`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "3.8302171323501`*^-6", ",", "4.356413560557826`*^-6", ",", 
     "3.6887004912777854`*^-6"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"5.13710905092335`*^-7", ",", 
     RowBox[{"-", "1.621993745981598`*^-6"}], ",", "8.377348778287885`*^-7"}],
     "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"6.3437250363753556`*^-9", ",", "1.3958582117208282`*^-8", ",", 
     RowBox[{"-", "3.064831323625061`*^-8"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "1.5280621263880079`*^-6"}], ",", "6.305279962592891`*^-6", 
     ",", 
     RowBox[{"-", "8.74118796012624`*^-6"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "3.995765263997697`*^-6"}], ",", "1.3507605758356078`*^-6", 
     ",", "0.00008694037211165462`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.00002109381877590799`"}], ",", "0.00001824674410788596`",
      ",", 
     RowBox[{"-", "5.776054182060619`*^-6"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "0.000010340647090600896`", ",", "0.000011029542640983436`", ",", 
     RowBox[{"-", "0.000019615671424055847`"}]}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
    "0.000013930101824044527`", ",", "0.000012258459514691289`", ",", 
     "3.9368360940095215`*^-6"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "0.00006975571373435407`"}], ",", "0.00002796659320969397`",
      ",", "0.00008062576388212506`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{
     RowBox[{"-", "5.732983856032714`*^-6"}], ",", "7.1045740073216734`*^-6", 
     ",", "0.00001030307119753715`"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.5139715992840033`*^9, 3.513972760739259*^9, 
  3.513972879589386*^9, 3.5139729538315487`*^9, 3.513975175655528*^9, 
  3.513976990334423*^9, 3.5140488769013453`*^9, 3.514057111026916*^9, 
  3.514058147068818*^9, 3.514058654402313*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0", ",", "0", ",", "0"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.5139715992840033`*^9, 3.513972760739259*^9, 
  3.513972879589386*^9, 3.5139729538315487`*^9, 3.513975175655528*^9, 
  3.513976990334423*^9, 3.5140488769013453`*^9, 3.514057111026916*^9, 
  3.514058147068818*^9, 3.514058654405797*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.5139715992840033`*^9, 3.513972760739259*^9, 
  3.513972879589386*^9, 3.5139729538315487`*^9, 3.513975175655528*^9, 
  3.513976990334423*^9, 3.5140488769013453`*^9, 3.514057111026916*^9, 
  3.514058147068818*^9, 3.5140586544072*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}], ",", 
   RowBox[{"{", 
    RowBox[{"0.`", ",", "0.`", ",", "0.`"}], "}"}]}], "}"}]], "Output",
 CellChangeTimes->{3.5139715992840033`*^9, 3.513972760739259*^9, 
  3.513972879589386*^9, 3.5139729538315487`*^9, 3.513975175655528*^9, 
  3.513976990334423*^9, 3.5140488769013453`*^9, 3.514057111026916*^9, 
  3.514058147068818*^9, 3.51405865440897*^9}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[
 RowBox[{
  RowBox[{"(*", " ", 
   RowBox[{
   "GET", " ", "THE", " ", "POSITION", " ", "AND", " ", "VALUE", " ", "OF", 
    " ", "THE", " ", "FORCES", " ", "WHERE", " ", "THE", " ", "MAXIMUM", " ", 
    "DIFFERENCE", " ", "OCCURS"}], " ", "*)"}], "\[IndentingNewLine]", 
  RowBox[{
   RowBox[{"m", "=", 
    RowBox[{"Max", "[", 
     RowBox[{"Flatten", "[", 
      RowBox[{"Abs", "[", 
       RowBox[{"fca", "+", "FCA"}], "]"}], "]"}], "]"}]}], "\n", 
   RowBox[{"p", "=", 
    RowBox[{"Position", "[", 
     RowBox[{
      RowBox[{"Flatten", "[", 
       RowBox[{"Abs", "[", 
        RowBox[{"fca", "+", "FCA"}], "]"}], "]"}], ",", "m"}], "]"}]}], "\n", 
   
   RowBox[{
    RowBox[{"Flatten", "[", "fca", "]"}], "[", 
    RowBox[{"[", 
     RowBox[{"Flatten", "[", 
      RowBox[{"Flatten", "[", "p", "]"}], "]"}], "]"}], "]"}], "\n", 
   RowBox[{
    RowBox[{"Flatten", "[", "FCA", "]"}], "[", 
    RowBox[{"[", 
     RowBox[{"Flatten", "[", 
      RowBox[{"Flatten", "[", "p", "]"}], "]"}], "]"}], "]"}]}]}]], "Input",
 CellChangeTimes->{{3.514052272974889*^9, 3.514052291071145*^9}, 
   3.514052351296309*^9, {3.514052550384282*^9, 3.5140526081344213`*^9}, {
   3.514052689964172*^9, 3.514052712000087*^9}}],

Cell[BoxData["0.00014068493772745327`"], "Output",
 CellChangeTimes->{
  3.514052399013033*^9, {3.514052553529213*^9, 3.5140525648011293`*^9}, 
   3.514052608768815*^9, 3.514058356967655*^9, 3.514058659436166*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"{", "42", "}"}], "}"}]], "Output",
 CellChangeTimes->{
  3.514052399013033*^9, {3.514052553529213*^9, 3.5140525648011293`*^9}, 
   3.514052608768815*^9, 3.514058356967655*^9, 3.514058659437374*^9}],

Cell[BoxData[
 RowBox[{"{", 
  RowBox[{"-", "2.28510522`"}], "}"}]], "Output",
 CellChangeTimes->{
  3.514052399013033*^9, {3.514052553529213*^9, 3.5140525648011293`*^9}, 
   3.514052608768815*^9, 3.514058356967655*^9, 3.514058659438414*^9}],

Cell[BoxData[
 RowBox[{"{", "2.2849645350622727`", "}"}]], "Output",
 CellChangeTimes->{
  3.514052399013033*^9, {3.514052553529213*^9, 3.5140525648011293`*^9}, 
   3.514052608768815*^9, 3.514058356967655*^9, 3.514058659439537*^9}]
}, Open  ]]
},
WindowSize->{1600, 1123},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
FrontEndVersion->"8.0 for Linux x86 (64-bit) (November 7, 2010)",
StyleDefinitions->"Default.nb"
]
(* End of Notebook Content *)

(* Internal cache information *)
(*CellTagsOutline
CellTagsIndex->{}
*)
(*CellTagsIndex
CellTagsIndex->{}
*)
(*NotebookFileOutline
Notebook[{
Cell[545, 20, 3805, 83, 534, "Text"],
Cell[4353, 105, 20016, 509, 1152, "Input"],
Cell[24372, 616, 13399, 340, 601, "Input"],
Cell[37774, 958, 13341, 336, 620, "Input"],
Cell[51118, 1296, 720, 16, 50, "Input"],
Cell[51841, 1314, 2191, 62, 19, "Input",
 CellOpen->False],
Cell[CellGroupData[{
Cell[54057, 1380, 1981, 40, 88, "Input"],
Cell[56041, 1422, 3641, 94, 126, "Output"],
Cell[59685, 1518, 3213, 84, 107, "Output"],
Cell[62901, 1604, 2163, 57, 69, "Output"]
}, Open  ]],
Cell[65079, 1664, 13786, 321, 772, "Input"],
Cell[78868, 1987, 11869, 350, 424, "Input"],
Cell[90740, 2339, 37178, 984, 19, "Input",
 CellOpen->False],
Cell[127921, 3325, 7579, 216, 19, "Input",
 CellOpen->False],
Cell[135503, 3543, 14550, 406, 19, "Input",
 CellOpen->False],
Cell[150056, 3951, 20869, 567, 19, "Input",
 CellOpen->False],
Cell[170928, 4520, 11073, 313, 19, "Input",
 CellOpen->False],
Cell[182004, 4835, 10858, 267, 19, "Input",
 CellOpen->False],
Cell[192865, 5104, 15476, 351, 560, "Input"],
Cell[208344, 5457, 7747, 202, 404, "Input"],
Cell[216094, 5661, 2540, 44, 107, "Input"],
Cell[CellGroupData[{
Cell[218659, 5709, 6331, 168, 164, "Input"],
Cell[224993, 5879, 1319, 19, 33, "Output"],
Cell[226315, 5900, 4870, 114, 126, "Output"],
Cell[231188, 6016, 4454, 106, 107, "Output"],
Cell[235645, 6124, 3085, 72, 50, "Output"]
}, Open  ]],
Cell[238745, 6199, 6160, 157, 19, "Input",
 CellOpen->False],
Cell[CellGroupData[{
Cell[244930, 6360, 1497, 37, 88, "Input"],
Cell[246430, 6399, 590, 9, 30, "Output"],
Cell[247023, 6410, 590, 9, 30, "Output"],
Cell[247616, 6421, 570, 9, 30, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[248223, 6435, 1499, 27, 50, "Input"],
Cell[249725, 6464, 1860, 27, 30, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[251622, 6496, 892, 18, 202, "Input"],
Cell[252517, 6516, 3889, 100, 126, "Output"],
Cell[256409, 6618, 3734, 95, 126, "Output"],
Cell[260146, 6715, 4192, 105, 187, "Output"],
Cell[264341, 6822, 3471, 92, 107, "Output"],
Cell[267815, 6916, 3308, 85, 107, "Output"],
Cell[271126, 7003, 3602, 90, 143, "Output"],
Cell[274731, 7095, 2102, 58, 50, "Output"],
Cell[276836, 7155, 2256, 58, 69, "Output"],
Cell[279095, 7215, 2257, 58, 69, "Output"]
}, Open  ]],
Cell[CellGroupData[{
Cell[281389, 7278, 1227, 32, 107, "Input"],
Cell[282619, 7312, 213, 3, 30, "Output"],
Cell[282835, 7317, 237, 5, 30, "Output"],
Cell[283075, 7324, 241, 5, 30, "Output"],
Cell[283319, 7331, 231, 4, 30, "Output"]
}, Open  ]]
}
]
*)

(* End of internal cache information *)
